unit PasCodegen;

interface

uses
  PasTypes, SysUtils, PasCpool, PasErrors, PasScanner;

var
  class_name: string[16];

  procedure bytecode_append(code: bytecode; c: Jmnem);
  procedure bytecode_append_short_int(code: bytecode; s: shortint);
  procedure initialize_variable(item: block; variable: identifier; name: string; is_static: integer; class_name: string);
  procedure create_record_class(records: types);
  function  block_create(parent_block: block; block_name: string): block;
  procedure create_constant_bytecode(item: identifier; code: bytecode);
  procedure create_variable_bytecode(item: identifier; code: bytecode; name: string; is_field: integer);
  procedure adjust_indices(code: bytecode; dimension: types);
  procedure create_put_variable_bytecode(item: identifier; code: bytecode; name: string; is_field: integer);

implementation

function bytecode_create(): bytecode;
var
  new_bytecode: bytecode;
begin
	New(new_bytecode);
	new_bytecode^.bytecode_allocated_size := 0;
	new_bytecode^.bytecode_pos := 0;
	Result:= new_bytecode;
end;

function block_create(parent_block: block; block_name: string): block;
var
  new_block: block;
begin
	New(new_block);
	new_block^.names := name_table_create();
	new_block^.parent_block := parent_block;
	new_block^.block_name := block_name;
	new_block^.next_variable_index := 0;
	new_block^.next_parameter_index := 0;
	new_block^.code := bytecode_create();
	new_block^.children_count := 0;
	result := new_block;
end;

procedure bytecode_append(code: bytecode; c: Jmnem);
begin
	if (code^.bytecode_pos = code^.bytecode_allocated_size) then inc(code^.bytecode_allocated_size,128);
  SetLength(code^.bytecode,length(code^.bytecode)+1);
	code^.bytecode[code^.bytecode_pos] := Char(c);
	Inc(code^.bytecode_pos);
end;

procedure bytecode_append_short_int(code: bytecode; s: shortint);
var
	ch: char;
begin
	ch := char(s shr 8);
	bytecode_append(code, Jmnem(ch));
	ch := char(s);
	bytecode_append(code, Jmnem(ch));
end;

procedure create_variable_bytecode(item: identifier; code: bytecode; name: string; is_field: integer);
var
	index: integer;
	item_type: types;
  class_index,
	method_index,
	constructor_index,
	copy_index: integer;
  type_name: string[16];
	copy_sig: string[64];
  type_descriptor: array [0..128] of char;
  unit_name: string;
begin

	if ((is_field<>0) or (item^.unit_function = 1)) then begin

		if (item^.identifier_class = variable_name) then begin

			name:=lowercase(name);
			bytecode_append(code, getstatic_);
			get_field_descriptor(item^.variable_type, type_descriptor);

			if (item^.unit_function = 0) then begin
				if (compiling_unit = 0) then
					bytecode_append_short_int(code, cp_add_fieldref('M', name, type_descriptor))
				else
					bytecode_append_short_int(code, cp_add_fieldref(str_program_name, name, type_descriptor));
			end	else begin
				if (item^.container_unit^.is_library = 0) then
					unit_name := ''
				else
					unit_name := 'Lib_';
				//lowercase(string_get_cstr(item->container_unit->unit_name));
				unit_name:=unit_name+item^.container_unit^.unit_name+#0;
				bytecode_append_short_int(code, cp_add_fieldref(unit_name, name, type_descriptor));
      end;
    end;
		Exit;
  end;

	case item^.identifier_class of
    variable_name:
                  begin
                    index := item^.variable_index;
                    item_type := item^.variable_type;
                  end;
    parameter_name:
                  begin
                    index := item^.parameter_index;
                    item_type := item^.parameter_type;
                  end;
    function_name:
                  begin
                    index := type_list_length(item^.parameters);
                    item_type := item^.return_type;
                  end;
	else
		MakeError(MSG_444, '');
	end;

	case (item_type^.type_class) of

      integer_type,
      char_type,
      boolean_type:
          begin
            case index of
              0:  bytecode_append(code, iload_0_);
              1:  bytecode_append(code, iload_1_);
              2:  bytecode_append(code, iload_2_);
              3:  bytecode_append(code, iload_3_);
            else
              bytecode_append(code, iload_);
              bytecode_append(code, Jmnem(index));
            end;
          end;

      real_type:
          begin
            if (mathType = 1) then begin
              case index of
                0:  bytecode_append(code, iload_0_);
                1:  bytecode_append(code, iload_1_);
                2:  bytecode_append(code, iload_2_);
                3:  bytecode_append(code, iload_3_);
              else
                bytecode_append(code, iload_);
                bytecode_append(code, Jmnem(index));
              end;
            end else begin
              case index of
                0:  bytecode_append(code, aload_0_);
                1:  bytecode_append(code, aload_1_);
                2:  bytecode_append(code, aload_2_);
                3:  bytecode_append(code, aload_3_);
              else
                bytecode_append(code, aload_);
                bytecode_append(code, Jmnem(index));
              end;
            end;
          end;

      string_type,
      record_type,
      array_type,
      image_type,
      command_type,
      stream_type,
      record_store_type,
      http_type,
      alert_type:
          begin
            case index of
              0:  bytecode_append(code, aload_0_);
              1:  bytecode_append(code, aload_1_);
              2:  bytecode_append(code, aload_2_);
              3:  bytecode_append(code, aload_3_);
            else
              bytecode_append(code, aload_);
              bytecode_append(code, Jmnem(index));
            end;
          end;

	else
		MakeError(MSG_015,'');
  end;

end;

procedure create_put_variable_bytecode(item: identifier; code: bytecode; name: string; is_field: integer);
var
	index: integer;
	item_type: types;
  class_index,
	method_index,
	constructor_index,
	copy_index: integer;
  type_name: string[16];
	copy_sig: string[64];
  type_descriptor: array [0..128] of char;
  unit_name: string;
begin

	case item^.identifier_class of
    variable_name:
                  begin
                    index := item^.variable_index;
                    item_type := item^.variable_type;
                  end;
    parameter_name:
                  begin
                    index := item^.parameter_index;
                    item_type := item^.parameter_type;
                  end;
    function_name:
                  begin
                    index := type_list_length(item^.parameters);
                    is_field := 0;
                    item_type := item^.return_type;
                  end;
	else
		MakeError(MSG_444, '');
	end;

	if (item_type^.type_class = string_type) then begin
		class_index := cp_add_class('java/lang/String');
		method_index := cp_add_methodref('java/lang/String', '<init>', '(Ljava/lang/String;)V');
		bytecode_append(code, new_);
		bytecode_append_short_int(code, class_index);
		bytecode_append(code, dup_x1_);
		bytecode_append(code, swap_);
		bytecode_append(code, invokespecial_);
		bytecode_append_short_int(code, method_index);
	end;

	if ((item_type^.type_class = real_type) and (mathType <> 1)) then begin
		usesFloat := 1;
		class_index := cp_add_class('Real');
		method_index := cp_add_methodref('Real', '<init>', '(LReal;)V');
		bytecode_append(code, new_);
		bytecode_append_short_int(code, class_index);
		bytecode_append(code, dup_x1_);
		bytecode_append(code, swap_);
		bytecode_append(code, invokespecial_);
		bytecode_append_short_int(code, method_index);
	end;

	if (item_type^.type_class = record_type) then begin
		type_name:= 'R_'+IntToStr(item_type^.unique_record_ID);
		copy_sig:='(L'+type_name+';)L'+type_name+';';
		class_index := cp_add_class(type_name);
		constructor_index := cp_add_methodref(type_name, '<init>', '()V');
		copy_index := cp_add_methodref(type_name, 'Copy', copy_sig);
		bytecode_append(code, new_);
		bytecode_append_short_int(code, class_index);
		bytecode_append(code, dup_x1_);
		bytecode_append(code, invokespecial_);
		bytecode_append_short_int(code, constructor_index);
		bytecode_append(code, invokevirtual_);
		bytecode_append_short_int(code, copy_index);
	end;

	if ((is_field<>0) or (item^.unit_function = 1)) then begin
		if (item^.identifier_class = variable_name) then begin
			name:=lowercase(name);
			bytecode_append(code, putstatic_);
			get_field_descriptor(item^.variable_type, type_descriptor);

			if (item^.unit_function = 0) then begin
				if (compiling_unit = 0) then
					bytecode_append_short_int(code, cp_add_fieldref('M', name, type_descriptor))
				else
					bytecode_append_short_int(code, cp_add_fieldref(str_program_name, name, type_descriptor));
			end else begin
				if (item^.container_unit^.is_library = 0) then
					unit_name := ''
				else
					unit_name := 'Lib_';
				//lowercase(string_get_cstr(item^.container_unit^.unit_name));
				unit_name:=unit_name+item^.container_unit^.unit_name+#0;
				bytecode_append_short_int(code, cp_add_fieldref(unit_name, name, type_descriptor));
			end;
		end;
		Exit;
	end;

	case (item_type^.type_class) of

      integer_type,
      char_type,
      boolean_type:
          begin
            case index of
              0:  bytecode_append(code, istore_0_);
              1:  bytecode_append(code, istore_1_);
              2:  bytecode_append(code, istore_2_);
              3:  bytecode_append(code, istore_3_);
            else
              bytecode_append(code, istore_);
              bytecode_append(code, Jmnem(index));
            end;
          end;

      real_type:
          begin
            if (mathType = 1) then begin
              case index of
                0:  bytecode_append(code, istore_0_);
                1:  bytecode_append(code, istore_1_);
                2:  bytecode_append(code, istore_2_);
                3:  bytecode_append(code, istore_3_);
              else
                bytecode_append(code, istore_);
                bytecode_append(code, Jmnem(index));
              end;
            end else begin
              case index of
                0:  bytecode_append(code, astore_0_);
                1:  bytecode_append(code, astore_1_);
                2:  bytecode_append(code, astore_2_);
                3:  bytecode_append(code, astore_3_);
              else
                bytecode_append(code, astore_);
                bytecode_append(code, Jmnem(index));
              end;
            end;
          end;

      string_type,
      record_type,
      array_type,
      image_type,
      command_type,
      stream_type,
      record_store_type,
      http_type,
      alert_type:
          begin
            case index of
              0:  bytecode_append(code, astore_0_);
              1:  bytecode_append(code, astore_1_);
              2:  bytecode_append(code, astore_2_);
              3:  bytecode_append(code, astore_3_);
            else
              bytecode_append(code, astore_);
              bytecode_append(code, Jmnem(index));
            end;
          end;

	else
		MakeError(MSG_015,'');
  end;

end;

procedure initialize_variable(item: block; variable: identifier; name: string; is_static: integer; class_name: string);
var
	storeIntoField: integer;
  method_index,
	class_index: integer;
  count, size: integer;
	it: type_list;
  descriptor: array [0..1024] of char;
  descriptor128: array [0..128] of char;
  descriptor512: array [0..512] of char;
  field_index: integer;
  c: shortint;
	num: shortint;
	offsets: array of integer;
  offset: shortint;
begin

  storeIntoField := 0;
	if variable^.belongs_to_program_block<>0 then storeIntoField := 1;

	case variable^.variable_type^.type_class of

	integer_type,
	char_type,
	boolean_type:
		begin
			bytecode_append(item^.code, iconst_0_);

			if storeIntoField<>0 then begin
				case variable^.variable_index of
          0:  bytecode_append(item^.code, istore_0_);
          1:  bytecode_append(item^.code, istore_1_);
          2:  bytecode_append(item^.code, istore_2_);
          3:  bytecode_append(item^.code, istore_3_);
				else
					bytecode_append(item^.code, istore_);
					bytecode_append(item^.code, Jmnem(variable^.variable_index));
				end;
			end;

		end;

	real_type:
		begin

			if mathType = 1 then begin

				bytecode_append(item^.code, iconst_0_);

				if storeIntoField<>0 then begin
					case (variable^.variable_index) of
            0:  bytecode_append(item^.code, istore_0_);
            1:  bytecode_append(item^.code, istore_1_);
            2:  bytecode_append(item^.code, istore_2_);
            3:  bytecode_append(item^.code, istore_3_);
					else
						bytecode_append(item^.code, istore_);
						bytecode_append(item^.code, Jmnem(variable^.variable_index));
          end;
        end;

      end else begin

				usesFloat := 1;
				
				class_index := cp_add_class('Real');
				method_index := cp_add_methodref('Real', '<init>', '()V');

				bytecode_append(item^.code, new_);
				bytecode_append_short_int(item^.code, shortint(class_index));
				bytecode_append(item^.code, dup_);
				bytecode_append(item^.code, invokespecial_);
				bytecode_append_short_int(item^.code, shortint(method_index));

				if storeIntoField<>0 then begin
					case (variable^.variable_index) of
            0:  bytecode_append(item^.code, astore_0_);
            1:  bytecode_append(item^.code, astore_1_);
            2:  bytecode_append(item^.code, astore_2_);
            3:  bytecode_append(item^.code, astore_3_);
					else
						bytecode_append(item^.code, astore_);
						bytecode_append(item^.code, Jmnem(variable^.variable_index));
          end;
        end;

      end;

		end;


	string_type:
		begin
			
			class_index := cp_add_class('java/lang/String');
			method_index := cp_add_methodref('java/lang/String', '<init>', '()V');

			bytecode_append(item^.code, new_);
			bytecode_append_short_int(item^.code, shortint(class_index));
			bytecode_append(item^.code, dup_);
			bytecode_append(item^.code, invokespecial_);
			bytecode_append_short_int(item^.code, shortint(method_index));

			if storeIntoField<>0 then begin
				case (variable^.variable_index) of
          0:  bytecode_append(item^.code, astore_0_);
          1:  bytecode_append(item^.code, astore_1_);
          2:  bytecode_append(item^.code, astore_2_);
          3:  bytecode_append(item^.code, astore_3_);
				else
					bytecode_append(item^.code, astore_);
					bytecode_append(item^.code, Jmnem(variable^.variable_index));
        end;
      end;

    end;

	image_type:
		begin
			method_index := cp_add_methodref('javax/microedition/lcdui/Image', 'createImage', '(II)Ljavax/microedition/lcdui/Image;');
			
			bytecode_append(item^.code, iconst_1_);
			bytecode_append(item^.code, iconst_1_);
			bytecode_append(item^.code, invokestatic_);
			bytecode_append_short_int(item^.code, method_index);

			if storeIntoField<>0 then begin
				case (variable^.variable_index) of
          0:  bytecode_append(item^.code, astore_0_);
          1:  bytecode_append(item^.code, astore_1_);
          2:  bytecode_append(item^.code, astore_2_);
          3:  bytecode_append(item^.code, astore_3_);
				else
					bytecode_append(item^.code, astore_);
					bytecode_append(item^.code, Jmnem(variable^.variable_index));
        end;
      end;

    end;

  stream_type:
		begin

			bytecode_append(item^.code, aconst_null_);

			if storeIntoField<>0 then begin
			  case (variable^.variable_index) of
				0:  bytecode_append(item^.code, astore_0_);
				1:  bytecode_append(item^.code, astore_1_);
				2:  bytecode_append(item^.code, astore_2_);
				3:  bytecode_append(item^.code, astore_3_);
				else
					bytecode_append(item^.code, astore_);
					bytecode_append(item^.code, Jmnem(variable^.variable_index));
        end;
      end;

    end;

	command_type:
		begin

			method_index := cp_add_methodref('javax/microedition/lcdui/Command', '<init>', '(Ljava/lang/String;II)V');
			
			bytecode_append(item^.code, new_);
			bytecode_append_short_int(item^.code, cp_add_class('javax/microedition/lcdui/Command'));
			bytecode_append(item^.code, dup_);
			
			bytecode_append(item^.code, new_);
			bytecode_append_short_int(item^.code, cp_add_class('java/lang/String'));
			bytecode_append(item^.code, dup_);
			bytecode_append(item^.code, invokespecial_);
			bytecode_append_short_int(item^.code, cp_add_methodref('java/lang/String', '<init>', '()V'));

			bytecode_append(item^.code, iconst_1_);
			bytecode_append(item^.code, iconst_1_);
			bytecode_append(item^.code, invokespecial_);
			bytecode_append_short_int(item^.code, method_index);

			if storeIntoField<>0 then begin
				case (variable^.variable_index) of
          0:  bytecode_append(item^.code, astore_0_);
          1:  bytecode_append(item^.code, astore_1_);
          2:  bytecode_append(item^.code, astore_2_);
          3:  bytecode_append(item^.code, astore_3_);
				else
					bytecode_append(item^.code, astore_);
					bytecode_append(item^.code, Jmnem(variable^.variable_index));
        end;
      end;

    end;

  http_type:
		begin

			method_index := cp_add_methodref('H', '<init>', '()V');
			usesHttp := 1;
			
			bytecode_append(item^.code, new_);
			bytecode_append_short_int(item^.code, cp_add_class('H'));
			bytecode_append(item^.code, dup_);

			bytecode_append(item^.code, invokespecial_);
			bytecode_append_short_int(item^.code, method_index);

			if storeIntoField<>0 then begin
				case (variable^.variable_index) of
				0:  bytecode_append(item^.code, astore_0_);
				1:  bytecode_append(item^.code, astore_1_);
				2:  bytecode_append(item^.code, astore_2_);
				3:  bytecode_append(item^.code, astore_3_);
				else
					bytecode_append(item^.code, astore_);
					bytecode_append(item^.code, Jmnem(variable^.variable_index));
        end;
      end;

    end;

	record_store_type:  usesRecordStore := 1;

	array_type:
		begin
			count := 0;
			it := variable^.variable_type^.dimensions_list;
			get_field_descriptor(variable^.variable_type, descriptor);

			while (it <> NIL) do begin
        if (it^.data <> NIL) then begin
					size := it^.data^.last_element - it^.data^.first_element + 1;
					if (size <= 0) then MakeError(MSG_437,'');
					if (size < 127) then begin
						bytecode_append(item^.code, bipush_);
						bytecode_append(item^.code, Jmnem(size));
					end	else if (size < 16000) then begin
						bytecode_append(item^.code, sipush_);
						bytecode_append_short_int(item^.code, shortint(size));
					end	else
						MakeError(MSG_438, '');
        end;
				it := it^.next;
        Inc(count);
			end;

			bytecode_append(item^.code, multianewarray_);
			bytecode_append_short_int(item^.code, cp_add_class(descriptor));
			bytecode_append(item^.code, Jmnem(count));

			// NEW:
			create_put_variable_bytecode(variable, item^.code, name, storeIntoField);

			if  ((variable^.variable_type^.element_type^.type_class = record_type)
				or (variable^.variable_type^.element_type^.type_class = string_type)
				or ((variable^.variable_type^.element_type^.type_class = real_type) and (mathType <> 1))
				or (variable^.variable_type^.element_type^.type_class = image_type)
				or (variable^.variable_type^.element_type^.type_class = command_type)) then begin

        c:=0;
				it := variable^.variable_type^.dimensions_list;

				while (it <> NIL) do begin
					it := it^.next;
					Inc(c);
				end;

				if variable^.variable_type^.element_type^.type_class = real_type then usesFloat := 1;

				num := c;

				it := variable^.variable_type^.dimensions_list;
				c := 0;
				while (it <> NIL) do begin
					if (it^.data <> NIL) then begin
						bytecode_append(item^.code, getstatic_);
						bytecode_append_short_int(item^.code, cp_add_fieldref('M', 'IC', '[I'));
						bytecode_append(item^.code, sipush_);
						bytecode_append_short_int(item^.code, c);
						bytecode_append(item^.code, iconst_0_);
						bytecode_append(item^.code, iastore_);
            SetLength(offsets,Length(offsets)+1);
						offsets[c] := item^.code^.bytecode_pos;
					end;
					Inc(c);
					it := it^.next;
				end;

				for c:=0 to num do begin
					if ((c < (num-1)) or (c = 0)) then
						create_variable_bytecode(variable, item^.code, name, storeIntoField);
					bytecode_append(item^.code, getstatic_);
					bytecode_append_short_int(item^.code, cp_add_fieldref('M', 'IC', '[I'));
					bytecode_append(item^.code, sipush_);
					bytecode_append_short_int(item^.code, c);
					bytecode_append(item^.code, iaload_);
					if (c < (num-1)) then
						bytecode_append(item^.code, aaload_);
				end;

				get_field_descriptor(variable^.variable_type^.element_type, descriptor128);
				descriptor128[length(descriptor128)-1] := #0;
				
				if (variable^.variable_type^.element_type^.type_class = image_type) then begin
					bytecode_append(item^.code, iconst_1_);
					bytecode_append(item^.code, iconst_1_);
					bytecode_append(item^.code, invokestatic_);
					bytecode_append_short_int(item^.code, cp_add_methodref('javax/microedition/lcdui/Image', 'createImage', '(II)Ljavax/microedition/lcdui/Image;'));
				end else if (variable^.variable_type^.element_type^.type_class = command_type) then begin
					method_index := cp_add_methodref('javax/microedition/lcdui/Command', '<init>', '(Ljava/lang/String;II)V');

					bytecode_append(item^.code, new_);
					bytecode_append_short_int(item^.code, cp_add_class('javax/microedition/lcdui/Command'));
					bytecode_append(item^.code, dup_);

					bytecode_append(item^.code, new_);
					bytecode_append_short_int(item^.code, cp_add_class('java/lang/String'));
					bytecode_append(item^.code, dup_);
					bytecode_append(item^.code, invokespecial_);
					bytecode_append_short_int(item^.code, cp_add_methodref('java/lang/String', '<init>', '()V'));

					bytecode_append(item^.code, iconst_1_);
					bytecode_append(item^.code, iconst_1_);
					bytecode_append(item^.code, invokespecial_);
					bytecode_append_short_int(item^.code, method_index);
				end else begin
					bytecode_append(item^.code, new_);
					bytecode_append_short_int(item^.code, cp_add_class(descriptor128+1));
					bytecode_append(item^.code, dup_);
					bytecode_append(item^.code, invokespecial_);
					bytecode_append_short_int(item^.code, cp_add_methodref(descriptor128+1, '<init>', '()V'));
				end;

				bytecode_append(item^.code, aastore_);
				it := variable^.variable_type^.dimensions_list;
				c := num-1;
				while (it <> NIL) do begin
					if (it^.data <> NIL) then begin
						bytecode_append(item^.code, getstatic_);
						bytecode_append_short_int(item^.code, cp_add_fieldref('M', 'IC', '[I'));
						bytecode_append(item^.code, dup_);
						bytecode_append(item^.code, sipush_);
						bytecode_append_short_int(item^.code, c);
						bytecode_append(item^.code, iaload_);
						bytecode_append(item^.code, iconst_1_);
						bytecode_append(item^.code, iadd_);
						bytecode_append(item^.code, sipush_);
						bytecode_append_short_int(item^.code, c);
						bytecode_append(item^.code, swap_);
						bytecode_append(item^.code, iastore_);

						bytecode_append(item^.code, getstatic_);
						bytecode_append_short_int(item^.code, cp_add_fieldref('M', 'IC', '[I'));
						bytecode_append(item^.code, sipush_);
						bytecode_append_short_int(item^.code, c);
						bytecode_append(item^.code, iaload_);
						bytecode_append(item^.code, sipush_);
						bytecode_append_short_int(item^.code,
							it^.data^.last_element-it^.data^.first_element);
						bytecode_append(item^.code, if_icmple_);
						offset := offsets[c] - item^.code^.bytecode_pos + 1;
						bytecode_append_short_int(item^.code, offset);
					end;
					Dec(c);
					it := it^.next;
				end;
			end;
		end;

	record_type:
		begin

			get_field_descriptor(variable^.variable_type, descriptor);
			descriptor[length(descriptor)-1]:=#0;
			bytecode_append(item^.code, new_);
			bytecode_append_short_int(item^.code, cp_add_class(descriptor+1));
			bytecode_append(item^.code, dup_);
			bytecode_append(item^.code, invokespecial_);
			bytecode_append_short_int(item^.code, cp_add_methodref(descriptor+1, '<init>', '()V'));

			if storeIntoField<>0 then begin
				case variable^.variable_index of
          0:  bytecode_append(item^.code, astore_0_);
          1:  bytecode_append(item^.code, astore_1_);
          2:  bytecode_append(item^.code, astore_2_);
          3:  bytecode_append(item^.code, astore_3_);
				else
					bytecode_append(item^.code, astore_);
					bytecode_append(item^.code, Jmnem(variable^.variable_index));
        end;
      end;
		end;

	error_type:;

	else
    MakeError(MSG_014,'');
  end;

	if ((storeIntoField<>0)
		and (variable^.variable_type^.type_class <> record_store_type)
		and (variable^.variable_type^.type_class <> array_type)) then begin
		get_field_descriptor(variable^.variable_type, descriptor512);
		name:=lowercase(name);
		field_index := cp_add_fieldref(class_name, name, descriptor512);
		if (is_static<>0) then
			bytecode_append(item^.code, putstatic_)
		else
			bytecode_append(item^.code, putfield_);
		bytecode_append_short_int(item^.code, shortint(field_index));
  end;
  
end;

function cp_create_new_entry(): constant_pool;
begin
	Inc(constant_pool_size);
  SetLength(constants,Length(constants)+1);
	Result:= @constants[constant_pool_size-1];
end;

function cp_add_nameandtype(name,types: string): integer;
var
	i,name_index,
	type_index: integer;
	entry: constant_pool;
begin

	name_index := cp_add_utf8(name);
	type_index := cp_add_utf8(types);

	for i:=0 to constant_pool_size-1 do begin
		if ((constants[i].tag = 12)
    and (constants[i].param1 = name_index)
    and (constants[i].param2 = type_index)) then begin
			Result:= i+1;
      Exit;
		end;
  end;

	entry := cp_create_new_entry();
	entry^.tag := 12;
	entry^.param1 := name_index;
	entry^.param2 := type_index;

	Result:= constant_pool_size;

end;

procedure create_record_fields(records: types);
var
	names: string_list;
	types: type_list;
	descriptor: array [0..128] of char;
begin
	names := records^.elements_name_list;
	types := records^.elements_type_list;
	while (names <> NIL) do begin
		if names^.data <> '' then begin
			names^.data:=lowercase(names^.data);
			get_field_descriptor(types^.data, descriptor);
			cp_add_nameandtype(names^.data, descriptor);
    end;
		names := names^.next;
		types := types^.next;
  end;
end;

procedure create_init_method(records: types; code: bytecode);
var
	type_it: type_list;
	name_it: string_list;
	counter: integer;
  code_wrapper: block;
	type_wrapper: identifier;
begin

	counter := 0;

	type_it := records^.elements_type_list;
	name_it := records^.elements_name_list;

	bytecode_append(code, aload_0_);
	bytecode_append(code, invokespecial_);
	bytecode_append_short_int(code, cp_add_methodref('java/lang/Object', '<init>', '()V'));


	while (type_it <> NIL) do begin
		if ((type_it^.data <> NIL) and (name_it^.data <> '')) then begin
			
			code_wrapper := block_create(NIL, '');
			code_wrapper^.code := code;

			type_wrapper := identifier_create();
			type_wrapper^.identifier_class := variable_name;
			type_wrapper^.variable_type := type_it^.data;
			type_wrapper^.belongs_to_program_block := 1;

			bytecode_append(code, aload_0_);

			initialize_variable(code_wrapper, type_wrapper, name_it^.data, 0, class_name);

			if (type_it^.data^.type_class = array_type) then
				MakeError(MSG_446, name_it^.data);

			code_wrapper^.code := bytecode_create();
			type_wrapper^.variable_type := type_create();
    end;

		type_it := type_it^.next;
		if (name_it <> NIL) then name_it := name_it^.next;

		Inc(counter);

  end;

	bytecode_append(code, return_);

end;

procedure create_copy_method(records: types; code: bytecode);
var
	type_name: string[64];
	type_it: type_list;
	name_it: string_list;
	counter: integer;
  copy_sig: string[64];
	type_descriptor: array [0..128] of char;
begin

  counter := 0;
	type_name:='R_'+IntToStr(records^.unique_record_ID);
	type_it := records^.elements_type_list;
	name_it := records^.elements_name_list;

	while type_it <> NIL do begin
		if ((type_it^.data <> NIL) and (name_it^.data <> '')) then begin

			name_it^.data:=lowercase(name_it^.data);

			if (type_it^.data^.type_class = record_type) then begin
				get_field_descriptor(type_it^.data, type_descriptor);
				copy_sig:='(LR_'+IntToStr(records^.unique_record_ID)+';)LR_'+IntToStr(records^.unique_record_ID)+';';
				bytecode_append(code, aload_0_);
				bytecode_append(code, getfield_);
				bytecode_append_short_int(code, cp_add_fieldref(type_name, name_it^.data, type_descriptor));
        bytecode_append(code, aload_1_);
				bytecode_append(code, getfield_);
				bytecode_append_short_int(code, cp_add_fieldref(type_name, name_it^.data, type_descriptor));
				bytecode_append(code, invokevirtual_);
				bytecode_append_short_int(code, cp_add_methodref(type_name, 'Copy', copy_sig));
      end	else if (type_it^.data^.type_class = string_type) then begin
				bytecode_append(code, aload_0_);
				bytecode_append(code, aload_1_);
				bytecode_append(code, getfield_);
				bytecode_append_short_int(code, cp_add_fieldref(type_name, name_it^.data, 'Ljava/lang/String;'));
				bytecode_append(code, new_);
				bytecode_append_short_int(code, cp_add_class('java/lang/String'));
				bytecode_append(code, dup_x1_);
				bytecode_append(code, swap_);
				bytecode_append(code, invokespecial_);
				bytecode_append_short_int(code, cp_add_methodref('java/lang/String', '<init>', '(Ljava/lang/String;)V'));
				bytecode_append(code, putfield_);
				bytecode_append_short_int(code, cp_add_fieldref(type_name, name_it^.data, 'Ljava/lang/String;'));
			end	else if ((type_it^.data^.type_class = real_type) and (mathType <> 1)) then begin
				bytecode_append(code, aload_0_);
				bytecode_append(code, aload_1_);
				bytecode_append(code, getfield_);
				bytecode_append_short_int(code, cp_add_fieldref(type_name, name_it^.data, 'LReal;'));
				bytecode_append(code, new_);
				bytecode_append_short_int(code, cp_add_class('Real'));
				bytecode_append(code, dup_x1_);
				bytecode_append(code, swap_);
				bytecode_append(code, invokespecial_);
				bytecode_append_short_int(code, cp_add_methodref('Real', '<init>', '(LReal;)V'));
				bytecode_append(code, putfield_);
				bytecode_append_short_int(code, cp_add_fieldref(type_name, name_it^.data, 'LReal;'));
      end	else begin
				get_field_descriptor(type_it^.data, type_descriptor);
				bytecode_append(code, aload_0_);
				bytecode_append(code, aload_1_);
				bytecode_append(code, getfield_);
				bytecode_append_short_int(code, cp_add_fieldref(type_name, name_it^.data, type_descriptor));
				bytecode_append(code, putfield_);
				bytecode_append_short_int(code, cp_add_fieldref(type_name, name_it^.data, type_descriptor));
			end;

    end;

		type_it := type_it^.next;
		if (name_it <> NIL) then name_it := name_it^.next;

		Inc(counter);
	end;

	bytecode_append(code, aload_0_);
	bytecode_append(code, areturn_);

end;

procedure create_constant_bytecode(item: identifier; code: bytecode);
var
	cp_index: integer;
  cst: real;
	csti,cstf,i: integer;
  number: string[64];
begin
	case item^.constant_type^.type_class of

	integer_type:
      begin
        cp_index := cp_add_integer(item^.constant_int_value);
        if (cp_index <= 255) then begin
          bytecode_append(code, ldc_);
          bytecode_append(code, Jmnem(cp_index));
        end	else begin
          bytecode_append(code, ldc_w_);
          bytecode_append_short_int(code, shortint(cp_index));
        end;
      end;

	real_type:
      begin
        cst := item^.constant_real_value;
        csti := round(cst);
        cstf := round((cst - round(cst))*100000);
        usesFloat := 1;

        if mathType = 1 then begin
          if (abs(csti) < 127) then begin
            bytecode_append(code, bipush_);
            bytecode_append(code, Jmnem(csti));
          end	else if (abs(csti) < 15000) then begin
            bytecode_append(code, sipush_);
            bytecode_append_short_int(code, csti);
          end else begin
            bytecode_append(code, ldc_w_);
            bytecode_append_short_int(code, cp_add_integer(csti));
          end;

          if (abs(cstf) < 127) then begin
            bytecode_append(code, bipush_);
            bytecode_append(code, Jmnem(cstf));
          end	else if (abs(cstf) < 15000) then begin
            bytecode_append(code, sipush_);
            bytecode_append_short_int(code, cstf);
          end	else begin
            bytecode_append(code, ldc_w_);
            bytecode_append_short_int(code, cp_add_integer(cstf));
          end;

          bytecode_append(code, invokestatic_);
          bytecode_append_short_int(code, cp_add_methodref('F', 'fI', '(II)I'));
        end	else begin

          bytecode_append(code, new_);
          bytecode_append_short_int(code, cp_add_class('Real'));
          bytecode_append(code, dup_);

          Val(number,TokenReal,i);
          bytecode_append(code, ldc_w_);
          bytecode_append_short_int(code, cp_add_string(number));

          bytecode_append(code, invokespecial_);
          bytecode_append_short_int(code, cp_add_methodref('Real', '<init>', '(Ljava/lang/String;)V'));
        end;
      end;

	char_type:
      begin
        cp_index := cp_add_integer(item^.constant_int_value);
        if (cp_index <= 255) then begin
          bytecode_append(code, ldc_);
          bytecode_append(code, Jmnem(cp_index));
        end	else begin
          bytecode_append(code, ldc_w_);
          bytecode_append_short_int(code, shortint(cp_index));
        end;
      end;

	boolean_type:
      begin
        if item^.constant_int_value<>0 then
          bytecode_append(code, iconst_m1_)
        else
          bytecode_append(code, iconst_0_);
      end;

	string_type:
      begin
        cp_index := cp_add_string(item^.constant_string_value);
        if cp_index <= 255 then begin
          bytecode_append(code, ldc_);
          bytecode_append(code, Jmnem(cp_index));
        end	else begin
          bytecode_append(code, ldc_w_);
          bytecode_append_short_int(code, shortint(cp_index));
        end;
      end;
  end;

end;

procedure adjust_indices(code: bytecode; dimension: types);
begin
	if dimension = NIL then Exit;

	case dimension^.first_element of
    0:  bytecode_append(code, iconst_0_);
    1:  bytecode_append(code, iconst_1_);
    2:  bytecode_append(code, iconst_2_);
    3:  bytecode_append(code, iconst_3_);
    4:  bytecode_append(code, iconst_4_);
    5:  bytecode_append(code, iconst_5_);
	else
		if (dimension^.first_element < 128) then begin
			bytecode_append(code, bipush_);
			bytecode_append(code, Jmnem(dimension^.first_element));
		end	else begin
			bytecode_append(code, sipush_);
			bytecode_append_short_int(code, dimension^.first_element);
    end;
  end;

	bytecode_append(code, isub_);
end;

procedure create_record_class(records: types);
var
	class_file_name: string[16];
	output_file_name: string;
	copy_sig: string[64];

	record_file: FILE;

	init_method: bytecode;
	copy_method: bytecode;

	global_cp: array of constant_pool;
	global_cp_count: integer;
begin
	global_cp := @constants;
	global_cp_count := constant_pool_size;

	constants := NIL;
	constant_pool_size := 0;

	if (records^.type_class <> record_type) then MakeError(MSG_012,'');

  class_name:= 'R_'+IntToStr(records^.unique_record_ID);
  class_file_name:=class_name+'.class';
	output_file_name:=output_path+class_file_name;

	//record_file := fopen(output_file_name, "wb");
	//mem_free(output_file_name);
	//if (record_file == NULL) die(MSG_003);

	//create_class_file_header(record_file);

	cp_add_class(class_name);
	cp_add_class('java/lang/Object');
	cp_add_utf8('<init>');
	cp_add_utf8('()V');
	cp_add_utf8('Code');
	cp_add_utf8('Copy');
	copy_sig:='(LR_'+IntToStr(records^.unique_record_ID)+';)LR_'+IntToStr(records^.unique_record_ID)+';';
	cp_add_utf8(copy_sig);

	create_record_fields(records);

	init_method := bytecode_create();
	create_init_method(records, init_method);

	copy_method := bytecode_create();
	create_copy_method(records, copy_method);

	//write_constant_pool(record_file);
	//write_short_int(record_file, 0x0021);
	//write_short_int(record_file, cp_add_class(class_name));
	//write_short_int(record_file, cp_add_class("java/lang/Object"));
	//write_short_int(record_file, 0);
	//write_short_int(record_file, (short) type_list_length(record->elements_type_list));
	//write_record_fields(record_file, record);
	//write_short_int(record_file, 2);
	//write_init_method(record_file, init_method);
	//write_copy_method(record, record_file, copy_method);
	//write_short_int(record_file, 0);

	//fclose(record_file);

	constants := @global_cp;
	constant_pool_size := global_cp_count;

end;

end.
