unit PasCpool;

interface

uses
  Windows, SysUtils, PasTypes;

type
  constant_pool=^constant_pool_struct;
  constant_pool_struct = record
    tag: byte;
    param1: dword;
    param2: dword;
    data: Pchar;
    data_len: integer;
  end;

var
  constants: array of constant_pool;
  constant_pool_size: integer = 0;

  function cp_add_fieldref(class_name, name, types: string): integer;
  function cp_add_class(class_name: string): integer;
  function cp_add_methodref(class_name, name, types: string): integer;
  function cp_add_utf8(str: string): integer;
  function cp_add_integer(data: integer): integer;
  function cp_add_string(str: string): integer;
  
implementation

function cp_create_new_entry(): constant_pool;
begin
	Inc(constant_pool_size);
  SetLength(constants,constant_pool_size);
  New(constants[constant_pool_size-1]);
	result := constants[constant_pool_size-1];
end;

function cp_add_utf8(str: string): integer;
var
	i: integer;
	size: integer;
	entry: constant_pool;
begin
  size := 0;

	for i:=0 to constant_pool_size-1 do begin
		if (constants[i].tag = 1) and SameText(constants[i].data, str) then begin Result:= i+1; Exit; end;
  end;
	entry := cp_create_new_entry();
	entry^.tag := 1;
	entry^.data_len := size;
	entry^.data := PChar(str+#0);
	result:= constant_pool_size;
end;

function cp_add_class(class_name: string): integer;
var
	i     : integer;
	entry : constant_pool;
  utf8_const: integer;
begin

	utf8_const := cp_add_utf8(class_name);
	for i:=0 to constant_pool_size-1 do
		if ((constants[i].tag = 7) and (constants[i].param1 = utf8_const)) then begin
			Result:= i+1; Exit;
		end;

	entry := cp_create_new_entry();
	entry^.tag := 7;
	entry^.param1 := utf8_const;
	result:= constant_pool_size;

end;

function cp_add_nameandtype(name,types: string): integer;
var
  name_index,
	type_index,
	i     : integer;
	entry : constant_pool;
begin

	name_index := cp_add_utf8(name);
	type_index := cp_add_utf8(types);

	for i:=0 to constant_pool_size-1 do begin
		if ((constants[i].tag = 12)
			and (constants[i].param1 = name_index)
			and (constants[i].param2 = type_index)) then begin
			result:= i+1;
      exit;
		end;
	end;

	entry := cp_create_new_entry();
	entry^.tag := 12;
	entry^.param1 := name_index;
	entry^.param2 := type_index;

	result:= constant_pool_size;
end;

function cp_add_fieldref(class_name, name, types: string): integer;
var
	i,class_index,
	nameandtype_index: integer;
	entry: constant_pool;
begin
	class_index := cp_add_class(class_name);
	nameandtype_index := cp_add_nameandtype(name, types);

	for i:=0 to constant_pool_size-1 do begin
		if ((constants[i].tag = 9)
			and (constants[i].param1 = class_index)
			and (constants[i].param2 = nameandtype_index)) then begin
			  result:= i+1;
        Exit;
    end;
  end;

	entry := cp_create_new_entry();
	entry^.tag := 9;
	entry^.param1 := class_index;
	entry^.param2 := nameandtype_index;

	result:= constant_pool_size;
end;

function cp_add_methodref(class_name, name, types: string): integer;
var
	i,class_index,
	nameandtype_index: integer;
	entry: constant_pool;
begin

	class_index := cp_add_class(class_name);
	nameandtype_index := cp_add_nameandtype(name, types);

	for i:=0 to constant_pool_size-1 do begin
		if ((constants[i].tag = 10)
			and (constants[i].param1 = class_index)
			and (constants[i].param2 = nameandtype_index)) then begin
			  result:= i+1;
        Exit;
      end;
  end;

	entry := cp_create_new_entry();
	entry^.tag := 10;
	entry^.param1 := class_index;
	entry^.param2 := nameandtype_index;

	result:= constant_pool_size;
end;

function cp_add_integer(data: integer): integer;
var
	i: integer;
	entry: constant_pool;
  p_value: integer;
begin

	for i:=0 to constant_pool_size-1 do begin

		if (constants[i].tag = 3) then begin
			p_value := integer(constants[i].data);
			if p_value = data then begin Result:= i+1; exit; end;
    end;

  end;

	entry := cp_create_new_entry();
	entry^.tag := 3;
	New(entry^.data);
	entry^.data := PChar(data);

	Result:= constant_pool_size;

end;

function cp_add_string(str: string): integer;
var
	i: integer;
	entry: constant_pool;
  utf8_const: integer;
begin

	utf8_const := cp_add_utf8(str);

	for i:=0 to constant_pool_size-1 do begin
		if ((constants[i].tag = 8)
			and (constants[i].param1 = utf8_const)) then begin
			Result:= i+1;
      Exit;
    end;
  end;

	entry := cp_create_new_entry();
	entry^.tag := 8;
	entry^.param1 := utf8_const;

	Result:= constant_pool_size;

end;

end.
