unit PasParser;

interface

uses
  PasScanner, PasTypes, PasErrors, SysUtils, PasCodegen, PasCpool;

  procedure Compile(P: string);
  function  RD_type(current_block: block): types;
  procedure RD_block(current_block: block);
  procedure RD_block_body(current_block: block);
  procedure RD_statement(current_block: block);
  function  RD_expression(current_block: block): types;

implementation

var
  StackPos    :integer;
  IdentPos    :integer;

  procedure RD_unit_interface(current_block: block);
begin
	if (Token <> _INTERFACE) then
		if Token in [_IMPLEMENTATION,_INITIALIZATION] then
      MakeError(MSG_203, 'interface');

	ReadToken;
	if (Token = _SEMICOLON) then ReadToken;

	inside_interface_part := 1;
	RD_block(current_block);
	inside_interface_part := 0;
end;

procedure RD_uses_list();
label
  following_unit;
begin
	while (Token = _USES) do begin

    repeat
		  ReadToken;
		  if Token <> _IDENT then MakeError(MSG_202, '');
		  //load_extern_library(string_from_cstr(TokenStr));
		  ReadToken;
		until Token = _COMMA;

		if (Token = _SEMICOLON) then
			MakeError(MSG_200, ';')
		else
			ReadToken;

  end;

	if (Token <> _SEMICOLON) then ReadToken;

	//if (detect_units_only)
	{
		mem_close();
		fclose(yyin);
		compile_terminate();
	}
end;

procedure RD_unit_implementation(current_block: block);
begin

	if (Token <> _IMPLEMENTATION) then
		MakeError(MSG_203, 'implementation');

	ReadToken;
	if (Token = _SEMICOLON) then ReadToken;

	RD_uses_list();

	inside_implementation_part := 1;
	RD_block(current_block);
	inside_implementation_part := 0;
end;

procedure RD_unit_initialization(current_block: block);
begin
	if Token = _INITIALIZATION then begin
		ReadToken;
		if Token = _SEMICOLON then ReadToken;
		RD_block_body(current_block);
	end;
end;

procedure RD_unit_finalization(current_block: block);
begin
	if Token = _FINALIZATION then
		MakeError(MSG_220, '');
end;

function RD_program_header(): string;
var
  program_name: string;
begin
	program_name := '';

	if (Token = _PROGRAM) OR (Token = _UNIT) then begin

		if Token = _UNIT then compiling_unit := 1;
		ReadToken;

		if Token <> _IDENT then
			MakeError(MSG_202, '')
		else
			program_name := TokenStr;

		if (compiling_unit<>0) and (length(program_name) < 3) then
			MakeError(MSG_454,'');

		ReadToken;
		Match(_SEMICOLON);

  end;

	result := program_name;
end;

//----------------------------------------------------------

function RD_basic_type(current_block: block): types;
var
	return_type: types;
  name: identifier;
	declared_type: types;
  type_identifier: identifier;
  constant_identifier: identifier;
  constant_type: types;
  S: integer;
begin

  return_type := type_create();
	return_type^.type_class := error_type;

  case Token of
    _PLUS:  begin ReadToken; S:=1; end;
    _MINUS: begin ReadToken; S:=-1;end
  else
    S:=0;
  end;

	case Token of

	_IDENT:
        begin
          declared_type := type_from_name(current_block, TokenStr);
          name := get_identifier(current_block, LowerCase(TokenStr));

          if name^.identifier_class = unit_name then begin

            ReadToken;
            Match(_POINT);

            type_identifier := name_table_find(name^.unit_block^.names, TokenStr);

            if type_identifier = NIL then
              MakeError(MSG_455, TokenStr)
            else
              declared_type := type_duplicate(type_identifier^.defined_type);
          end;

          if declared_type^.type_class <> error_type then begin
            return_type := type_duplicate(declared_type);
            ReadToken;
          end	else begin
            constant_identifier := get_constant_identifier(current_block, TokenStr);
            if ((constant_identifier^.identifier_class = constant_name)
            and ((constant_identifier^.constant_type^.type_class = integer_type)
            or (constant_identifier^.constant_type^.type_class = char_type))) then begin

              return_type^.type_class := interval_type;
              return_type^.interval_base_type := constant_identifier^.constant_type^.type_class;
              if (constant_identifier^.constant_type^.type_class = integer_type) then
                return_type^.first_element := constant_identifier^.constant_int_value
              else
                return_type^.first_element := constant_identifier^.constant_int_value;
              return_type^.last_element := return_type^.first_element + 1;

              ReadToken;
              Match(_PERIOD);

              case Token of

                _IDENT:
                        begin
                          constant_identifier := get_constant_identifier(current_block, TokenStr);
                          if ((constant_identifier^.identifier_class <> constant_name)
                            or (constant_identifier^.constant_type^.type_class <> return_type^.interval_base_type)) then
                            MakeError(MSG_416, '')
                          else begin
                            if (return_type^.interval_base_type = integer_type) then
                              return_type^.last_element := constant_identifier^.constant_int_value
                            else
                              return_type^.last_element := constant_identifier^.constant_int_value;
                          end;
                        end;
                _INTEGER:
                        begin
                          if (return_type^.interval_base_type <> integer_type) then
                            MakeError(MSG_416, '');
                          return_type^.last_element := TokenInt;
                        end;
                _CHAR:
                        begin
                          if (return_type^.interval_base_type <> char_type) then
                            MakeError(MSG_416, '');
                          return_type^.last_element := ord(TokenStr[1]);
                        end;

              else
                MakeError(MSG_416, '');
              end;

              ReadToken;

            end	else
              MakeError(MSG_415, TokenStr);
          end;
        end;

	_INTEGER,
	_CHAR:
        begin
          return_type^.type_class := interval_type;

          if Token = _INTEGER then begin
            return_type^.interval_base_type := integer_type;
            return_type^.first_element := TokenInt*S;
          end	else begin
            return_type^.interval_base_type := char_type;
            return_type^.first_element := ord(TokenStr[1]);
          end;

          return_type^.last_element := return_type^.first_element + 1;
          ReadToken;
          Match(_PERIOD);

            case Token of

              _IDENT:
                          begin
                            constant_type := get_constant_type(current_block, TokenStr);
                            if (constant_type^.type_class <> return_type^.interval_base_type) then
                              MakeError(MSG_413, '');
                            constant_identifier := get_constant_identifier(current_block, TokenStr);
                            if constant_identifier^.identifier_class <> constant_name then
                              MakeError(MSG_414, TokenStr)
                            else begin
                              if (return_type^.interval_base_type = integer_type) then
                                return_type^.last_element := constant_identifier^.constant_int_value
                              else
                                return_type^.last_element := constant_identifier^.constant_int_value;
                            end;
                          end;
              _INTEGER  : return_type^.last_element := TokenInt;
              _CHAR     : return_type^.last_element := ord(TokenStr[1]);
              
            else
              MakeError(MSG_217, '');
            end;

          ReadToken;
        end;

	_RBRACKET: MakeError(MSG_212,'');

	else
    MakeError(MSG_204, TokenStr);
	end;

	result:= return_type;
end;

function RD_identifier_list(current_block: block; check_valid_names: integer): string_list;
var
	return_list: string_list;
begin
	return_list := string_list_create();
	if Token <> _IDENT then MakeError(MSG_202, '');

	if ((check_valid_names<>0)
  and (block_check_name(current_block, TokenStr)=0)) then
		MakeError(MSG_400, TokenStr);

	string_list_append(return_list, TokenStr);
	ReadToken;

	while Token = _COMMA do begin
		ReadToken;
		if Token <> _IDENT then MakeError(MSG_202, '');
		if ((check_valid_names<>0)
    and ( block_check_name(current_block, TokenStr)=0 )) then
			MakeError(MSG_400, TokenStr);
		string_list_append(return_list, TokenStr);
		ReadToken;
	end;

	result:= return_list;
end;

function RD_array_declaration(current_block: block): types;
var
	new_type : types;
	dimension: types;
begin
	new_type := type_create();
	new_type^.type_class := array_type;
	new_type^.dimensions_list := type_list_create();
	if Token <> _LBRACKET then MakeError(MSG_200,'[');

	repeat
		ReadToken;
		dimension := RD_basic_type(current_block);
		if (dimension^.type_class <> interval_type) then MakeError(MSG_430, '');
		type_list_append(new_type^.dimensions_list, dimension);
		if not (Token in [_RBRACKET,_COMMA]) then MakeError(MSG_200, TokenStr);
	until (Token = _RBRACKET);

	ReadToken;
	Match(_OF);

	new_type^.element_type := RD_type(current_block);
	result:= new_type;
end;

function RD_record_declaration(current_block: block): types;
var
	new_type: types;
	identifier_list: string_list;
	element_type: types;
  first_pass: shortint;
begin
	first_pass := 1;

	new_type := type_create();
	new_type^.type_class := record_type;
	new_type^.elements_name_list := string_list_create();
	new_type^.elements_type_list := type_list_create();
	new_type^.unique_record_ID := next_record_ID;

	Inc(next_record_ID);
	
	repeat
		if first_pass<>0 then
			first_pass := 0
		else
			ReadToken;

		if Token = _END then break;
		identifier_list := RD_identifier_list(current_block, 0);

		Match(_COLON);

		element_type := RD_type(current_block);

		if (element_type^.type_class = interval_type) then
			MakeError(MSG_440, '');

		type_add_record(new_type, identifier_list, element_type);

	until Token <> _SEMICOLON;

  Match(_END);

	create_record_class(new_type);
	result:= new_type;
end;

function RD_file_declaration(current_block: block):types;
var
	file_type: types;
begin
	file_type := type_create();
	file_type^.type_class := error_type;
	if Token <> _OF then
		MakeError(MSG_203,'of'+TokenStr);
	ReadToken;
  //!!! TODO FILE
	ReadToken;
	MakeError(MSG_435, '');
	result:= file_type;
end;

procedure RD_set_declaration(current_block: block);
begin
	ReadToken;
	RD_basic_type(current_block);
end;

function RD_type(current_block: block): types;
var
  error: types;
begin

	if Token=_PACKED then begin
    MakeWarning(WMSG_003, '');
    ReadToken;
	end;

	case Token of

	_ARRAY:
        begin
          ReadToken;
          result:= RD_array_declaration(current_block);
        end;

	_RECORD:
        begin
          ReadToken;
          result:= RD_record_declaration(current_block);
        end;

	_FILE:
        begin
          ReadToken;
          result:= RD_file_declaration(current_block);
        end;

	_SET:
        begin
          error := type_create();
          error^.type_class := error_type;
          MakeError(MSG_211, '');
          ReadToken;
          RD_set_declaration(current_block);
          result:= error;
        end;

	else
		result:= RD_basic_type(current_block);
	end;

end;

procedure VarDeclaration(current_block: block);
var
	identifier_list: string_list;
	var_type: types;
begin
		identifier_list := RD_identifier_list(current_block, 1);
		Match(_COLON);
		var_type := RD_type(current_block);
		if var_type^.type_class = interval_type then MakeError(MSG_440, '');
		add_variables(current_block, identifier_list, var_type);
		//if (inside_interface_part<>0) then
			//bsf_write_variables(identifier_list, var_type);
		Match(_SEMICOLON);
end;

procedure Constant(var Value: Values; var Typs: TToken; current_block: block);
var
  S:integer;
  type_identifier: identifier;
begin

  case Token of
    _PLUS:  begin ReadToken; S:=1; end;
    _MINUS: begin ReadToken; S:=-1;end
  else
    S:=0;
  end;

  if S<>0 then begin
    TokenInt:=TokenInt*S;
    TokenReal:=TokenReal*S;
  end;

  Typs:=Token;

  case Token of
    _IDENT  :
            begin
              type_identifier := name_table_find(current_block^.names,TokenStr);
              if type_identifier^.identifier_class <> constant_name then MakeError(MSG_416, '');
              case type_identifier^.constant_type^.type_class of
                integer_type: begin Typs:=_INTEGER; Value.Int_Val     :=type_identifier^.constant_int_value; end;
                real_type   : begin Typs:=_REAL;    Value.Real_Val    :=type_identifier^.constant_real_value; end;
                char_type   : begin Typs:=_CHAR;    Value.Char_Val    :=type_identifier^.constant_string_value[1]; end;
                boolean_type: begin Typs:=_BOOLEAN; Value.Bool_Val    :=type_identifier^.constant_int_value; end;
                string_type : begin Typs:=_STRING;  Value.String_Val  :=type_identifier^.constant_string_value; end;
              end;
            end;
    _INTEGER: Value.Int_Val   :=TokenInt;
    _REAL:    Value.Real_Val  :=TokenReal;
    _BOOLEAN: Value.Bool_Val  :=TokenBool;
    _CHAR:    Value.Char_Val  :=TokenStr[1];
    _STRING:  Value.String_Val:=TokenStr;
  else
    MakeError(MSG_206, '');
  end;

  ReadToken;
end;

procedure ConstDeclaration(current_block: block);
var
  Name  :string;
  Value : Values;
  Typs  : TToken;
begin
  if block_check_name(current_block, TokenStr)=0 then MakeError(MSG_400, TokenStr);
  Name:=TokenStr;
  ReadToken;
  Match(_EQUAL);
  Constant(Value,Typs,current_block);
  Match(_SEMICOLON);

  case Typs of
    _INTEGER:
          begin
            add_integer_constant(current_block,Value.Int_Val, name);
            //if inside_interface_part<>0 then
              //bsf_write_integer_constant(Value.Int_Val, name);
          end;

    _REAL:
          begin
            add_real_constant(current_block,Value.Real_Val,name);
            //if inside_interface_part<>0 then
              //bsf_write_real_constant(Value.Real_Val,name);
          end;

    _BOOLEAN:
          begin
            add_boolean_constant(current_block,Value.Bool_Val,name);
            //if inside_interface_part<>0 then
              //bsf_write_boolean_constant(Value.Bool_Val, name);
          end;

    _CHAR:
          begin
            add_char_constant(current_block,Value.Char_Val,name);
            //if inside_interface_part<>0 then
              //bsf_write_char_constant(Value.Char_Val,name);
          end;

    _STRING:
          begin
            add_string_constant(current_block,Value.String_Val,name);
            //if inside_interface_part<>0 then
              //bsf_write_string_constant(Value.String_Val,name);
          end;
    end;

end;

procedure TypeDeclaration(current_block: block);
var
	type_name: string;
	type_type: types;
begin
  type_name := TokenStr;
  if block_check_name(current_block, TokenStr)=0 then MakeError(MSG_400, TokenStr);
  ReadToken;
  Match(_EQUAL);
  type_type := RD_type(current_block);
  add_type(current_block, type_name, type_type);
  //if inside_interface_part<>0 then
    //bsf_write_type(type_name, type_type);
  Match(_SEMICOLON);
end;

function RD_param_list(current_block: block): type_list;
var
	parameter_list: type_list;
	identifier_list: string_list;
	parameters_type: types;
	is_parameter_variable,
	i, len: integer;
  units: identifier;
  type_identifier: identifier;
begin
	parameter_list := type_list_create();

	if Token = _LPAREN then begin

		repeat

			ReadToken;
			if Token = _RPAREN then break;
			is_parameter_variable := 0;

			if Token = _VAR then begin
				is_parameter_variable := 1;
				ReadToken;
				MakeWarning(WMSG_004,'');
			end;

			identifier_list := RD_identifier_list(current_block, 0);
      Match(_COLON);
			parameters_type := type_from_name(current_block, TokenStr);

			if parameters_type^.type_class = error_type then begin
				TokenStr:=lowercase(TokenStr);
				units := get_identifier(current_block, TokenStr);
				if units^.identifier_class = unit_name then begin
					ReadToken;
          Match(_POINT);
					type_identifier := name_table_find(units^.unit_block^.names, TokenStr);
					if type_identifier = NIL then
						MakeError(MSG_455, TokenStr)
					else
						parameters_type := type_identifier^.defined_type;
				end else
          MakeError(MSG_455, TokenStr);
			end;

			if parameters_type^.type_class = interval_type then
			  MakeError(MSG_440, '');

			len := string_list_length(identifier_list);
			for i:=0 to len do
				type_list_append(parameter_list, type_duplicate(parameters_type));
			add_parameters(current_block, identifier_list, parameters_type,is_parameter_variable);

      Match(_IDENT);
		
		until Token <> _SEMICOLON;

		Match(_RPAREN);
    
	end;

	result:= parameter_list;

end;

function RD_expression_list_array(current_block: block; code: bytecode; array_type: types): type_list;
var
	return_list: type_list;
	expression_type: types;
	oldBytecode: bytecode;
	it: type_list;
begin
	it := array_type^.dimensions_list;

	oldBytecode := current_block^.code;
	current_block^.code := code;
	return_list := type_list_create();
	expression_type := RD_expression(current_block);
	adjust_indices(code, it^.data);
	if it <> NIL then it := it^.next;
	bytecode_append(code, aaload_);
	type_list_append(return_list, expression_type);

	while Token = _COMMA do begin
		ReadToken;
		expression_type := RD_expression(current_block);
		adjust_indices(code, it^.data);
		if (it <> NIL) then it := it^.next;
		bytecode_append(code, aaload_);
		type_list_append(return_list, expression_type);
  end;

	current_block^.code := oldBytecode;

	Result:= return_list;
end;

function RD_expression_list_cast(current_block: block; formal_params: type_list): type_list;
var
	return_list: type_list;
	expression_type: types;
	it: type_list;
  cast,method_index: integer;
begin
	it := formal_params;
	return_list := type_list_create();
	expression_type := RD_expression(current_block);

	if ((it <> NIL) and (it^.data <> NIL)) then begin

		cast := type_equal_cast(it^.data, expression_type);

		if ((cast = 2) and (it^.data^.type_class = string_type)) then begin
			method_index := cp_add_methodref('java/lang/String', 'valueOf', '(C)Ljava/lang/String;');
			bytecode_append(current_block^.code, invokestatic_);
			bytecode_append_short_int(current_block^.code, method_index);
		end;

		if ((cast = 2) and (it^.data^.type_class = real_type)) then begin
			usesFloat := 1;

			if (mathType = 1) then begin
				bytecode_append(current_block^.code, invokestatic_);
				bytecode_append_short_int(current_block^.code, cp_add_methodref('F', 'fI', '(I)I'));
			end	else begin
				bytecode_append(current_block^.code, new_);
				bytecode_append_short_int(current_block^.code, cp_add_class('Real'));
				bytecode_append(current_block^.code, dup_x1_);
				bytecode_append(current_block^.code, swap_);
				bytecode_append(current_block^.code, invokespecial_);
				bytecode_append_short_int(current_block^.code, cp_add_methodref('Real', '<init>', '(I)V'));
      end;

    end;

		it := it^.next;

  end;

	type_list_append(return_list, expression_type);

	while Token = _COMMA do begin

		ReadToken;

		expression_type := RD_expression(current_block);

		if ((it <> NIL) and (it^.data <> NIL)) then begin

			cast := type_equal_cast(it^.data, expression_type);

			if ((cast = 2) and (it^.data^.type_class = string_type)) then begin
				method_index := cp_add_methodref('java/lang/String', 'valueOf', '(C)Ljava/lang/String;');
				bytecode_append(current_block^.code, invokestatic_);
				bytecode_append_short_int(current_block^.code, method_index);
			end;

			if ((cast = 2) and (it^.data^.type_class = real_type)) then begin

				usesFloat := 1;

				if (mathType = 1) then begin
					bytecode_append(current_block^.code, invokestatic_);
					bytecode_append_short_int(current_block^.code, cp_add_methodref('F', 'fI', '(I)I'));
				end	else begin
					bytecode_append(current_block^.code, new_);
					bytecode_append_short_int(current_block^.code, cp_add_class('Real'));
					bytecode_append(current_block^.code, dup_x1_);
					bytecode_append(current_block^.code, swap_);
					bytecode_append(current_block^.code, invokespecial_);
					bytecode_append_short_int(current_block^.code, cp_add_methodref('Real', '<init>', '(I)V'));
        end;

      end;

			it := it^.next;

    end;

		type_list_append(return_list, expression_type);

  end;

	Result:= return_list;

end;

function RD_value(current_block: block): types;
var
	return_type: types;
  cp_index, i: integer;
  cst: real;
	csti,cstf: integer;
  number: string[64];
  name: identifier;
	identifier_text: string;
  units_name: identifier;
  current_type,new_type: types;
	is_field: integer;
  record_name: string[5];
	field_descriptor: array [0..128] of char;
	fieldref_index: integer;
  array_elements: type_list;
  num1,num2: string[6];

  methodref_index: integer;
  method_type: array [0..128] of char;
  method_index: integer;

  br_depth: integer;
  full_unit_name: string;

	compare_result: integer;
	parameter_list: type_list;
  par_num: string[4];

	pos: integer;
	it: type_list;

label
  constant_value,
  function_call,
  variable_value;
begin
	return_type := type_create();

	return_type^.type_class := error_type;

	if (Token = _INTEGER) then begin

		case TokenInt of
		  -1: bytecode_append(current_block^.code, iconst_m1_);
		  0:  bytecode_append(current_block^.code, iconst_0_);
		  1:  bytecode_append(current_block^.code, iconst_1_);
		  2:  bytecode_append(current_block^.code, iconst_2_);
		  3:  bytecode_append(current_block^.code, iconst_3_);
		  4:  bytecode_append(current_block^.code, iconst_4_);
		  5:  bytecode_append(current_block^.code, iconst_5_);
		else
				if (abs(TokenInt) < 127) then begin
					bytecode_append(current_block^.code, bipush_);
					bytecode_append(current_block^.code, Jmnem(TokenInt));
				end	else if (abs(TokenInt) < 16000) then begin
					bytecode_append(current_block^.code, sipush_);
					bytecode_append_short_int(current_block^.code, shortint(TokenInt));
				end	else begin
					cp_index := cp_add_integer(TokenInt);
					if (cp_index <= 255) then begin
						bytecode_append(current_block^.code, ldc_);
						bytecode_append(current_block^.code, Jmnem(cp_index));
					end	else begin
						bytecode_append(current_block^.code, ldc_w_);
						bytecode_append_short_int(current_block^.code, shortint(cp_index));
          end;
        end;
    end;

		ReadToken;
		return_type^.type_class := integer_type;
		Result:= return_type;
    exit;
  end;

	if (Token = _REAL) then begin

		cst := TokenReal;
		csti := round(cst);
		cstf := round((cst - round(cst))*100000);
		usesFloat := 1;

		if mathType = 1 then begin

			if (abs(csti) < 127) then begin
				bytecode_append(current_block^.code, bipush_);
				bytecode_append(current_block^.code, Jmnem(csti));
			end	else if (abs(csti) < 15000) then begin
				bytecode_append(current_block^.code, sipush_);
				bytecode_append_short_int(current_block^.code, csti);
			end	else begin
				bytecode_append(current_block^.code, ldc_w_);
				bytecode_append_short_int(current_block^.code, cp_add_integer(csti));
      end;

			if (abs(cstf) < 127) then begin
				bytecode_append(current_block^.code, bipush_);
				bytecode_append(current_block^.code, Jmnem(cstf));
			end	else if (abs(cstf) < 15000) then begin
				bytecode_append(current_block^.code, sipush_);
				bytecode_append_short_int(current_block^.code, cstf);
			end	else begin
				bytecode_append(current_block^.code, ldc_w_);
				bytecode_append_short_int(current_block^.code, cp_add_integer(cstf));
      end;

			bytecode_append(current_block^.code, invokestatic_);
			bytecode_append_short_int(current_block^.code, cp_add_methodref('F', 'fI', '(II)I'));
		end	else begin

			bytecode_append(current_block^.code, new_);
			bytecode_append_short_int(current_block^.code, cp_add_class('Real'));
			bytecode_append(current_block^.code, dup_);

			Val(number,TokenReal,i);
			bytecode_append(current_block^.code, ldc_w_);
			bytecode_append_short_int(current_block^.code, cp_add_string(number));

			bytecode_append(current_block^.code, invokespecial_);
			bytecode_append_short_int(current_block^.code, cp_add_methodref('Real', '<init>', '(Ljava/lang/String;)V'));
    end;

		ReadToken;
		return_type^.type_class := real_type;
		Result:= return_type;
    Exit;
    
  end;

	if (Token = _BOOLEAN) then begin
		if TokenBool<>0 then
			bytecode_append(current_block^.code, iconst_m1_)
		else
			bytecode_append(current_block^.code, iconst_0_);

		ReadToken;
		return_type^.type_class := boolean_type;
		Result:= return_type;
  end;

	if Token = _CHAR then begin
		bytecode_append(current_block^.code, bipush_);
		bytecode_append(current_block^.code, Jmnem(TokenStr[1]));

		ReadToken;
		return_type^.type_class := char_type;
		Result:= return_type;
    Exit;
  end;

	if Token =_STRING then begin
		cp_index := cp_add_string(TokenStr);

		if (cp_index <= 255) then begin
			bytecode_append(current_block^.code, ldc_);
			bytecode_append(current_block^.code, Jmnem(cp_index));
		end	else begin
			bytecode_append(current_block^.code, ldc_w_);
			bytecode_append_short_int(current_block^.code, shortint(cp_index));
    end;

		ReadToken;
		return_type^.type_class := string_type;
		Result:= return_type;
    Exit;
  end;

	if Token = _IDENT then begin

		identifier_text:=lowercase(TokenStr);

		name := get_identifier(current_block, identifier_text);

		if ((name^.identifier_class = none)
			or (name^.identifier_class = program_name)
			or (name^.identifier_class = type_name)
			or (name^.identifier_class = procedure_name)) then
			MakeError(MSG_429, TokenStr);

		ReadToken;

		if (name^.identifier_class = constant_name) then begin
constant_value:
			create_constant_bytecode(name, current_block^.code);
			return_type := type_duplicate(name^.constant_type);
			Result:= return_type;
      Exit;
    end;

		if name^.identifier_class = unit_name then begin

			Match(_POINT);

			ReadToken;
			units_name := name_table_find(name^.unit_block^.names, TokenStr);
			identifier_text:=TokenStr;
			if units_name = NIL then MakeError(MSG_453, TokenStr);
			name := units_name;
			ReadToken;

			if (name^.identifier_class = function_name) then goto function_call;
			if (name^.identifier_class = constant_name) then goto constant_value;
			if (name^.identifier_class = variable_name) then goto variable_value;

			MakeError(MSG_458, TokenStr);

    end;

		if ((name^.identifier_class = variable_name)
    or (name^.identifier_class = parameter_name)) then begin

variable_value:
			is_field := name^.belongs_to_program_block;

			if (name^.identifier_class = variable_name) then begin
				create_variable_bytecode(name, current_block^.code, identifier_text, is_field);
				current_type := type_duplicate(name^.variable_type);
      end;

			if(name^.identifier_class = parameter_name) then begin
				create_variable_bytecode(name, current_block^.code, identifier_text, is_field);
				current_type := type_duplicate(name^.parameter_type);
      end;

			while ((Token = _POINT) or (Token = _LPAREN)) do begin

				if (Token = _POINT) then begin
					if (current_type^.type_class <> record_type) then MakeError(MSG_424, '');
					ReadToken;
					new_type := type_find_record_element(current_type, TokenStr);
					record_name:= 'R_'+IntToStr(current_type^.unique_record_ID);
					current_type := new_type;
					if (current_type = NIL) then MakeError(MSG_425, TokenStr);
					TokenStr:=lowercase(TokenStr);
					get_field_descriptor(new_type, field_descriptor);
					fieldref_index := cp_add_fieldref(record_name, TokenStr, field_descriptor);
					bytecode_append(current_block^.code, getfield_);
					bytecode_append_short_int(current_block^.code, fieldref_index);
					ReadToken;
				end	else if Token = _LPAREN then begin
					if current_type^.type_class <> array_type then MakeError(MSG_426, '');
					ReadToken;
					array_elements := RD_expression_list_array(current_block, current_block^.code, current_type);
					replace_aaload_instruction(current_block^.code, current_type^.element_type);
					if (type_list_different_parameter_array(array_elements, current_type^.dimensions_list) <> 0) then begin
						num1:=IntToStr(type_list_length(current_type^.dimensions_list));
						num2:=IntToStr(type_list_length(array_elements));
						MakeError(MSG_427, num1+' '+num2);
          end;
					Match(_RPAREN);
					new_type := type_duplicate(current_type^.element_type);
					current_type := new_type;
        end;

      end;

			return_type := current_type;

			Result:= return_type;
      Exit;

  end;

function_call:

		if (name^.identifier_class = function_name) then begin

			if (name^.standard_function<>0) then
				;//create_std_function_prefix(current_block^.code, identifier_text);

			if (type_list_length(name^.parameters) = 0) then begin

				if (Token = _LBRACKET) then MakeError(MSG_419, '');
				if (name^.unit_function<>0) then begin

					if (name^.container_unit^.is_library = 1) then
						full_unit_name := 'Lib_'
					else
						full_unit_name := '';
					full_unit_name:=full_unit_name+name^.container_unit^.unit_name;

					identifier_text:=lowercase(identifier_text);
					method_type:='()';
					get_field_descriptor(name^.return_type, method_type+2);
					method_index := cp_add_methodref(full_unit_name,identifier_text, method_type);
					bytecode_append(current_block^.code, invokestatic_);
					bytecode_append_short_int(current_block^.code, method_index);

				end	else if name^.standard_function=0 then begin
					method_type:='()';
					get_field_descriptor(name^.return_type, method_type + 2);
					if (compiling_unit = 0) then
						methodref_index := cp_add_methodref('M', identifier_text, method_type)
					else
						methodref_index := cp_add_methodref(str_program_name, identifier_text, method_type);
					bytecode_append(current_block^.code, invokestatic_);
					bytecode_append_short_int(current_block^.code, methodref_index);
				end	else
					;//create_std_function_code(current_block^.code, identifier_text);

				return_type := type_duplicate(name^.return_type);
				Result:= return_type;
        Exit;
			end	else begin

				Match(_LBRACKET);

				parameter_list := RD_expression_list_cast(current_block, name^.parameters);
				compare_result := type_list_different_parameter_cast(parameter_list, name^.parameters);
				if (compare_result = -1) then MakeError(MSG_421, '');

				if (compare_result > 0) then begin
					par_num:=IntToStr(compare_result);
					MakeError(MSG_422, par_num);
        end;

				Match(_RBRACKET);

				if name^.standard_function=0 then begin

						pos := 1;
						it := name^.parameters;
						method_type[0] := '(';
						while (it <> NIL) or (it^.data = NIL) do begin
							get_field_descriptor(it^.data, method_type + pos);
							pos := strlen(method_type);
							it := it^.next;
						end;
						method_type[strlen(method_type)+1]:= ')';


					get_field_descriptor(name^.return_type, method_type + strlen(method_type));

					if (name^.unit_function<>0) then begin

						if (name^.container_unit^.is_library = 1) then
							full_unit_name := 'Lib_'
						else
							full_unit_name := '';
						full_unit_name:=full_unit_name+name^.container_unit^.unit_name;
						identifier_text:=lowercase(identifier_text);
						methodref_index := cp_add_methodref(full_unit_name, identifier_text, method_type);
					end	else begin
						if (compiling_unit = 0) then
							methodref_index := cp_add_methodref('M', identifier_text, method_type)
						else
							methodref_index := cp_add_methodref(str_program_name, identifier_text, method_type);
					end;

					bytecode_append(current_block^.code, invokestatic_);
					bytecode_append_short_int(current_block^.code, methodref_index);
				end else
					;//create_std_function_code(current_block^.code, identifier_text);

				return_type := type_duplicate(name^.return_type);
				Result:= return_type;
        Exit;
      end;

    end;
		Result:= return_type;
    Exit;
	end;

	if (Token = _LPAREN) then begin
		ReadToken;
		return_type :=  RD_expression(current_block);
		Match(_RPAREN);
		Result:= return_type;
    Exit;
	end;

	MakeError(MSG_204, TokenStr);

end;

function RD_neg(current_block: block): types;
var
	type1: types;
	is_neg: integer;
begin

  is_neg := 0;

	if Token = _MINUS then begin
	  ReadToken;
		is_neg := 1;
	end;

	type1 := RD_value(current_block);

	if (is_neg<>0) then begin
		if ((type1^.type_class = integer_type)
			or ((type1^.type_class = real_type) and (mathType = 1))) then
			bytecode_append(current_block^.code, ineg_);

		if ((type1^.type_class = real_type) and (mathType <> 1)) then begin
			usesFloat := 1;

			bytecode_append(current_block^.code, new_);
			bytecode_append_short_int(current_block^.code, cp_add_class('Real'));
			bytecode_append(current_block^.code, dup_x1_);
			bytecode_append(current_block^.code, swap_);
			bytecode_append(current_block^.code, invokespecial_);
			bytecode_append_short_int(current_block^.code, cp_add_methodref('Real', '<init>', '(LReal;)V'));
			bytecode_append(current_block^.code, dup_);
			bytecode_append(current_block^.code, invokevirtual_);
			bytecode_append_short_int(current_block^.code, cp_add_methodref('Real', 'neg', '()V'));
    end;
  end;

	if ((type1^.type_class <> real_type)
		and (type1^.type_class <> integer_type)
		and (type1^.type_class <> error_type)
		and (is_neg<>0)) then MakeError(MSG_417, '');

	Result:= type1;
  
end;

function RD_not(current_block: block): types;
var
	type1: types;
	is_not: integer;
begin

  is_not := 0;
	if Token = _NOT then begin
	  ReadToken;
		is_not := 1;
  end;

	type1 := RD_neg(current_block);

	if is_not<>0 then begin
		bytecode_append(current_block^.code, iconst_m1_);
		bytecode_append(current_block^.code, ixor_);
  end;

	if ((is_not<>0)
		and (type1^.type_class <> integer_type)
		and (type1^.type_class <> boolean_type)
		and (type1^.type_class <> error_type)) then
		MakeError(MSG_417, '');

	Result:= type1;
end;

function RD_mult(current_block: block): types;
var
	old_token: TToken;
	type1, type2: types;
  cast: integer;
  tmp: types;
begin

	type1 := RD_not(current_block);

	while Token in [_MUL,_SLASH,_DIV,_MOD,_AND] do begin

		old_token := Token;
		ReadToken;
		type2 := RD_not(current_block);

		if ((type2^.type_class <> integer_type)
			and (type2^.type_class <> real_type)
			and (type2^.type_class <> boolean_type)
			and (type2^.type_class <> error_type)) then
			MakeError(MSG_417,'')
		else begin

			cast := type_equal_cast(type1, type2);
			if (cast = 0) then MakeError(MSG_417, '');
			if (cast <> 1) then begin

				usesFloat:=1;
				if (cast = 3) then bytecode_append(current_block^.code, swap_);
				if (mathType = 1) then begin
					bytecode_append(current_block^.code, invokestatic_);
					bytecode_append_short_int(current_block^.code, cp_add_methodref('F', 'fI', '(I)I'));
				end	else begin
					bytecode_append(current_block^.code, new_);
					bytecode_append_short_int(current_block^.code, cp_add_class('Real'));
					bytecode_append(current_block^.code, dup_x1_);
					bytecode_append(current_block^.code, swap_);
					bytecode_append(current_block^.code, invokespecial_);
					bytecode_append_short_int(current_block^.code, cp_add_methodref('Real', '<init>', '(I)V'));
				end;
				if (cast = 3) then bytecode_append(current_block^.code, swap_);

				if (cast = 3) then begin
					tmp := type2;
					type2 := type1;
					type1 := tmp;
        end;

      end;

    end;

		if ((type1^.type_class = real_type)
			and (old_token <> _MUL)
			and (old_token <> _SLASH)) then MakeError(MSG_417, '');

		if ((type1^.type_class = boolean_type)
			and (old_token <> _AND)) then MakeError(MSG_417, '');

		if (type1^.type_class = integer_type) then begin
			case old_token of
			  _MUL:   bytecode_append(current_block^.code, imul_);
			  _SLASH,
			  _DIV:   bytecode_append(current_block^.code, idiv_);
			  _MOD:   bytecode_append(current_block^.code, irem_);
			  _AND:   bytecode_append(current_block^.code, iand_);
      end;
    end;

		if type1^.type_class = real_type then begin

			usesFloat := 1;
			case old_token of

			_MUL:
				begin
					if (mathType = 1) then begin
						bytecode_append(current_block^.code, invokestatic_);
						bytecode_append_short_int(current_block^.code,cp_add_methodref('F', 'M', '(II)I'));
					end	else begin
						bytecode_append(current_block^.code, new_);
						bytecode_append_short_int(current_block^.code, cp_add_class('Real'));
						bytecode_append(current_block^.code, dup_x1_);
						bytecode_append(current_block^.code, swap_);
						bytecode_append(current_block^.code, invokespecial_);
						bytecode_append_short_int(current_block^.code, cp_add_methodref('Real', '<init>', '(LReal;)V'));

						bytecode_append(current_block^.code, swap_);

						bytecode_append(current_block^.code, new_);
						bytecode_append_short_int(current_block^.code, cp_add_class('Real'));
						bytecode_append(current_block^.code, dup_x1_);
						bytecode_append(current_block^.code, swap_);
						bytecode_append(current_block^.code, invokespecial_); ///
						bytecode_append_short_int(current_block^.code, cp_add_methodref('Real', '<init>', '(LReal;)V'));

						bytecode_append(current_block^.code, dup_x1_);
						bytecode_append(current_block^.code, swap_);

						bytecode_append(current_block^.code, invokevirtual_);
						bytecode_append_short_int(current_block^.code, cp_add_methodref('Real', 'mul', '(LReal;)V'));
          end;
        end;
    _SLASH:
				begin
					if (mathType = 1) then begin
						bytecode_append(current_block^.code, invokestatic_);
						bytecode_append_short_int(current_block^.code, cp_add_methodref('F', 'D', '(II)I'));
					end else begin
						bytecode_append(current_block^.code, new_);
						bytecode_append_short_int(current_block^.code, cp_add_class('Real'));
						bytecode_append(current_block^.code, dup_x1_);
						bytecode_append(current_block^.code, swap_);
						bytecode_append(current_block^.code, invokespecial_);
						bytecode_append_short_int(current_block^.code, cp_add_methodref('Real', '<init>', '(LReal;)V'));

						bytecode_append(current_block^.code, swap_);

						bytecode_append(current_block^.code, new_);
						bytecode_append_short_int(current_block^.code, cp_add_class('Real'));
						bytecode_append(current_block^.code, dup_x1_);
						bytecode_append(current_block^.code, swap_);
						bytecode_append(current_block^.code, invokespecial_);
						bytecode_append_short_int(current_block^.code, cp_add_methodref('Real', '<init>', '(LReal;)V'));

						bytecode_append(current_block^.code, dup_x1_);
						bytecode_append(current_block^.code, swap_);

						bytecode_append(current_block^.code, invokevirtual_);
						bytecode_append_short_int(current_block^.code, cp_add_methodref('Real', 'div', '(LReal;)V'));
          end;
        end;

    end;

  end;

		if ((type1^.type_class = boolean_type)
			and (old_token = _AND)) then
			bytecode_append(current_block^.code, iand_);

  end;

	Result:= type1;
end;


function RD_sum(current_block: block): types;
var
	old_token: TToken;
	type1, type2: types;
  cast,method_index: integer;
  tmp: types;
  class_index,
	init_index,
	append_index,
	append2_index,
	toString_index: integer;
begin
	type1 := RD_mult(current_block);

	while Token in [_PLUS,_MINUS,_OR,_XOR] do begin

		old_token := Token;
    ReadToken;
		type2 := RD_mult(current_block);

		if not (type2^.type_class in [integer_type..string_type,error_type])
		    or ((type1^.type_class = string_type) and (old_token <> _PLUS)) then
			MakeError(MSG_417, '')
		else begin

			if (type1^.type_class = char_type) then begin
				method_index := cp_add_methodref('java/lang/String', 'valueOf', '(C)Ljava/lang/String;');
				bytecode_append(current_block^.code, swap_);
				bytecode_append(current_block^.code, invokestatic_);
				bytecode_append_short_int(current_block^.code, method_index);
				bytecode_append(current_block^.code, swap_);
				type1^.type_class := string_type;
      end;

			cast := type_equal_cast(type1, type2);
			
			if ((cast = 0) and (type1^.type_class <> string_type)) then MakeError(MSG_417, '');

			if ((type1^.type_class <> string_type) and (cast <> 1)) then begin

				usesFloat:=1;

				if (cast = 3) then bytecode_append(current_block^.code, swap_);
				if (mathType = 1) then begin
					bytecode_append(current_block^.code, invokestatic_);
					bytecode_append_short_int(current_block^.code, cp_add_methodref('F', 'fI', '(I)I'));
				end	else begin
					bytecode_append(current_block^.code, new_);
					bytecode_append_short_int(current_block^.code, cp_add_class('Real'));
					bytecode_append(current_block^.code, dup_x1_);
					bytecode_append(current_block^.code, swap_);
					bytecode_append(current_block^.code, invokespecial_);
					bytecode_append_short_int(current_block^.code, cp_add_methodref('Real', '<init>', '(I)V'));
        end;
				if (cast = 3) then bytecode_append(current_block^.code, swap_);
				if (cast = 3) then begin
					tmp := type2;
					type2 := type1;
					type1 := tmp;
        end;

      end;

    end;

		if ((type1^.type_class <> integer_type)
			and (type1^.type_class <> boolean_type)
			and (type1^.type_class <> error_type)
			and ((old_token = _OR) or (old_token = _XOR))) then
			MakeError(MSG_417, '');

		if ((type1^.type_class = integer_type)
			or (type1^.type_class = boolean_type)
			or ((type1^.type_class = real_type) and (mathType = 1))) then begin

			if (type1^.type_class = real_type) then usesFloat := 1;

			case old_token of
			  _PLUS : bytecode_append(current_block^.code, iadd_);
			  _MINUS: bytecode_append(current_block^.code, isub_);
			  _OR   : bytecode_append(current_block^.code, ior_);
			  _XOR  : bytecode_append(current_block^.code, ixor_);
      end;

    end;

		if ((type1^.type_class = real_type) and (mathType <> 1)) then begin

			usesFloat := 1;
			if old_token = _PLUS then begin
				bytecode_append(current_block^.code, new_);
				bytecode_append_short_int(current_block^.code, cp_add_class('Real'));
				bytecode_append(current_block^.code, dup_x1_);
				bytecode_append(current_block^.code, swap_);
				bytecode_append(current_block^.code, invokespecial_);
				bytecode_append_short_int(current_block^.code, cp_add_methodref('Real', '<init>', '(LReal;)V'));

				bytecode_append(current_block^.code, swap_);

				bytecode_append(current_block^.code, new_);
				bytecode_append_short_int(current_block^.code, cp_add_class('Real'));
				bytecode_append(current_block^.code, dup_x1_);
				bytecode_append(current_block^.code, swap_);
				bytecode_append(current_block^.code, invokespecial_);
				bytecode_append_short_int(current_block^.code, cp_add_methodref('Real', '<init>', '(LReal;)V'));

				bytecode_append(current_block^.code, dup_x1_);
				bytecode_append(current_block^.code, swap_);

				bytecode_append(current_block^.code, invokevirtual_);
				bytecode_append_short_int(current_block^.code, cp_add_methodref('Real', 'add', '(LReal;)V'));
      end;

			if old_token = _MINUS then begin

				bytecode_append(current_block^.code, new_);
				bytecode_append_short_int(current_block^.code, cp_add_class('Real'));
				bytecode_append(current_block^.code, dup_x1_);
				bytecode_append(current_block^.code, swap_);
				bytecode_append(current_block^.code, invokespecial_);
				bytecode_append_short_int(current_block^.code, cp_add_methodref('Real', '<init>', '(LReal;)V'));

				bytecode_append(current_block^.code, swap_);

				bytecode_append(current_block^.code, new_);
				bytecode_append_short_int(current_block^.code, cp_add_class('Real'));
				bytecode_append(current_block^.code, dup_x1_);
				bytecode_append(current_block^.code, swap_);
				bytecode_append(current_block^.code, invokespecial_);
				bytecode_append_short_int(current_block^.code, cp_add_methodref('Real', '<init>', '(LReal;)V'));

				bytecode_append(current_block^.code, dup_x1_);
				bytecode_append(current_block^.code, swap_);

				bytecode_append(current_block^.code, invokevirtual_);
				bytecode_append_short_int(current_block^.code, cp_add_methodref('Real', 'sub', '(LReal;)V'));
      end;

    end;

		if ((type1^.type_class = string_type) and (old_token = _PLUS)) then begin

			if (type2^.type_class = real_type) then begin

				usesFloat := 1;

				if (mathType = 1) then begin
					bytecode_append(current_block^.code, invokestatic_);
					bytecode_append_short_int(current_block^.code, cp_add_methodref('F', 'tS', '(I)Ljava/lang/String;'));
				end	else begin
					bytecode_append(current_block^.code, invokevirtual_);
					bytecode_append_short_int(current_block^.code, cp_add_methodref('Real', 'toString', '()Ljava/lang/String;'));
        end;

      end;
			
			class_index := cp_add_class('java/lang/StringBuffer');
			init_index := cp_add_methodref('java/lang/StringBuffer', '<init>', '()V');
			append_index := cp_add_methodref('java/lang/StringBuffer', 'append', '(Ljava/lang/String;)Ljava/lang/StringBuffer;');
			toString_index := cp_add_methodref('java/lang/StringBuffer', 'toString', '()Ljava/lang/String;');

			case (type2^.type_class) of
    			string_type,
    			real_type:    append2_index := cp_add_methodref('java/lang/StringBuffer', 'append', '(Ljava/lang/String;)Ljava/lang/StringBuffer;');
    			integer_type: append2_index := cp_add_methodref('java/lang/StringBuffer', 'append', '(I)Ljava/lang/StringBuffer;');
    			char_type:
                      begin
                				bytecode_append(current_block^.code, i2c_);
                				append2_index := cp_add_methodref('java/lang/StringBuffer', 'append', '(C)Ljava/lang/StringBuffer;');
                      end;
    			boolean_type:
                      begin
                				bytecode_append(current_block^.code, i2b_);
                				append2_index := cp_add_methodref('java/lang/StringBuffer', 'append', '(Z)Ljava/lang/StringBuffer;');
                			end;
			else
				MakeError(MSG_434,'');
      end;

			bytecode_append(current_block^.code, swap_);

			bytecode_append(current_block^.code, new_);
			bytecode_append_short_int(current_block^.code, class_index);
			bytecode_append(current_block^.code, dup_);
			bytecode_append(current_block^.code, invokespecial_);
			bytecode_append_short_int(current_block^.code, init_index);

			bytecode_append(current_block^.code, swap_);
			bytecode_append(current_block^.code, invokevirtual_);
			bytecode_append_short_int(current_block^.code, append_index);

			bytecode_append(current_block^.code, swap_);

			bytecode_append(current_block^.code, invokevirtual_);
			bytecode_append_short_int(current_block^.code, append2_index);

			bytecode_append(current_block^.code, invokevirtual_);
			bytecode_append_short_int(current_block^.code, toString_index);

    end;

  end;

	Result:= type1;

end;

function RD_expression(current_block: block): types;
var
	old_token: TToken;
	type1, type2: types;
  cast,method_index: integer;
  tmp: types;
begin
	type1 := RD_sum(current_block);

	if Token in [_EQUAL.._NOT_EQ] then begin

		old_token := Token;
		ReadToken;

		type2 := RD_sum(current_block);

		if not (type2^.type_class in [integer_type..string_type,error_type,command_type]) then
			MakeError(MSG_417, '')
		else begin

			cast := type_equal_cast(type1, type2);
			if cast = 0 then MakeError(MSG_417, '');

			if ((type1^.type_class = integer_type)
				or (type2^.type_class = integer_type)) then begin

				if (cast <> 1) then begin
					usesFloat:=1;
					if (cast = 3) then bytecode_append(current_block^.code, swap_);
					if (mathType = 1) then begin
						bytecode_append(current_block^.code, invokestatic_);
						bytecode_append_short_int(current_block^.code, cp_add_methodref('F', 'fI', '(I)I'));
					end else begin
						bytecode_append(current_block^.code, new_);
						bytecode_append_short_int(current_block^.code, cp_add_class('Real'));
						bytecode_append(current_block^.code, dup_x1_);
						bytecode_append(current_block^.code, swap_);
						bytecode_append(current_block^.code, invokespecial_);
						bytecode_append_short_int(current_block^.code, cp_add_methodref('Real', '<init>', '(I)V'));
					end;
					if (cast = 3) then bytecode_append(current_block^.code, swap_);
				end

			end else begin

				if (cast = 2) or (cast = 3) then begin
					method_index := cp_add_methodref('java/lang/String', 'valueOf', '(C)Ljava/lang/String;');
					if (cast = 3) then	bytecode_append(current_block^.code, swap_);
					bytecode_append(current_block^.code, invokestatic_);
					bytecode_append_short_int(current_block^.code, method_index);
					if (cast = 3) then bytecode_append(current_block^.code, swap_);
        end;

      end;

			if (cast = 3) then begin
				tmp := type2;
				type2 := type1;
				type1 := tmp;
      end;

			if ((type1^.type_class = integer_type)
				or (type1^.type_class = boolean_type)
				or (type1^.type_class = char_type)
				or ((type1^.type_class = real_type) and (mathType = 1))) then begin

				if type1^.type_class = real_type then usesFloat := 1;

				case old_token of
				  _LESS:	      bytecode_append(current_block^.code, if_icmplt_);
				  _LESS_EQ:     bytecode_append(current_block^.code, if_icmple_);
				  _GREATER_EQ:  bytecode_append(current_block^.code, if_icmpge_);
				  _GREATER:     bytecode_append(current_block^.code, if_icmpgt_);
				  _EQUAL:       bytecode_append(current_block^.code, if_icmpeq_);
				  _NOT_EQ:      bytecode_append(current_block^.code, if_icmpne_);
        end;

				bytecode_append_short_int(current_block^.code, 7);
				bytecode_append(current_block^.code, iconst_0_);
				bytecode_append(current_block^.code, goto_);
				bytecode_append_short_int(current_block^.code, 4);
				bytecode_append(current_block^.code, iconst_m1_);

      end;

			if ((type1^.type_class = real_type) and (mathType <> 1)) then begin

				usesFloat := 1;
				bytecode_append(current_block^.code, invokevirtual_);

				case old_token of
				  _LESS:      bytecode_append_short_int(current_block^.code, cp_add_methodref('Real', 'lessThan', '(LReal;)Z'));
				  _LESS_EQ:   bytecode_append_short_int(current_block^.code, cp_add_methodref('Real', 'lessEqual', '(LReal;)Z'));
				  _GREATER_EQ:bytecode_append_short_int(current_block^.code, cp_add_methodref('Real', 'greaterEqual', '(LReal;)Z'));
				  _GREATER:   bytecode_append_short_int(current_block^.code, cp_add_methodref('Real', 'greaterThan', '(LReal;)Z'));
				  _EQUAL:     bytecode_append_short_int(current_block^.code, cp_add_methodref('Real', 'equalTo', '(LReal;)Z'));
				  _NOT_EQ:    bytecode_append_short_int(current_block^.code, cp_add_methodref('Real', 'notEqualTo', '(LReal;)Z'));
        end;

				bytecode_append(current_block^.code, ifne_);
				bytecode_append_short_int(current_block^.code, 7);
				bytecode_append(current_block^.code, iconst_0_);
				bytecode_append(current_block^.code, goto_);
				bytecode_append_short_int(current_block^.code, 4);
				bytecode_append(current_block^.code, iconst_m1_);

      end;

			if (type1^.type_class = command_type) then begin

				if (old_token = _EQUAL) then
					bytecode_append(current_block^.code, if_acmpeq_)
				else if (old_token = _NOT_EQ) then
					bytecode_append(current_block^.code, if_acmpne_)
				else
					MakeError(MSG_417, '');

				bytecode_append_short_int(current_block^.code, 7);
				bytecode_append(current_block^.code, iconst_0_);
				bytecode_append(current_block^.code, goto_);
				bytecode_append_short_int(current_block^.code, 4);
				bytecode_append(current_block^.code, iconst_m1_);

      end;

    end;

		if (type1^.type_class = string_type) then begin

			method_index := cp_add_methodref('java/lang/String', 'compareTo', '(Ljava/lang/String;)I');
			bytecode_append(current_block^.code, invokevirtual_);
			bytecode_append_short_int(current_block^.code, method_index);

			case old_token of
			  _LESS       : bytecode_append(current_block^.code, iflt_);
			  _LESS_EQ    : bytecode_append(current_block^.code, ifle_);
			  _GREATER_EQ : bytecode_append(current_block^.code, ifge_);
			  _GREATER    : bytecode_append(current_block^.code, ifgt_);
			  _EQUAL      : bytecode_append(current_block^.code, ifeq_);
			  _NOT_EQ     : bytecode_append(current_block^.code, ifne_);
      end;

			bytecode_append_short_int(current_block^.code, 7);
      bytecode_append(current_block^.code, iconst_0_);
      bytecode_append(current_block^.code, goto_);
      bytecode_append_short_int(current_block^.code, 4);
      bytecode_append(current_block^.code, iconst_m1_);

    end;

		type1 := type_create();
		type1^.type_class := boolean_type;

  end;

	Result:= type1;

end;

procedure RD_if_statement(current_block: block);
var
	expression_type: types;
	jump_offset_position: longint;
	offset: shortint ;
  yycopy: char;
	i: integer;
begin
	expression_type := RD_expression(current_block);

	if ((expression_type^.type_class <> error_type)
		and (expression_type^.type_class <> boolean_type)) then
		MakeError(MSG_407, 'boolean');

	Match(_THEN);

	bytecode_append(current_block^.code, ifeq_);
	jump_offset_position := current_block^.code^.bytecode_pos;
	bytecode_append(current_block^.code, nop_);
	bytecode_append(current_block^.code, nop_);

	RD_statement(current_block);

	while Token = _SEMICOLON do ReadToken;

	if Token = _ELSE then begin
		offset := current_block^.code^.bytecode_pos + 4 - jump_offset_position;
		current_block^.code^.bytecode[jump_offset_position] := char(offset shr 8);
		current_block^.code^.bytecode[jump_offset_position+1] := char(offset);

		bytecode_append(current_block^.code, goto_);
		jump_offset_position := current_block^.code^.bytecode_pos;
		bytecode_append(current_block^.code, nop_);
		bytecode_append(current_block^.code, nop_);

		ReadToken;

		RD_statement(current_block);

		offset := current_block^.code^.bytecode_pos - jump_offset_position + 1;
		current_block^.code^.bytecode[jump_offset_position] := char(offset shr 8);
		current_block^.code^.bytecode[jump_offset_position+1] := char(offset);
	end	else begin

		offset := current_block^.code^.bytecode_pos - jump_offset_position + 1;
		current_block^.code^.bytecode[jump_offset_position] := char(offset shr 8);
		current_block^.code^.bytecode[jump_offset_position+1] := char(offset);
    {
		yycopy = strdup(yytext);
		for(i = strlen(yytext) - 1; i >= 0; i--) yyunput(yycopy[i], yytext);
		yyunput(';', yytext);
		free(yycopy);
		}
		ReadToken;
  end;
  
end;

procedure RD_case_list(current_block: block; case_type: types);
var
  first_pass: shortint;
  expected_type_name: string;
  constant_type: types;
	constant_type_name: string;
begin
	first_pass := 1;

	repeat

		if (first_pass<>0) then
			first_pass := 0
		else
			ReadToken;

		case Token of

		_INTEGER:
			begin
				if (case_type^.type_class <> integer_type) then begin
					expected_type_name := type_get_name(case_type);
					MakeError(MSG_409, expected_type_name+'integer');
				end;
				ReadToken;
      end;

		_REAL:
			begin
				expected_type_name := type_get_name(case_type);
				MakeError(MSG_409, expected_type_name+'real');
				ReadToken;
			end;

		_BOOLEAN:
			begin
				expected_type_name := type_get_name(case_type);
				MakeError(MSG_409, expected_type_name+'boolean');
				ReadToken;
			end;

		_CHAR:
			begin
				if (case_type^.type_class <> char_type) then begin
					expected_type_name := type_get_name(case_type);
					MakeError(MSG_409, expected_type_name+'char');
				end;
				ReadToken;
      end;

		_STRING:
			begin
				if (case_type^.type_class <> string_type) then begin
					expected_type_name := type_get_name(case_type);
					MakeError(MSG_409, expected_type_name+'string');
        end;
        ReadToken;
      end;

		_IDENT:
			begin
				constant_type := get_constant_type(current_block, TokenStr);
				if (constant_type^.type_class = error_type) then
					MakeError(MSG_410, TokenStr)
				else
					if (type_equal(constant_type, case_type)=0) then begin
						expected_type_name := type_get_name(case_type);
						constant_type_name := type_get_name(constant_type);
						MakeError(MSG_409, expected_type_name+constant_type_name);
          end;
					ReadToken;
      end;

		else
				MakeError(MSG_206, '');
		end;


	until (Token <> _COMMA);

	ReadToken;
	Match(_COLON);
	RD_statement(current_block);

end;


procedure RD_case_statement(current_block: block);
var
	expression_type: types;
begin
  expression_type := RD_expression(current_block);
	MakeError(MSG_442, '');

	if ((expression_type^.type_class <> error_type)
		and (expression_type^.type_class <> integer_type)
		and (expression_type^.type_class <> char_type)
		and (expression_type^.type_class <> string_type)) then
		MakeError(MSG_408,'');

	Match(_OF);
	RD_case_list(current_block, expression_type);
	Match(_END);

end;

procedure RD_with_statement(current_block: block);
begin
	while Token <> _DO do begin
		if Token = _TOEOF then Exit;
		ReadToken;
	end;
	ReadToken;
	RD_statement(current_block);
end;

procedure RD_repeat_statement(current_block: block);
var
	expression_type: types;
	pos1,
	break_pos1, break_pos2: integer;
begin
	pos1 := current_block^.code^.bytecode_pos;
	Inc(inside_loop);
	break_pos1 := current_block^.code^.bytecode_pos;
	RD_block_body(current_block);
	break_pos2 := current_block^.code^.bytecode_pos;
	Dec(inside_loop);

	Match(_UNTIL);

	expression_type := RD_expression(current_block);

	bytecode_append(current_block^.code, ifeq_);
	bytecode_append_short_int(current_block^.code, pos1 - current_block^.code^.bytecode_pos + 1);

	if ((expression_type^.type_class <> error_type)
  and (expression_type^.type_class <> boolean_type)) then
		MakeError(MSG_407, 'boolean');

	transform_break_stmts(current_block^.code, break_pos1, break_pos2, current_block^.code^.bytecode_pos);

end;

procedure RD_while_statement(current_block: block);
var
	expression_type: types;
	pos1, pos2,
	jump_offset_position,
	jump_offset,
	break_pos1, break_pos2: integer;
begin
	pos1 := current_block^.code^.bytecode_pos;
	
	expression_type := RD_expression(current_block);

	pos2 := current_block^.code^.bytecode_pos;
	bytecode_append(current_block^.code, ifeq_);
	jump_offset_position := current_block^.code^.bytecode_pos;
	bytecode_append_short_int(current_block^.code, 0);

	if ((expression_type^.type_class <> error_type)
		and (expression_type^.type_class <> boolean_type)) then
		MakeError(MSG_407, 'boolean');

	Match(_DO);

	Inc(inside_loop);
	break_pos1 := current_block^.code^.bytecode_pos;
	RD_statement(current_block);
	break_pos2 := current_block^.code^.bytecode_pos;
	Dec(inside_loop);

	bytecode_append(current_block^.code, goto_);
	bytecode_append_short_int(current_block^.code, pos1 - current_block^.code^.bytecode_pos + 1);

	jump_offset := current_block^.code^.bytecode_pos - pos2;
	current_block^.code^.bytecode[jump_offset_position] := char(jump_offset shr 8);
	current_block^.code^.bytecode[jump_offset_position + 1] := char(jump_offset);

	transform_break_stmts(current_block^.code, break_pos1, break_pos2, current_block^.code^.bytecode_pos);

end;

procedure RD_for_statement(current_block: block);
var
	iterator_type: types;
	expression1_type,
  expression2_type: types;
	iterator: identifier;
	iterator_name: string;
	is_field,
	check_bytecode_pos,
	direction,
	evaluate_pos,
	break_pos1, break_pos2: integer;
begin
	iterator_type := get_variable_type(current_block, TokenStr);
	iterator := get_identifier(current_block, TokenStr);
	iterator_name := TokenStr;

	if (iterator^.identifier_class = unit_name) then
		MakeError(MSG_456, '');

	if ((iterator_type^.type_class <> integer_type)
		and (iterator_type^.type_class <> char_type)) then
		MakeError(MSG_411, '');

	Match(_IDENT);
	Match(_ASSIGN);

	expression1_type := RD_expression(current_block);

	is_field := iterator^.belongs_to_program_block;
	create_put_variable_bytecode(iterator, current_block^.code, iterator_name, is_field);

	if (type_equal(iterator_type, expression1_type)=0) then
		MakeError(MSG_412, '');

	case (Token) of
	_TO:
		begin
			direction := 1;
			ReadToken;
		end;

	_DOWNTO:
		begin
			direction := -1;
			ReadToken;;
		end;

	else
    MakeError(MSG_204, '');
	end;

	evaluate_pos := current_block^.code^.bytecode_pos;
	expression2_type := RD_expression(current_block);

	create_variable_bytecode(iterator, current_block^.code, iterator_name, iterator^.belongs_to_program_block);

	check_bytecode_pos := current_block^.code^.bytecode_pos;
	if (direction = 1) then
		bytecode_append(current_block^.code, if_icmplt_)
	else
		bytecode_append(current_block^.code, if_icmpgt_);

	bytecode_append_short_int(current_block^.code, 0);

	if (type_equal(iterator_type, expression2_type)=0) then
		MakeError(MSG_412, '');

	Match(_DO);

	Inc(inside_loop);
	break_pos1 := current_block^.code^.bytecode_pos;
	RD_statement(current_block);
	break_pos2 := current_block^.code^.bytecode_pos;
	Dec(inside_loop);

	// TODO:: moguca optimizacija za neke slucajeve sa iinc_ naredbom
	create_variable_bytecode(iterator, current_block^.code, iterator_name, iterator^.belongs_to_program_block);

	if (direction = 1) then
		bytecode_append(current_block^.code, iconst_1_)
	else
		bytecode_append(current_block^.code, iconst_m1_);

	bytecode_append(current_block^.code, iadd_);

	create_put_variable_bytecode(iterator, current_block^.code, iterator_name, iterator^.belongs_to_program_block);

	bytecode_append(current_block^.code, goto_);
	bytecode_append_short_int(current_block^.code, evaluate_pos - current_block^.code^.bytecode_pos + 1);

	current_block^.code^.bytecode[check_bytecode_pos + 1] := char((current_block^.code^.bytecode_pos - check_bytecode_pos) shr 8);
	current_block^.code^.bytecode[check_bytecode_pos + 2] := char(current_block^.code^.bytecode_pos - check_bytecode_pos);


	transform_break_stmts(current_block^.code, break_pos1, break_pos2, current_block^.code^.bytecode_pos);

end;
(*
procedure RD_assignment_or_procedure_call(current_block: block);
var
	name: identifier;
	element_name: string;
	identifier_text: string;
  params: type_list;
  units_name: identifier;
label:
  procedure_call,
  variable_lvalue;
begin
	name := get_identifier(current_block, TokenStr);
	element_name := TokenStr;

	if (Token <> _IDENT) then
	  MakeError(MSG_202, '');

	if ((name^.identifier_class = none)
		or (name^.identifier_class = program_name)
		or (name^.identifier_class = constant_name)
		or (name^.identifier_class = type_name)) then
		MakeError(MSG_428, '');

	ReadToken;

	if (name^.identifier_class = unit_name) then begin

		Match(_POINT);


		units_name := name_table_find(name^.unit_block^.names, TokenStr);
		identifier_text:=TokenStr;
		element_name := TokenStr;

		if (units_name = NIL) then
			MakeError(MSG_453, '');

		name := units_name;
		ReadToken;

		if (name^.identifier_class = procedure_name) then goto procedure_call;
		if (name^.identifier_class = variable_name) then goto variable_lvalue;

		MakeError(MSG_457, linenum, TokenStr, '');
	}

	if (name^.identifier_class = procedure_name) then begin
procedure_call:
		if (name^.standard_function) then
			;//create_std_function_prefix(current_block^.code, element_name^.cstr);

		if (type_list_length(name^.parameters) == 0) then begin
			if (Token == _LBRACKET) then
				MakeError(MSG_419, '');
		end	else begin
			if (Token <> _LBRACKET) then
				MakeError(MSG_420, '')
			else begin
				int compare_result;
				int old_linenum;
				type_list *parameter_list;
				Token = yylex();

				old_linenum = linenum;

				parameter_list = RD_expression_list_cast(current_block, name^.parameters);

				compare_result = type_list_different_parameter_cast(parameter_list, name^.parameters);

				if (compare_result == -1)
					MakeError(MSG_421, old_linenum, '', '');

				if (compare_result > 0)
				{
					char par_num[4];
					sprintf(par_num, '%d', compare_result);

					MakeError(MSG_422, old_linenum, par_num, '');
				}

				type_list_destroy(parameter_list);

				if (Token <> CLOSE_BR)
				{
					MakeError(MSG_200, linenum, ')', TokenStr);
				}
				else
					Token = yylex();
			}
		}

		if (!name^.standard_function)
		{
			int methodref_index;
			identifier *block_identifier;
			type_list *it;
			int pos;

			char* descriptor = (char) mem_alloc(5*1024);
			if (descriptor == NULL)
				die(MSG_001);
			descriptor[0] = '(';

			if (!name^.unit_function)
			{
				block_identifier = name_table_find(current_block^.names, element_name);

				if (block_identifier == NULL)
					block_identifier = name_table_find(current_block^.parent_block^.names, element_name);
			}
			else
			{
				block_identifier = name;
			}

			it = block_identifier^.parameters;
			pos = 1;

			while(it <> NULL)
			{
				if (it^.data == NULL)
					break;

				get_field_descriptor(it^.data, descriptor + pos);
				pos = strlen(descriptor);
				it = it^.next;
			}

			descriptor[pos] = ')';
			pos ++;
			descriptor[pos] = '\0';

			if (block_identifier^.identifier_class == procedure_name)
				strcat(descriptor, 'V');
			else
				get_field_descriptor(block_identifier^.return_type, descriptor + pos);


			lowercase(element_name^.cstr);
			if (name^.unit_function)
			{
				string *full_unit_name;

				if (name^.container_unit^.is_library)
					full_unit_name = string_from_cstr('Lib_');
				else
					full_unit_name = string_create();
				string_append(full_unit_name, name^.container_unit^.unit_name);
				methodref_index = cp_add_methodref(full_unit_name^.cstr, element_name^.cstr, descriptor);
				string_destroy(full_unit_name);
			}
			else
			{
				if (compiling_unit == 0)
					methodref_index = cp_add_methodref('M', element_name^.cstr, descriptor);
				else
					methodref_index = cp_add_methodref(string_get_cstr(str_program_name), element_name^.cstr, descriptor);
			}
			bytecode_append(current_block^.code, invokestatic_); /* call */
			bytecode_append_short_int(current_block^.code, methodref_index);

			mem_free(descriptor);
		}
		else
		{
			/* handle standard functions and procedures */
			create_std_function_code(current_block^.code, element_name^.cstr);
		}
	}

	/* if the identifier is the function name and that is not the name of the current
	block, report an error */
	if (name^.identifier_class == function_name)
	{
		if (STRING_COMPARE(element_name^.cstr, current_block^.block_name^.cstr) <> 0)
		{
			MakeError(MSG_432, linenum, element_name^.cstr, '');
		}
	}

	/* if an identifier is a valid l-value */
	if ((name^.identifier_class == variable_name)
		|| (name^.identifier_class == parameter_name)
		|| (name^.identifier_class == function_name)) /* function names are treated as variables
													     for return values */
	{
		type *current_type;
		type *new_type;
		type *expression_type;
		int assign_linenum;
		int is_field = 0;
		bytecode *get_bytecode;
		bytecode *put_bytecode;
variable_lvalue:
		is_field = 0;
		get_bytecode = bytecode_create();
		put_bytecode = bytecode_create();

		if (name^.identifier_class == variable_name)
			current_type = type_duplicate(name^.variable_type);

		if (name^.identifier_class == parameter_name)
			current_type = type_duplicate(name^.parameter_type);

		if (name^.identifier_class == function_name)
			current_type = type_duplicate(name^.return_type);

		/* create the get and put bytecodes */
		is_field = name^.belongs_to_program_block;

		if (name^.identifier_class == function_name)
			is_field = 0;

		create_variable_bytecode(name, get_bytecode, element_name^.cstr, is_field);
		create_put_variable_bytecode(name, put_bytecode, element_name^.cstr, is_field);

		identifier_destroy(name);
		name = NULL;
		string_destroy(element_name);
		element_name = NULL;

		while (Token <> OP_ASSIGN)
		{
			if ((Token == SEMI_COLON)
				|| (Token == KWD_END)
				|| (Token == END_OF_INPUT))
			{
				MakeError(MSG_218, linenum, '', '');
				type_destroy(current_type);
				bytecode_destroy(get_bytecode);
				bytecode_destroy(put_bytecode);
				return;
			}

			if (Token == DOT)
			{
				int fieldref_index;
				char record_name[15];
				char field_descriptor[128];

				Token = yylex();

				/* append the old get bytecode */
				bytecode_append_bytecode(current_block^.code, get_bytecode);

				/* reset the bytecodes */
				put_bytecode^.bytecode_pos = 0;
				get_bytecode^.bytecode_pos = 0;

				if (current_type^.type_class <> record_type)
				{
					MakeError(MSG_424, linenum, '', '');

					/* Error-recovery */
					while ((Token <> SEMI_COLON)
						&& (Token <> KWD_END)
						&& (Token <> END_OF_INPUT))
					{
						Token = yylex();
					}

					type_destroy(current_type);
					bytecode_destroy(get_bytecode);
					bytecode_destroy(put_bytecode);
					return;
				}

				new_type = type_find_record_element(current_type, TokenStr);

				if (new_type == NULL)
				{
					MakeError(MSG_447, linenum, TokenStr, '');
				}
				else
				{
					/* create the new put and get bytecodes */
					sprintf(record_name, 'R_%d', current_type^.unique_record_ID);
					get_field_descriptor(new_type, field_descriptor);
					lowercase(TokenStr);
					fieldref_index = cp_add_fieldref(record_name, TokenStr, field_descriptor);

					bytecode_append(get_bytecode, getfield_);
					bytecode_append_short_int(get_bytecode, fieldref_index);

					bytecode_append(put_bytecode, putfield_);
					bytecode_append_short_int(put_bytecode, fieldref_index);


					type_destroy(current_type);
					current_type = new_type;
				}


				if (current_type == NULL)
				{
					MakeError(MSG_425, linenum, TokenStr, '');

					/* Error-recovery */
					while ((Token <> SEMI_COLON)
						&& (Token <> KWD_END)
						&& (Token <> END_OF_INPUT))
					{
						Token = yylex();
					}

					type_destroy(current_type);
					bytecode_destroy(get_bytecode);
					bytecode_destroy(put_bytecode);
					return;
				}

				Token = yylex();
			}
			else if (Token == OPEN_SQ_BR)
			{
				type_list *array_elements;
				bytecode *tmp_code;
				tmp_code = bytecode_create();

				// ako je [, a zadnji je array, procitaj listu, usporedi, vrati slijedeci, skoci gore
				if (current_type^.type_class <> array_type)
				{
					MakeError(MSG_426, linenum, '', '');

					/* Error-recovery */
					while ((Token <> SEMI_COLON)
						&& (Token <> KWD_END)
						&& (Token <> END_OF_INPUT))
					{
						Token = yylex();
					}

					type_destroy(current_type);
					bytecode_destroy(get_bytecode);
					bytecode_destroy(put_bytecode);
					return;
				}

				Token = yylex();

				array_elements = RD_expression_list_array(current_block, tmp_code, current_type);

				bytecode_append_bytecode(get_bytecode, tmp_code);
				//bytecode_append_bytecode(put_bytecode, tmp_code);
				bytecode_destroy(put_bytecode);
				put_bytecode = bytecode_duplicate(get_bytecode);

				replace_aaload_instruction(get_bytecode, current_type^.element_type);

				/* remove the last aaload from put bytecode, adjust the stack and create the
				 put bytecode */
				put_bytecode^.bytecode_pos --;
				bytecode_append(put_bytecode, dup2_x1_);
				bytecode_append(put_bytecode, pop2_);

				switch(current_type^.element_type^.type_class)
				{
				case integer_type:
				case boolean_type:
				case char_type:
					bytecode_append(put_bytecode, iastore_);
					break;

				case real_type:
					if (mathType == 1)
						bytecode_append(put_bytecode, iastore_);
					else
						bytecode_append(put_bytecode, aastore_);
					break;


				case image_type:
				case string_type:
				case record_type:
				case command_type:
				case record_store_type:
				case http_type:
				case alert_type:
				case stream_type:
					bytecode_append(put_bytecode, aastore_);
					break;

				}

				bytecode_destroy(tmp_code);

				if (type_list_different_parameter_array(array_elements, current_type^.dimensions_list) <> 0)
				{
					char num1[6], num2[6];

					sprintf(num1, '%d', type_list_length(current_type^.dimensions_list));
					sprintf(num2, '%d', type_list_length(array_elements));
					MakeError(MSG_427, linenum, num1, num2);

					/* Error-recovery */
					while ((Token <> SEMI_COLON)
						&& (Token <> KWD_END)
						&& (Token <> END_OF_INPUT))
					{
						Token = yylex();
					}

					type_destroy(current_type);
					bytecode_destroy(get_bytecode);
					bytecode_destroy(put_bytecode);
					return;
				}

				if (Token <> CLOSE_SQ_BR)
				{
					MakeError(MSG_200, linenum, ']', TokenStr);

					/* Error-recovery */
					while ((Token <> SEMI_COLON)
						&& (Token <> KWD_END)
						&& (Token <> END_OF_INPUT))
					{
						Token = yylex();
					}

					type_destroy(current_type);
					bytecode_destroy(get_bytecode);
					bytecode_destroy(put_bytecode);
					return;
				}

				Token = yylex();

				new_type = type_duplicate(current_type^.element_type);
				type_destroy(current_type);

				current_type = new_type;

				type_list_destroy(array_elements);
			}
			else
			{
				MakeError(MSG_204, linenum, TokenStr, '');

				/* Error-recovery */
				while ((Token <> SEMI_COLON)
					&& (Token <> KWD_END)
					&& (Token <> END_OF_INPUT))
				{
					Token = yylex();
				}
				bytecode_destroy(get_bytecode);
				bytecode_destroy(put_bytecode);
				return;
			}
		}

		assign_linenum = linenum;

		Token = yylex();

		expression_type = RD_expression(current_block);

		if ((type_equal_cast(current_type, expression_type) <> 1)
			&& (current_type^.type_class == real_type))
		{
			usesFloat = 1;

			if (mathType == 1)
			{
				bytecode_append(current_block^.code, invokestatic_);
				bytecode_append_short_int(current_block^.code, cp_add_methodref('F', 'fI', '(I)I'));
			}
			else
			{
				bytecode_append(current_block^.code, new_);
				bytecode_append_short_int(current_block^.code, cp_add_class('Real'));
				bytecode_append(current_block^.code, dup_x1_);
				bytecode_append(current_block^.code, swap_);
				bytecode_append(current_block^.code, invokespecial_);
				bytecode_append_short_int(current_block^.code, cp_add_methodref('Real', '<init>', '(I)V'));
			}
		}

		if ((type_equal_cast(current_type, expression_type) == 2)
			&& (current_type^.type_class == string_type))
		{
			int method_index = cp_add_methodref('java/lang/String', 'valueOf', '(C)Ljava/lang/String;');

			/* string cast */
			bytecode_append(current_block^.code, invokestatic_);
			bytecode_append_short_int(current_block^.code, method_index);
		}

		bytecode_append_bytecode(current_block^.code, put_bytecode);

		if ((type_equal_cast(current_type, expression_type) <> 1)
			&& (type_equal_cast(current_type, expression_type) <> 2))
		{
			MakeError(MSG_423, assign_linenum, '', '');
		}
		else
			if (current_type^.type_class == array_type)
			{
				MakeError(MSG_443,assign_linenum, '', '');
			}



		type_destroy(expression_type);
		type_destroy(current_type);
		bytecode_destroy(get_bytecode);
		bytecode_destroy(put_bytecode);
	}

	if (name <> NULL)
		identifier_destroy(name);

	if (element_name <> NULL)
		string_destroy(element_name);
}
*)
procedure RD_statement(current_block: block);
begin
	case Token of

	_BEGIN:
		begin
			ReadToken;
			RD_block_body(current_block);
			Match(_END);
		end;

	_END,
  _UNTIL: exit;

	_IF:
		begin
			ReadToken;
			RD_if_statement(current_block);
		end;

	_CASE:
		begin
			ReadToken;
			RD_case_statement(current_block);
    end;

	_WHILE:
		begin
			ReadToken;
			RD_while_statement(current_block);
		end;

	_REPEAT:
		begin
			ReadToken;
			RD_repeat_statement(current_block);
		end;

	_FOR:
		begin
			ReadToken;;
			RD_for_statement(current_block);
		end;

	_WITH:
    begin
			ReadToken;
			RD_with_statement(current_block);
		end;

	_IDENT:writeconsole('i',1,2);
		{
			RD_assignment_or_procedure_call(current_block);
		}

	_BREAK:
		begin
			ReadToken;
			if (inside_loop <= 0) then
				MakeError(MSG_219, '')
			else begin
				bytecode_append(current_block^.code, break_stmt_);
				bytecode_append_short_int(current_block^.code, 0);
			end;
		end;
	else
		MakeError(MSG_204, TokenStr);
  end;

end;

procedure RD_block_body(current_block: block);
var
  first_pass: shortint;
begin
	repeat
    if (first_pass<>0) then
			first_pass := 0
		else
			ReadToken;

		RD_statement(current_block);
	until Token <>_SEMICOLON;
end;

function RD_proc_block(current_block: block; return_type: types; parameters: type_list): integer;
var
	forward_declaration: integer;
begin

  forward_declaration := 0;

	if ((Token = _FORWARD)
  or (inside_interface_part<>0)) then begin
		forward_declaration := 1;
		if (inside_interface_part = 0) then ReadToken;
	end	else begin

		if (inside_interface_part<>0) then MakeError(MSG_221, '');

		RD_block(current_block);
    SetLength(current_block^.parent_block^.children,Length(current_block^.parent_block^.children)+1);
		current_block^.parent_block^.children_count:=Length(current_block^.parent_block^.children)+1;

		case return_type^.type_class of

        void_type:  bytecode_append(current_block^.code, return_);
        real_type:
                    begin
                      usesFloat:=1;
                      if (mathType = 1) then begin
                        bytecode_append(current_block^.code, iload_);
                        bytecode_append(current_block^.code, Jmnem(type_list_length(parameters)));
                        bytecode_append(current_block^.code, ireturn_);
                      end	else begin
                        case type_list_length(parameters) of
                          0:  bytecode_append(current_block^.code, aload_0_);
                          1:  bytecode_append(current_block^.code, aload_1_);
                          2:  bytecode_append(current_block^.code, aload_2_);
                          3:  bytecode_append(current_block^.code, aload_3_);
                        else
                          bytecode_append(current_block^.code, aload_);
                          bytecode_append(current_block^.code, Jmnem(type_list_length(parameters)));
                        end;

                        bytecode_append(current_block^.code, areturn_);
                      end;
                    end;

        integer_type,
        boolean_type,
        char_type:
                    begin
                      bytecode_append(current_block^.code, iload_);
                      bytecode_append(current_block^.code, Jmnem(type_list_length(parameters)));
                      bytecode_append(current_block^.code, ireturn_);
                    end;

        string_type,
        array_type,
        record_type,
        image_type,
        command_type,
        stream_type,
        record_store_type,
        http_type:
                    begin
                      case type_list_length(parameters) of
                        0:  bytecode_append(current_block^.code, aload_0_);
                        1:  bytecode_append(current_block^.code, aload_1_);
                        2:  bytecode_append(current_block^.code, aload_2_);
                        3:  bytecode_append(current_block^.code, aload_3_);
                      else
                        bytecode_append(current_block^.code, aload_);
                        bytecode_append(current_block^.code, Jmnem(type_list_length(parameters)));
                      end;
                      bytecode_append(current_block^.code, areturn_);
                    end;
		end;

		current_block^.parent_block^.children[current_block^.parent_block^.children_count-1] := current_block;
	end;

	result:= forward_declaration;

end;

procedure RD_procedure_declaration(current_block: block);
var
	procedure_name: string;
	procedure_block: block;
	parameters: type_list;
	forward_declaration: integer;
	void_return_type: types;
	procedure_linenum: integer;
  old_inside_implementation: integer;
begin

  //procedure_linenum := linenum;
	old_inside_implementation := inside_implementation_part;
	inside_implementation_part := 0;
	if block_check_name(current_block, TokenStr)=0 then MakeError(MSG_400, TokenStr);

	procedure_name := TokenStr;
	
	if Token <> _IDENT then
		MakeError(MSG_202, '')
	else
	  ReadToken;

	if CompareText(procedure_name, 'run')=0 then
		MakeError(MSG_433, '');
	procedure_block := block_create(current_block, procedure_name);
  parameters := RD_param_list(procedure_block);

	Match(_SEMICOLON);

	void_return_type := type_create();
	void_return_type^.type_class := void_type;

	add_procedure(current_block, procedure_name, parameters, -1, procedure_linenum);

	//if inside_interface_part<>0 then
		//bsf_write_procedure(procedure_name, parameters);

	forward_declaration := RD_proc_block(procedure_block, void_return_type, parameters);

	add_procedure(current_block, procedure_name, parameters, forward_declaration, procedure_linenum);

	Match(_SEMICOLON);

	inside_implementation_part := old_inside_implementation;

	RD_block(current_block);

end;

procedure RD_function_declaration(current_block: block);
var
	function_name: string;
	function_block: block;
	parameters: type_list;
	forward_declaration: integer;
	return_type: types;
	procedure_linenum: integer;
  old_inside_implementation: integer;
  return_value: identifier;
  function_linenum: integer;
begin
  ReadToken;
	old_inside_implementation := inside_implementation_part;
	inside_implementation_part := 0;
	if block_check_name(current_block, TokenStr)=0 then MakeError(MSG_400, TokenStr);

	function_name := TokenStr;

	Match(_IDENT);
	if CompareText(function_name, 'run')=0 then MakeError(MSG_433, '');

	function_block := block_create(current_block,function_name);
	function_block^.next_parameter_index := 0;
	function_block^.next_variable_index := 0;
  parameters := RD_param_list(function_block);
	Inc(function_block^.next_variable_index);

  Match(_COLON);

  return_type := type_from_name(current_block, TokenStr);
	if return_type^.type_class = error_type then MakeError(MSG_406, TokenStr);
	if return_type^.type_class = interval_type then MakeError(MSG_440, '');

  Match(_IDENT);
  Match(_SEMICOLON);

  return_value := identifier_create();
	return_value^.identifier_class := variable_name;
	return_value^.variable_index := type_list_length(parameters);
	return_value^.variable_type := type_duplicate(return_type);
	return_value^.belongs_to_program_block := 0;

  if compiling_unit = 0 then
		initialize_variable(function_block, return_value, '', 1, 'M')
	else
		initialize_variable(function_block, return_value, '', 1, str_program_name);

	add_function(current_block, function_name, parameters, return_type, -1, function_linenum);
	forward_declaration := RD_proc_block(function_block, return_type, parameters);
	add_function(current_block, function_name, parameters,return_type, forward_declaration, function_linenum);

  //if inside_interface_part<>0 then
		//bsf_write_function(function_name, parameters, return_type);

  Match(_SEMICOLON);
  inside_implementation_part := old_inside_implementation;
  RD_block(current_block);

end;

procedure RD_block(current_block: block);
begin

  while Token in [_CONST,_TYPE,_VAR,_FUNCTION,_PROCEDURE,_BEGIN] do begin

    case Token of
      _CONST    :
                begin
                  ReadToken;
                  Check(_IDENT);
                  while Token=_IDENT do ConstDeclaration(current_block);
                end;
      _TYPE     :
                begin
                  ReadToken;
                  Check(_IDENT);
                  while Token=_IDENT do TypeDeclaration(current_block);
                end;
      _VAR      :
                begin
                  ReadToken;
                  Check(_IDENT);
                  while Token=_IDENT do VarDeclaration(current_block);
                end;
      _PROCEDURE:
                begin
                  if (current_block^.parent_block <> root_block) then MakeError(MSG_431, '');
                  RD_procedure_declaration(current_block);
                end;

      _FUNCTION:
                begin
                  if (current_block^.parent_block <> root_block) then MakeError(MSG_431, '');
                  RD_function_declaration(current_block);
                end;

	    _BEGIN    :
                begin
                  ReadToken;
                  RD_block_body(current_block);
                  Match(_END);
                end;
    else
      MakeError(MSG_204, TokenStr);
    end;

  end;

end;

procedure DoPROGRAM(program_block: block; P: string);
var
	random_field_index  : integer;
	random_class_index  : integer;
	random_method_index : integer;
  symbols_filename: string;
begin
  Match(_PROGRAM);
  Match(_IDENT);
  Match(_SEMICOLON);

	compiling_unit := 0;
  str_program_name := RD_program_header();

  if compiling_unit=0 then begin

    program_block := block_create(root_block, str_program_name);

    random_field_index := cp_add_fieldref('M', 'RNG', 'Ljava/util/Random;');
		random_class_index := cp_add_class('java/util/Random');
		random_method_index := cp_add_methodref('java/util/Random', '<init>', '()V');

    bytecode_append(program_block^.code, new_);
		bytecode_append_short_int(program_block^.code, random_class_index);

		bytecode_append(program_block^.code, dup_);

		bytecode_append(program_block^.code, invokespecial_);
		bytecode_append_short_int(program_block^.code, random_method_index);

    bytecode_append(program_block^.code, putstatic_);
		bytecode_append_short_int(program_block^.code, random_field_index);

		bytecode_append(program_block^.code, bipush_);
		bytecode_append(program_block^.code, Jmnem(25));
		bytecode_append(program_block^.code, newarray_);
		bytecode_append(program_block^.code, Jmnem(10));
		bytecode_append(program_block^.code, putstatic_);
		bytecode_append_short_int(program_block^.code, cp_add_fieldref('M', 'IC', '[I'));

		bytecode_append(program_block^.code, iconst_0_);
		bytecode_append(program_block^.code, putstatic_);
		bytecode_append_short_int(program_block^.code, cp_add_fieldref('M', 'KC', 'I'));
		bytecode_append(program_block^.code, iconst_0_);
		bytecode_append(program_block^.code, putstatic_);
		bytecode_append_short_int(program_block^.code, cp_add_fieldref('M', 'KP', 'I'));

    RD_block(program_block);

		bytecode_append(program_block^.code, getstatic_);
		bytecode_append_short_int(program_block^.code, cp_add_fieldref('FW', 'fw', 'LFW;'));
		bytecode_append(program_block^.code, iconst_1_);
		bytecode_append(program_block^.code, invokevirtual_);
		bytecode_append_short_int(program_block^.code, cp_add_methodref('FW', 'destroyApp', '(Z)V'));

		bytecode_append(program_block^.code, return_);

  end else begin

		//int filepos;

    symbols_filename:=ExtractFilePath(P)+Copy(str_program_name,1,Pos('.',str_program_name));

		//symbols_filename = (char*) mem_alloc(strlen(output_path) + string_length(str_program_name) + 10);
		//sprintf(symbols_filename, "%s\\%s.bsf", output_path, string_get_cstr(str_program_name));
		//symbols_file = fopen(symbols_filename, "wb");
		//mem_free(symbols_filename);

		//if (symbols_file == NULL) die(MSG_007);

		program_block := block_create(root_block, str_program_name);

		RD_unit_interface(program_block);
		RD_unit_implementation(program_block);
		RD_unit_initialization(program_block);
		RD_unit_finalization(program_block);

		if (Token <> _END) then
      MakeError(MSG_203, 'end')
		else
		  ReadToken;

		//fclose(symbols_file);

		bytecode_append(program_block^.code, return_);
  end;


  Match(_END);
  Check(_POINT);
end;

procedure Compile(P: string);
var
  program_block : block;
  Scanner       : PScanner;
begin

  Finish:=False;
  New( Scanner, Init);
  Scanner^.OpenFile(P);
  PushScannerItem(Scanner);
  ReadChar;
  AsmMode:=false;

  ReadToken;

  root_block := initialize_root_block();
  program_block := block_create(root_block, P);
  DoPROGRAM(program_block,P);

end;

end.
