unit PasScanner;

interface

uses
  Classes, SysUtils, PasErrors, SPageControl, SMemo;

type
  TToken= (
    _NULL,_TOEOF,_TOEOLN,
    _POINT,_COMMA,_PLUS,_MINUS,_MUL,_AT,_CARET,_SLASH,_INTER,
    _LPAREN,_RPAREN,_SEMICOLON,_ASSIGN,_EQUAL,_GREATER,_LESS,_LESS_EQ,
    _GREATER_EQ,_NOT_EQ,_COLON,_LBRACKET,_RBRACKET,_PERIOD,

    _PROGRAM,_VAR,_CONST,_TYPE,_BEGIN,_WHILE,_DOWNTO,_DO,
    _REPEAT,_UNTIL,_ASM,_END,_DIV,_MOD,_SHL,_SHR,_IN,_IF,
    _THEN,_ELSE,_FOR,_TO,_SET,_PROCEDURE,_FUNCTION,_OBJECT,
    _CONSTRUCTOR,_DESTRUCTOR,_VIRTUAL,_DYNAMIC,_INHERITED,_HALT,
    _OF,_AND,_OR,_XOR,_NIL,_NOT,_ARRAY,_RECORD,_STRTYPE,_PACKED,
    _FILE,_GOTO,_CASE,_WITH,_UNIT,_IMPLEMENTATION,_FORWARD,
    _FINALIZATION,_INITIALIZATION,_TRUE,_FALSE,_INTERFACE,
    _USES,_BREAK,_CONTINUE,_EXIT,_LABEL,_ABSOLUTE,
    _UNKNOWN,_INTEGER,_REAL,_CHAR,_STRING,_BOOLEAN,_NAME,_IDENT
  );

const
  FileScanner_BufSize = 4096;

  Msg: array [_integer.._ident] of string = (
    '  ',
    ' ',
    '',
    ' ',
    '',
    '',
    ' '
  );

  NameToken: array [TToken] of String[14] = (
    '','','',
    '.',',','+','-','*','@','^','/','?',
    '(',')',';',':=','=','>','<','<=',
    '>=','<>',':','[',']','..',

    'PROGRAM','VAR','CONST','TYPE','BEGIN','WHILE','DOWNTO','DO',
    'REPEAT','UNTIL','ASM','END','DIV','MOD','SHL','SHR','IN','IF',
    'THEN','ELSE','FOR','TO','SET','PROCEDURE','FUNCTION','OBJECT',
    'CONSTRUCTOR','DESTRUCTOR','VIRTUAL','DYNAMIC','INHERITED','HALT',
    'OF','AND','OR','XOR','NIL','NOT','ARRAY','RECORD','STRING','PACKED',
    'FILE','GOTO','CASE','WITH','UNIT','IMPLEMENTATION','FORWARD',
    'FINALIZATION','INITIALIZATION','TRUE','FALSE','INTERFACE',
    'USES','BREAK','CONTINUE','EXIT','LABEL','ABSOLUTE',
    '','','','','','','',''
  );

  _EOLN   = #13;
  _EOF    = #26;

type
  pScannerItem = ^tScannerItem;
  tScannerItem = Object
    SrcName  : String;
    Next     : pScannerItem;
    LastRow  : Integer;
    LastCol  : Integer;
    LastChar : Char;
    LastPrev: Integer;
    constructor Init;
    destructor Done;          Virtual; Abstract;
    destructor SelfDestroy;   Virtual; Abstract;
    function   ReadChar:Char; Virtual; Abstract;
    function   NextChar:Char; Virtual; Abstract;
  end;

  //-----------------------------------------

  TBufferInfo = Record
    buf       : pchar;
    Start     : Longint;    // File Postion for the Start of the Buffer
    Size      : Longint;    // Data Size within the Buffer
    Index     : Longint;    // Current Cursor Position
  end;

  TFileInfo = Record
    s         : TMemoryStream;
    Name      : String;
    Size      : Integer;
    MSize     : Integer;
    Closed    : Boolean;
    ClosePos  : Integer;
  end;

  PScanner = ^TScanner;
  TScanner = Object(tScannerItem)
    fi      : TFileInfo;
    bi      : TBufferInfo;
    constructor Init;
    destructor  Done; Virtual;
    destructor  SelfDestroy; Virtual;
    function  OpenFile(fn: String): boolean;
    procedure CloseFile;
    procedure ReloadBuf;
    function  ReadChar:Char; Virtual;
    function  NextChar:Char; Virtual;
  end;

  procedure InitScanner;
  procedure DoneScanner;
  procedure PushScannerItem(Item:pScannerItem);
  procedure PopScannerItem;
  function  ReadChar:Char;
  PROCEDURE SkipSPACE;
  PROCEDURE Readstring;
  PROCEDURE ReadNumber;
  function  ReadToken: tToken;
  procedure Match(X: TToken);
  procedure Check(X: TToken);

Var
  LastChar  : Char;
  NexChar   : Char;
  LastTok   : integer;
  ScanItem  : pScannerItem;
  ModNAME   : string;
  Token,TokE: tToken;
  TokenStr	: ShortString; // Current Token String
  TokenUpStr: ShortString; // UpperCase of the Token String
  TokenInt	: Longint;
  TokenReal : Extended;
  TokenBool : byte;
  TokenRow  : Integer;
  TokenCol  : Integer;
  TokenPrev : Integer;
  ComTerm   : Char;
  AsmMode   : boolean;
  CurStrLen : integer;

implementation

{--------------------TSCANNER---------------------}

Constructor TScanner.Init;
begin
  fi.Closed:=True;
  GetMem( bi.buf,FileScanner_BufSize );
end;

Destructor TScanner.Done;
begin
  CloseFile;
  SetLength(SrcName,0);
  if assigned(bi.buf) then FreeMem(bi.buf,FileScanner_BufSize);
end;

Destructor TScanner.SelfDestroy;
begin
  Done;
  FreeMem( @Self, Sizeof(TScanner) );
end;

function TScanner.OpenFile(fn: String): boolean;
var
  SubPage   : TSPageControl;
  Memo      : TSMemo;
begin
  fi.Closed:=True;
  GetMem( bi.buf,FileScanner_BufSize );
  fi.Closed:=false;
  fi.s:=TMemoryStream.Create;
  
  LastChar:=#0;
  LastRow:=1;
  LastCol:=1;
  LastTok:=0;
  LastPrev:=0;

  fi.s.LoadFromFile(fn);

  fi.s.Position:=0;
  fi.Size:=fi.s.Size;

  bi.Start:=0;
  bi.Size:=0;
  bi.Index:=0;

  SrcName:=fn;
  Result:=true;
end;

procedure TScanner.CloseFile;
begin
 if not(fi.Closed) then begin
    fi.s.Free;
    fi.Closed:=true;
 end;
end;

procedure TScanner.ReloadBuf;
begin
  inc( bi.Start, bi.Size );
  bi.Size:=fi.s.Read(bi.buf^,FileScanner_BufSize);
  bi.Index:=0;
end;

function TScanner.NextChar:Char;
begin
  if bi.Index<bi.Size then
      NextChar:= bi.buf[bi.Index]
  else begin
    ReloadBuf;
    if (bi.Size>0) then begin
        NextChar:=bi.Buf[0];
        bi.Index:=0;
    end else
      NextChar:=_EOF;
  end;
end;

function TScanner.ReadChar: Char;
var
  C : Char;
begin
  LastPrev:=LastCol;
  if (bi.Index<bi.Size) then begin
    C:=bi.buf[bi.Index];
    Inc( bi.Index );
  end else begin
    ReloadBuf;
    if (bi.Size>0) then begin
        C:=bi.Buf[0];
        bi.Index:=1;
    end else
      C:=_EOF;
  end;
  if (C=#13) then begin
    Inc(LastRow);
    LastCol:=0;
    if AsmMode then Token:=_toEOLN;
    TokE:=_toEOLN;
  end else if (C=#10) then begin
    if (LastChar<>#13) then begin
      Inc(LastRow);
      LastCol:=0;
    end;
    TokE:=Token;
  end else
    Inc(LastCol);
  LastChar:=C;
  Result:=LastChar;
end;

////////////////////////////////////////////

constructor tScannerItem.Init;
begin
end;

procedure InitScanner;
Begin
  ScanItem := nil;
End;

procedure DoneScanner;
begin
  while (ScanItem<>NIL) do PopScannerItem;
end;

procedure PushScannerItem(Item:pScannerItem);
begin
  Item^.Next:=ScanItem;
  ScanItem := Item;
end;

procedure PopScannerItem;
var
  Item : pScannerItem;
begin
  if ScanItem=NIL then Exit;
  Item:=ScanItem.Next;
  ScanItem.SelfDestroy;
  ScanItem:=Item;
end;

function ReadChar:Char;
 Var  C:Char;
begin
  if (ScanItem<>NIL) then begin
    C:=ScanItem^.ReadChar;
    while (C=_EOF) do begin
      PopScannerItem;
      if (ScanItem=NIL) then break;
      C:=ReadChar;
    end;
  end else C:=_EOF;
  LastChar:=C;
  ReadChar:=C;
end;

function NextChar:Char;
 Var  C:Char;
begin
  if (ScanItem<>NIL)
    then C:=ScanItem^.NextChar
    else C:=_EOF;
  NextChar:=C;
  NexChar :=C;
end;

procedure SkipSpace;
begin
  if AsmMode then
    while (LastChar in [^I, ' '] ) do ReadChar
  else
    while (LastChar in [#32, #9, #10, #13]) do begin
      ReadChar;
      if (LastChar in [#10, #13]) then LastTok:=0;
    end;
  Inc(LastTok);
end;

procedure Readstring;
var
  St  : String;
  i   : tToken;
Begin
  St:=LastChar;
  while ReadChar in ['0'..'9','A'..'Z','_','a'..'z',''..'',''..'','',''] do begin
     St:=St+LastChar;
  end;
  TokenStr:=St;
  TokenUpStr:=UpperCase(TokenStr);
  Token:=_unknown;
  for i:=_program to _interface do
    if TokenUpStr=NameToken[i] then begin
      Token:=i;
      if Token in [_true,_false] then begin
        Token:=_boolean;
        TokenBool:=ord(Token=_true);
      end;
      Break;
    end;
  if Token = _unknown then Token := _ident;
end;

PROCEDURE ReadNumber;
CONST
  MaxHDig = 8;    (* maximal hexadecimal longint length *)
  MaxRExp = 38;   (* maximal real exponent *)
  MaxLExp = 308;  (* maximal longreal exponent *)
VAR
  i, m, n, d, e: INTEGER;
  dig: ARRAY [0..24] OF CHAR;
  f: DOUBLE;
  expCh: CHAR;
  neg: BOOLEAN;

  FUNCTION Ten(e: INTEGER): DOUBLE;
  VAR
    x, p: DOUBLE;
  BEGIN
    x:=1;
    p:=10;
    WHILE e > 0 DO BEGIN
      IF ODD(e) THEN x:=x*p;
      e:=e DIV 2;
      IF e > 0 THEN p:=p*p;
    END;
    RESULT:= x;
  END;

  FUNCTION FOrd(LastChar: CHAR; hex: BOOLEAN): INTEGER;
  BEGIN
    IF LastChar <= '9' THEN
      RESULT:= ORD(LastChar) - ORD('0')
    ELSE IF hex THEN
      RESULT:= ORD(LastChar) - ORD('A') + 10
    ELSE BEGIN
      MakeError(MSG_025,'');
      RESULT:= 0;
    END
  END;

BEGIN
  Token:=_integer;
  i:=0; m:=0; n:=0; d:=0;
  
  REPEAT

    IF ('0' <= LastChar) AND (LastChar <= '9') OR (d = 0)
    AND ('A' <= LastChar) AND (LastChar <= 'F') THEN BEGIN
      IF (m > 0) OR (LastChar <> '0') THEN BEGIN
        IF n < LENGTH(dig) THEN BEGIN
          dig[n]:=LastChar;
          INC(n);
        END;
        INC(m);
      END;
      ReadChar;
      INC(i);
    END ELSE IF LastChar = '.' THEN BEGIN
      IF NextChar = '.' THEN BEGIN
        EXIT;
      END ELSE IF d = 0 THEN BEGIN
        ReadChar;
        d:=i;
      END ELSE
        MakeError(MSG_025,'');
    END ELSE
      BREAK;

  UNTIL FALSE;

  IF d = 0 THEN BEGIN
    IF n = m THEN BEGIN
      TokenInt:=0;
      i:=0;

      IF LastChar = 'X' THEN BEGIN
        ReadChar;
        Token:=_char;
        IF n <= 2 THEN BEGIN
          WHILE i < n DO BEGIN
            TokenInt:=TokenInt*$10 + FOrd(dig[i], TRUE);
            INC(i);
          END;
        END ELSE
          MakeError(MSG_027,'');
      END ELSE IF LastChar = 'H' THEN BEGIN
        ReadChar;
        Token:=_integer;
        IF n <= MaxHDig THEN BEGIN
          IF (n = MaxHDig) AND (dig[0] > '7') THEN
            MakeError(MSG_025,'');
          WHILE i < n DO BEGIN
            TokenInt:=TokenInt*$10 + FOrd(dig[i], TRUE);
            INC(i);
          END;
        END ELSE
          MakeError(MSG_027,'');
      END ELSE BEGIN
        Token:=_integer;
        WHILE i < n DO BEGIN
          d:=FOrd(dig[i], FALSE);
          INC(i);
          IF TokenInt <= (high(LONGINT) - d) DIV 10 THEN
            TokenInt:=TokenInt*10 + d
          ELSE
            MakeError(MSG_027,'');
        END;
      END;

    END ELSE
      MakeError(MSG_027,'');
  END ELSE BEGIN
    f:=0; e:=0; expCh:='E';
    WHILE n > 0 DO BEGIN
      DEC(n);
      f:=(FOrd(dig[n], FALSE) + f)/10;
    END;
    IF (LastChar = 'E') OR (LastChar = 'D') THEN BEGIN
      expCh:=LastChar;
      ReadChar;
      neg:=FALSE;
      IF LastChar = '-' THEN BEGIN
        neg:=TRUE;
        ReadChar;
      END ELSE IF LastChar = '+' THEN ReadChar;
      IF ('0' <= LastChar) AND (LastChar <= '9') THEN BEGIN
        REPEAT
          n:=FOrd(LastChar, FALSE);
          ReadChar;
          IF e <= (high(INTEGER) - n) DIV 10 THEN
            e:=e*10 + n
          ELSE
            MakeError(MSG_027,'');
        UNTIL (LastChar < '0') OR ('9' < LastChar);
        IF neg THEN e:=-e;
      END ELSE
        MakeError(MSG_025,'');
    END;
    DEC(e, i-d-m);
    IF expCh = 'E' THEN BEGIN
      Token:=_real;
      IF (1-MaxRExp < e) AND (e <= MaxRExp) THEN BEGIN
        IF e < 0 THEN
          TokenReal:=f / Ten(-e)
        ELSE
          TokenReal:=f * Ten(e);
      END ELSE
        MakeError(MSG_027,'');
    END ELSE BEGIN
      Token:=_real;//_longreal;
      IF (1-MaxLExp < e) AND (e <= MaxLExp) THEN BEGIN
        IF e < 0 THEN
          TokenReal:=f / Ten(-e)
          //TokenDReal:=f / Ten(-e)
        ELSE
          TokenReal:=f * Ten(e);
          //TokenDReal:=f * Ten(e);
      END ELSE
        MakeError(MSG_027,'');
    END;
  END;
END;

procedure DoComments;
var
  EndComment: Boolean;
begin
  ReadChar;

  //if look = '$' then DoDirectives;
  EndComment := False;
  repeat
    case LastChar of
      '}' : if ComTerm = '}' then EndComment := True;
      '*' : if NextChar = ComTerm then begin
              ReadChar;
              EndComment := True;
            end;
      '{' : if ComTerm = '}' then DoComments;
      '/' : if (NextChar = '*') and (ComTerm = '/') then begin
              ReadChar;
              DoComments;
            end;
      '(' : if (NextChar = '*') and (ComTerm = ')') then begin
              ReadChar;
              DoComments;
            end;
      _EOF: Exit;
    end;
    //if WasInclude then ProcessInclude;
    ReadChar;
  until EndComment;
end;

function ReadToken: tToken;
var
  code      : integer;
  open      : boolean;
  tmp       : String;
  i         : tToken;

begin{ Readtoken }

  SkipSpace;
  TokenStr := LastChar;

  if LastChar=#26 then begin
    Result:=_toEOF;
    Exit;
  end;

  if (ScanItem<>NIL) then begin
    TokenRow:=ScanItem^.LastRow;
    TokenCol:=ScanItem^.LastCol;
    TokenPrev:=ScanItem^.LastPrev;
  end;

  case LastChar of

      '{'  :
                begin {normal comments}
                  ComTerm := '}';
                  DoComments;
                  ReadToken;
                end;

      '('  :
                if NextChar = '*' then begin
                  ReadChar;
                  ComTerm := ')';
                  DoComments;
                  ReadToken;
                end else begin
                  Token:=_lparen;
                  ReadChar;
                end;

      '/'  :    begin
                  if NextChar= '*' then begin
                    ReadChar;
                    ComTerm := '/';
                    DoComments;
                    ReadToken;
                  end else if NextChar = '/' then begin
                    while not (LastChar in [#10, #13, #26]) do ReadChar;
                    ReadToken;
                  end else begin
                    Token :=  _slash;
                    ReadChar;
                  end;
                end;

      'A'..'Z','_',
      ''..'',
      ''..'',
      '','',
      'a'..'z': ReadString;
      '0'..'9': ReadNumber;

      '#',
      ''''    : begin {string constant}
                  TokenStr := '';
                  Token := _string;
                  CurStrLen:=0;
                  open:=false;

                  repeat
                    case LastChar of
                      #13   : begin MakeWarning(WMSG_001,''); break; end;
                      '#'   :
                              begin
                                ReadChar;
                                tmp:='';
                                while (UpCase(LastChar) in ['$','0'..'9', 'A'..'F'] ) do begin
                                  tmp:=tmp+LastChar;
                                  ReadChar;
                                end;
                                val(tmp, TokenInt, code);
                                Inc(CurStrLen);
                                TokenStr:=TokenStr+chr(TokenInt);
                                open:=false;
                              end;
                      ''''  :
                              begin
                                ReadChar;
                                repeat
                                  case LastChar of
                                    #13   : begin MakeWarning(WMSG_001,''); break; end;
                                    ''''  :
                                            begin
                                              ReadChar;
                                              if LastChar<>'''' then begin
                                                open:=false;
                                                break;
                                              end else begin
                                                Inc(CurStrLen);
                                                TokenStr := TokenStr + LastChar;
                                                ReadChar;
                                                open:=true;
                                              end;
                                            end;
                                    else begin
                                      Inc(CurStrLen);
                                      TokenStr := TokenStr + LastChar;
                                      ReadChar;
                                      open:=true;
                                    end;
                                  end;
                                until false;
                              end;
                    else
                      Inc(CurStrLen);
                      TokenStr := TokenStr + LastChar;
                      ReadChar;
                      open:=true;
                    end;
                  until (not open) and not (LastChar in ['#','''',#26]);
                  if length(TokenStr)=1 then Token:=_char;
                end;

      '$':      begin {hex-number}
                  TokenStr:=LastChar;
                  ReadChar;
                  while (upcase(LastChar) in ['$','0'..'9', 'A'..'F'] ) do begin
                    TokenStr:=TokenStr+LastChar;
                    TokenUpStr:=TokenUpStr+UpCase(LastChar);
                    ReadChar;
                  end;
                  val(TokenStr, TokenInt, code);
                  Token := _integer;
                end;

      ':':
                begin
                  ReadChar;
                  if LastChar='=' then begin
                    TokenStr := TokenStr + LastChar;
                    ReadChar;
                    Token := _assign;
                  end else
                    Token :=  _colon;
                end;

      '.':
                begin
                  ReadChar;
                  if LastChar='.' then begin
                    TokenStr := TokenStr + LastChar;
                    ReadChar;
                    Token :=  _period;
                  end else
                    Token :=  _point;
                end;

      '<':
                begin
                    ReadChar;
                    case LastChar of
                    '=':
                          begin
                            TokenStr := TokenStr + LastChar;
                            ReadChar;
                            Token :=  _less_eq;
                          end;
                    '>':
                          begin
                            TokenStr := TokenStr + LastChar;
                            ReadChar;
                            Token :=  _not_eq;
                          end;
                    else
                      Token :=  _less;
                    end;
                end;


      '>':
                begin
                    ReadChar;
                    if LastChar='=' then begin
                      TokenStr := TokenStr + LastChar;
                      ReadChar;
                      Token := _greater_eq;
                    end else
                      Token := _greater;
                end;

      _EOF :    Token :=  _toEOF;

  else
    Token:=_unknown;
    for i:=_point to _period do
      if LastChar=NameToken[i] then begin
        Token:=i;
        TokenStr := LastChar;
        Break;
      end;
    ReadChar;
  end;
  ReadToken:=Token;
end;


procedure Check(X: TToken);
begin
  if Token<>X then begin
    if X in [_string.._ident] then
      MakeError(MSG_031,Msg[X])
    else
      MakeError(MSG_031,' "'+NameToken[X]+'"');
  end;
end;

procedure Match(X: TToken);
begin
  Check(X);
  ReadToken;
end;

end.

