unit PasTypes;

interface

uses
  SysUtils, PasErrors;

type

  en_type_class = (
    error_type,void_type,integer_type,real_type,char_type,boolean_type,
    string_type,array_type,record_type,interval_type,image_type,
    command_type,stream_type,record_store_type,http_type,alert_type
  );
  
  en_identifier_class = (
    none,program_name,constant_name,variable_name,type_name,
    procedure_name,function_name,parameter_name,unit_name
  );

  string_list             = ^string_list_struct;
  type_list               = ^type_list_struct;
  types                   = ^type_struct;
  identifier              = ^identifier_descriptor_struct;
  name_table              = ^name_table_struct;
  block                   = ^block_struct;
  units                   = ^unit_struct;
  bytecode                = ^bytecode_struct;

  values = record
    Int_Val: integer;
    Real_Val: real;
    Char_Val: char;
    Bool_Val: byte;
    String_Val: string;
  end;

  bytecode_struct = record
    bytecode_pos          : integer;
    bytecode_allocated_size: integer;
    bytecode              : array of char;
  end;

  string_list_struct = record
    next                  : string_list;
    data                  : string;
  end;

  type_list_struct = record
    next                  : type_list;
    data                  : types;
  end;

  type_struct = record
    type_class            : en_type_class;
    element_type          : types;
    dimensions_list       : type_list;

    elements_name_list    : string_list;
    elements_type_list    : type_list;
    unique_record_ID      : integer;

    interval_base_type    : en_type_class;
    first_element         : longint;
    last_element          : longint;
  end;

  unit_struct = record
    names                 : name_table;
    unit_name             : string;
    is_library            : integer;
  end;

  identifier_descriptor_struct = record
    identifier_class      : en_identifier_class ;

    constant_type         : types;
    constant_int_value    : integer;
    constant_real_value   : real;
    constant_string_value : string;

    variable_type         : types;
    variable_index        : integer;

    defined_type          : types;

    return_type           : types;
    parameters            : type_list;
    variables             : type_list;
    forward_declaration,
    standard_function,
    subprogram_linenum,
    unit_function         : integer;
    container_unit        : units;

    parameter_type        : types;
    is_parameter_variable : shortint;
    parameter_index       : integer;

    unit_block            : units;
    belongs_to_program_block: integer;
  end;

  name_table_struct = record
    name                  : string;
    left_child,
    right_child,
    last,next             : name_table;
    descriptor            : identifier;
  end;

  block_struct = record
    parent_block          : block;
    children              : array of block;
    names                 : name_table;
    block_name            : string;
    next_variable_index   : integer;
    next_parameter_index  : integer;
    code                  : bytecode;
    children_count        : integer;
  end;

  Jmnem = (
	  nop_,aconst_null_,iconst_m1_,iconst_0_,iconst_1_,iconst_2_,
    iconst_3_,iconst_4_,iconst_5_,lconst_0_,lconst_1_,fconst_0_,
    fconst_1_,fconst_2_,dconst_0_,dconst_1_,bipush_,sipush_,ldc_,
    ldc_w_,ldc2_w_,iload_,lload_,fload_,dload_,aload_,iload_0_,
    iload_1_,iload_2_,iload_3_,lload_0_,lload_1_,lload_2_,lload_3_,
    fload_0_,fload_1_,fload_2_,fload_3_,dload_0_,dload_1_,dload_2_,
    dload_3_,aload_0_,aload_1_,aload_2_,aload_3_,iaload_,laload_,
    faload_,daload_,aaload_,baload_,caload_,saload_,istore_,lstore_,
    fstore_,dstore_,astore_,istore_0_,istore_1_,istore_2_,istore_3_,
    lstore_0_,lstore_1_,lstore_2_,lstore_3_,fstore_0_,fstore_1_,
    fstore_2_,fstore_3_,dstore_0_,dstore_1_,dstore_2_,dstore_3_,
    astore_0_,astore_1_,astore_2_,astore_3_,iastore_,lastore_,fastore_,
    dastore_,aastore_,bastore_,castore_,sastore_,pop_,pop2_,dup_,dup_x1_,
    dup_x2_,dup2_,dup2_x1_,dup2_x2_,swap_,iadd_,ladd_,fadd_,dadd_,isub_,
    lsub_,fsub_,dsub_,imul_,lmul_,fmul_,dmul_,idiv_,ldiv_,fdiv_,ddiv_,
    irem_,lrem_,frem_,drem_,ineg_,lneg_,fneg_,dneg_,ishl_,lshl_,ishr_,
    lshr_,iushr_,lushr_,iand_,land_,ior_,lor_,ixor_,lxor_,iinc_,i2l_,
    i2f_,i2d_,l2i_,l2f_,l2d_,f2i_,f2l_,f2d_,d2i_,d2l_,d2f_,i2b_,i2c_,
    i2s_,lcmp_,fcmpl_,fcmpg_,dcmpl_,dcmpg_,ifeq_,ifne_,iflt_,ifge_,ifgt_,
    ifle_,if_icmpeq_,if_icmpne_,if_icmplt_,if_icmpge_,if_icmpgt_,if_icmple_,
    if_acmpeq_,if_acmpne_,goto_,jsr_,ret_,tableswitch_,lookupswitch_,
    ireturn_,lreturn_,freturn_,dreturn_,areturn_,return_,getstatic_,
    putstatic_,getfield_,putfield_,invokevirtual_,invokespecial_,
    invokestatic_,invokeinterface_,xxxunusedxxx1_,new_,newarray_,
    anewarray_,arraylength_,athrow_,checkcast_,instanceof_,monitorenter_,
    monitorexit_,wide_,multianewarray_,ifnull_,ifnonnull_,goto_w_,jsr_w_,

    nop1 ,nop2 ,nop3 ,nop4 ,nop5 ,nop6 ,nop7 ,nop8 ,nop9 ,nop10,nop11,nop12,
    nop13,nop14,nop15,nop16,nop17,nop18,nop19,nop20,nop21,nop22,nop23,nop24,
    nop25,nop26,nop27,nop28,nop29,nop30,nop31,nop32,nop33,nop34,nop35,nop36,
    nop37,nop38,nop39,nop40,nop41,nop42,nop43,nop44,nop45,nop46,nop47,

    break_stmt_
  );

var
  root_block            : block;
  compiling_unit        : integer = 0;
  inside_interface_part : integer = 0;
  inside_implementation_part: integer = 0;
  str_program_name      : string;
  next_record_ID        : integer;
  mathType              : integer = 1;
  usesFloat             : integer;
  usesHttp              : integer;
  usesRecordStore       : integer;
  output_path           : string;
  inside_loop           : integer = 0;

  procedure add_real_constant(item: block; value: real; name: string);
  procedure add_integer_constant(item: block; value: longint; name: string);
  procedure add_char_constant(item: block; value: char; name: string);
  procedure add_boolean_constant(item: block; value: byte; name: string);
  procedure add_string_constant(item: block; value,name: string);
  function  type_duplicate(item: types): types;
  procedure Init(item: block);
  function  initialize_root_block(): block;
  function  name_table_create(): name_table;
  function  get_identifier(current_block: block; cstr: string): identifier;
  function  name_table_find(item: name_table; name: string): identifier;
  function  block_check_name(item: block; cstr: string): integer;
  function  string_list_create(): string_list;
  procedure string_list_append(item: string_list; data: string);
  function  type_create(): types;
  function  type_list_create(): type_list;
  procedure type_list_append(item: type_list; data: types);
  function  type_from_name(current_block: block; cstr: string): types;
  function  get_constant_identifier(current_block: block; cstr: string): identifier;
  function  get_constant_type(current_block: block; cstr: string): types;
  procedure type_add_record(item: types; element_names: string_list; element_type: types);
  procedure add_variables(item: block; identifier_list: string_list; var_type: types);
  procedure get_field_descriptor(field_type: types; descriptor: Pchar);
  function  type_list_length(item: type_list): integer;
  function  identifier_create(): identifier;
  procedure add_type(item: block; name: string; type_type: types);
  function  string_list_length(item: string_list): integer;
  procedure add_parameters(item: block; parameters_list: string_list;
					      parameters_type: types; is_parameter_variable: integer);
  procedure add_procedure(item: block; name: string; parameters: type_list;
				      forward_declaration,linenum: integer);
  procedure add_function(item: block; name: string; parameters: type_list;
              return_type: types; forward_declaration,linenum: integer);
  function type_list_different_parameter(list1,list2: type_list): integer;
  function type_equal_cast(type1,type2: types): integer;
  function type_find_record_element(records: types; cstr: string): types;
  procedure replace_aaload_instruction(code: bytecode; element_type: types);
  function type_list_different_parameter_array(real_types: type_list; interval_types: type_list): integer;
  function type_list_different_parameter_cast(list1,list2: type_list): integer;
  function type_get_name(basic_type: types): string;
  function type_equal(type1,type2: types): integer;
  procedure transform_break_stmts(code: bytecode; start_offset,end_offset,jump_pos: integer);
  function get_variable_type(current_block: block; name: string): types;
  
implementation

uses
  PasCodegen;

function identifier_create(): identifier;
var
  new_identifier: identifier;
begin
	New(new_identifier);
	new_identifier^.identifier_class := none;
	new_identifier^.standard_function := 0;
	new_identifier^.variables := NIL;
	new_identifier^.forward_declaration := 0;
	new_identifier^.unit_block := NIL;
	new_identifier^.unit_function := 0;
	new_identifier^.container_unit := NIL;
	result:= new_identifier;
end;

function string_list_create(): string_list;
var
	new_list: string_list;
begin
  New(new_list);
	new_list^.data := '';
	new_list^.next := NIL;
	result:= new_list;
end;

procedure string_list_append(item: string_list; data: string);
var
	new_element: string_list;
begin
	if item^.data = '' then
		item^.data := data
	else begin
		New(new_element);
		new_element^.data := data;
		new_element^.next := NIL;
		while item^.next <> NIL do item := item^.next;
		item^.next := new_element;
  end;
end;

function string_list_duplicate(item: string_list): string_list;
var
	new_list: string_list;
begin
	new_list := string_list_create();
	if item^.data = '' then begin result:=new_list; exit; end;
  while (item<>NIL) and (item^.data='') do begin
		string_list_append(new_list, item^.data);
		item := item^.next;
  end;
	result:= new_list;
end;

function string_list_length(item: string_list): integer;
var
  counter: integer;
begin
	counter := 0;
	if item^.data = '' then begin Result:= counter; exit; end;
	repeat
		item := item^.next;
		Inc(counter);
	until item = NIL;
	Result:= counter;
end;

function type_list_create(): type_list;
var
  new_list: type_list;
begin
	New(new_list);
	new_list^.data := NIL;
	new_list^.next := NIL;
	Result:= new_list;
end;

procedure type_list_append(item: type_list; data: types);
var
	new_element: type_list;
begin
	if item^.data=NIL then
		item^.data := type_duplicate(data)
	else begin
		New(new_element);
		new_element^.data := type_duplicate(data);
		new_element^.next := NIL;
		while item^.next<>NIL do item := item^.next;
		item^.next:=new_element;
  end;
end;

function type_list_duplicate(item: type_list): type_list;
var
	new_list: type_list;
begin
	if item = NIL then begin result:=NIL; Exit; end;
	new_list := type_list_create();
	if (item = NIL) or (item^.data = NIL) then begin result:=new_list; Exit; end;
	repeat
		type_list_append(new_list, item^.data);
		item := item^.next;
	until (item = NIL);
	result := new_list;
end;

function type_from_name(current_block: block; cstr: string): types;
var
	found_type: types;
	name: string;
	declaration: identifier;
begin
	found_type := type_create();
	found_type^.type_class := error_type;
	name := cstr;
	while current_block <> NIL do begin
		declaration := name_table_find(current_block^.names, name);
		if declaration <> NIL then begin
			if declaration^.identifier_class = type_name then
				found_type := type_duplicate(declaration^.defined_type);
			break;
		end;
		current_block := current_block^.parent_block;
	end;
	Result:= found_type;
end;

function type_create(): types;
var
  new_type: types;
begin
	New(new_type);
	new_type^.type_class := void_type;
	result:= new_type;
end;

function type_duplicate(item: types): types;
var
	new_item: types;
begin
	new_item:=item;
	case new_item^.type_class of
    array_type:
          begin
            new_item^.dimensions_list     := type_list_duplicate(item^.dimensions_list);
            new_item^.element_type        := type_duplicate(item^.element_type);
          end;
    record_type:
          begin
            new_item^.elements_type_list  := type_list_duplicate(item^.elements_type_list);
            new_item^.elements_name_list  := string_list_duplicate(item^.elements_name_list);
            new_item^.unique_record_ID    := item^.unique_record_ID;
          end;
  end;
	result:= new_item;
end;

function name_table_create(): name_table;
var
  new_item: name_table;
begin
	New(new_item);
	new_item^.left_child := NIL;
	new_item^.right_child := NIL;
	new_item^.name := '';
	new_item^.descriptor := NIL;
	new_item^.last := new_item;
	new_item^.next := NIL;
	result := new_item;
end;

procedure name_table_insert(root: name_table; name: string; descriptor: identifier);
var
  string_compare: integer;
	item,new_item :name_table;
begin
	if root=NIL then Exit;
	if root^.name='' then begin
		root^.name := name;
		root^.descriptor := descriptor;
		Exit;
	end;
	item := root;
	repeat
		string_compare := CompareText(name, item^.name);
		if string_compare=0 then exit;
		if string_compare<0 then begin
			if item^.left_child=NIL then break;
			item := item^.left_child;
		end;
		if string_compare>0 then begin
			if item^.right_child=NIL then break;
			item := item^.right_child;
		end;
	until false;
	new_item := name_table_create();
	new_item^.name := name;
	new_item^.descriptor := descriptor;
	new_item^.next := NIL;
	root^.last^.next := new_item;
	root^.last := new_item;
	if string_compare<0 then item^.left_child := new_item;
	if string_compare>0 then item^.right_child := new_item;
end;

procedure add_std_type(item: block; type_class:en_type_class; name: string);
var
	descriptor: identifier;
begin
	descriptor := identifier_create();
	descriptor^.identifier_class := type_name;
	descriptor^.defined_type := type_create();
	descriptor^.defined_type^.type_class := type_class;
	name_table_insert(item^.names, name, descriptor);
end;

procedure add_real_constant(item: block; value: real; name: string);
var
	descriptor: identifier;
begin
	descriptor := identifier_create();
	descriptor^.identifier_class := constant_name;
	descriptor^.constant_type := type_create();
	descriptor^.constant_type^.type_class := real_type;
	descriptor^.constant_real_value := value;
	name_table_insert(item^.names, name, descriptor);
end;

procedure add_integer_constant(item: block; value: longint; name: string);
var
	descriptor: identifier;
begin
	descriptor := identifier_create();
	descriptor^.identifier_class := constant_name;
	descriptor^.constant_type := type_create();
	descriptor^.constant_type^.type_class := integer_type;
	descriptor^.constant_int_value := value;
	name_table_insert(item^.names, name, descriptor);
end;

procedure add_char_constant(item: block; value: char; name: string);
var
	descriptor: identifier;
begin
	descriptor := identifier_create();
	descriptor^.identifier_class := constant_name;
	descriptor^.constant_type := type_create();
	descriptor^.constant_type^.type_class := char_type;
	descriptor^.constant_int_value := ord(value);
	name_table_insert(item^.names, name, descriptor);
end;

procedure add_boolean_constant(item: block; value: byte; name: string);
var
	descriptor: identifier;
begin
	descriptor := identifier_create();
	descriptor^.identifier_class := constant_name;
	descriptor^.constant_type := type_create();
	descriptor^.constant_type^.type_class := boolean_type;
	descriptor^.constant_int_value := ord(value);
	name_table_insert(item^.names, name, descriptor);
end;

procedure add_string_constant(item: block; value,name: string);
var
	descriptor: identifier;
begin
	descriptor := identifier_create();
	descriptor^.identifier_class := constant_name;
	descriptor^.constant_type := type_create();
	descriptor^.constant_type^.type_class := string_type;
	descriptor^.constant_string_value := value;
	name_table_insert(item^.names, name, descriptor);
end;

procedure add_std_procedure(item: block; cstr_name: string; params: type_list; return_type: types);
var
	name: string;
	descriptor: identifier;
begin
	name := cstr_name;
	descriptor := identifier_create();
	descriptor^.identifier_class := procedure_name;
  descriptor^.return_type := return_type;
	descriptor^.parameters := params;
	descriptor^.variables := NIL;
	descriptor^.standard_function := 1;
	name_table_insert(item^.names, name, descriptor);
end;

procedure add_special_function(item: block; parameter_en: array of en_type_class;
                      return_type_en: en_type_class; name: string);
var
	params: type_list;
	param: types;
	return_type: types;
  i: integer;
begin
	params := type_list_create();
	param := type_create();
	return_type := type_create();
	param^.type_class := parameter_en[0];
	return_type^.type_class := return_type_en;
	if length(parameter_en)=1 then begin
    if parameter_en[0] <> void_type then type_list_append(params, param);
  end else
    type_list_append(params, param);

  for i:=1 to length(parameter_en)-1 do begin
    param^.type_class := parameter_en[i];
	  type_list_append(params, param);
  end;

	if return_type_en=void_type then
    add_std_procedure(item, name, params, NIL)
	else
		add_std_procedure(item, name, params, return_type);
    
end;

//######################################################

function name_table_find(item: name_table; name: string): identifier;
var
	string_compare: integer;
begin
  result:=NIL;
	while (item <> NIL) and (item^.name <> '') do begin
		string_compare := CompareText(name, item^.name);
		if (string_compare = 0) then begin result:= item^.descriptor; break; end;
		if (string_compare < 0) then item := item^.left_child;
		if (string_compare > 0) then item := item^.right_child;
	end;
end;

function block_check_name(item: block; cstr: string): integer;
var
	return_value: integer;
	name: string;
	descriptor: identifier;
begin
  return_value := 1;
	name := cstr;
	
	descriptor := name_table_find(item^.names, name);

	if descriptor <> NIL then begin
		if (((descriptor^.identifier_class = procedure_name)
			    or (descriptor^.identifier_class = function_name))
			and (descriptor^.forward_declaration = 1)) then
			return_value := 1
		else
			return_value := 0;
	end;

	if (return_value = 1) then begin
		while (item^.parent_block <> NIL) do item := item^.parent_block;
		descriptor := name_table_find(item^.names, name);
		if ((descriptor <> NIL) and (descriptor^.identifier_class = type_name)) then
			return_value := 0;
	end;

	Result:= return_value;
end;

function unit_duplicate(item: units): units;
var
	new_item: units;
begin
	New(new_item);
	new_item^.names := item^.names;
	new_item^.unit_name := item^.unit_name;
	new_item^.is_library := item^.is_library;
	result:= new_item;
end;

function identifier_duplicate(item: identifier): identifier;
var
	new_identifier: identifier;
begin
	New(new_identifier);
	new_identifier^.identifier_class := item^.identifier_class;
	new_identifier^.unit_function := item^.unit_function;

	case new_identifier^.identifier_class of

    constant_name:
      begin
        new_identifier^.constant_type := type_duplicate(item^.constant_type);
        case item^.constant_type^.type_class of
          integer_type: new_identifier^.constant_int_value := item^.constant_int_value;
          char_type:    new_identifier^.constant_int_value := item^.constant_int_value;
          string_type:  new_identifier^.constant_string_value := item^.constant_string_value;
          boolean_type: new_identifier^.constant_int_value := item^.constant_int_value;
          real_type:    new_identifier^.constant_real_value := item^.constant_real_value;
          command_type,
          record_store_type,
          http_type,
          image_type,
          stream_type,
          alert_type: ;
        else
          MakeError(MSG_011,'');
        end;
      end;

    variable_name:
          begin
            new_identifier^.variable_type := type_duplicate(item^.variable_type);
            new_identifier^.variable_index := item^.variable_index;
          end;

    type_name:  new_identifier^.defined_type := type_duplicate(item^.defined_type);

    procedure_name:
          begin
            new_identifier^.parameters := type_list_duplicate(item^.parameters);
            new_identifier^.forward_declaration := item^.forward_declaration;
            new_identifier^.standard_function := item^.standard_function;
            new_identifier^.variables := type_list_duplicate(item^.variables);
            if item^.container_unit <> NIL then
              new_identifier^.container_unit := unit_duplicate(item^.container_unit)
            else
              new_identifier^.container_unit := NIL;
          end;

    function_name:
          begin
            new_identifier^.parameters := type_list_duplicate(item^.parameters);
            new_identifier^.return_type := type_duplicate(item^.return_type);
            new_identifier^.forward_declaration := item^.forward_declaration;
            new_identifier^.standard_function := item^.standard_function;
            new_identifier^.variables := type_list_duplicate(item^.variables);
          end;

    parameter_name:
      begin
        new_identifier^.parameter_type := type_duplicate(item^.parameter_type);
        new_identifier^.is_parameter_variable := item^.is_parameter_variable;
        new_identifier^.parameter_index := item^.parameter_index;
      end;

    unit_name: new_identifier^.unit_block := unit_duplicate(item^.unit_block);

	end;

	if item^.container_unit <> NIL then
		new_identifier^.container_unit := unit_duplicate(item^.container_unit)
	else
		new_identifier^.container_unit := NIL;

	if item^.unit_block <> NIL then
		new_identifier^.unit_block := unit_duplicate(item^.unit_block)
	else
		new_identifier^.unit_block := NIL;

	result:= new_identifier;

end;

function get_identifier(current_block: block; cstr: string): identifier;
var
	name: string;
	found_identifier: identifier;
	declaration: identifier;
	name_table_it: name_table;
  descriptor: identifier;
  unit_table_it: name_table;
begin
	name := cstr;
  found_identifier := NIL;

	while current_block <> NIL do begin
		declaration := name_table_find(current_block^.names, name);
		if declaration <> NIL then begin
			found_identifier := identifier_duplicate(declaration);
			if current_block^.parent_block = root_block then
				found_identifier^.belongs_to_program_block := 1
			else
				found_identifier^.belongs_to_program_block := 0;
			break;
		end;
		current_block := current_block^.parent_block;
	end;

	if found_identifier <> NIL then begin result:= found_identifier; exit; end;

	if root_block^.names^.name <> '' then begin
		name_table_it := root_block^.names;
		while name_table_it <> NIL do begin
			descriptor := name_table_it^.descriptor;
			if descriptor^.identifier_class = unit_name then begin
				unit_table_it := descriptor^.unit_block^.names;
				if unit_table_it^.name <> '' then begin
					while unit_table_it <> NIL do begin
						if CompareText(unit_table_it^.name,cstr)=0 then begin
							if found_identifier = NIL then
								found_identifier := identifier_duplicate(unit_table_it^.descriptor)
							else
								MakeError(MSG_459, cstr);
						end;
						unit_table_it := unit_table_it^.next;
					end;
				end;
			end;
			name_table_it := name_table_it^.next;
    end;
  end;

	if found_identifier = NIL then begin
		found_identifier := identifier_create();
		found_identifier^.belongs_to_program_block := 0;
	end;

	result:= found_identifier;
end;

function get_constant_identifier(current_block: block; cstr: string): identifier;
var
	name: string;
	found_constant: identifier;
	declaration: identifier;
begin
	name := cstr;
  found_constant := NIL;
	while current_block <> NIL do begin
		declaration := name_table_find(current_block^.names, name);
		if declaration <> NIL then begin
			if declaration^.identifier_class = constant_name then
				found_constant := identifier_duplicate(declaration);
			break;
		end;
		current_block := current_block^.parent_block;
	end;
	if found_constant = NIL then found_constant := identifier_create();
	result:= found_constant;
end;

function get_constant_type(current_block: block; cstr: string): types;
var
	found_type: types;
	name: string;
	declaration: identifier;
begin
	found_type := type_create();
	found_type^.type_class := error_type;
	name := cstr;
	while current_block <> NIL do begin
		declaration := name_table_find(current_block^.names, name);
		if declaration <> NIL then begin
			if (declaration^.identifier_class = constant_name) then
				found_type := type_duplicate(declaration^.constant_type);
			break;
		end;
		current_block := current_block^.parent_block;
	end;
	result:= found_type;
end;

procedure type_record_check_duplicate_names(item: types; element_names: string_list);
var
	item_list: string_list;
begin
	while element_names <> NIL do begin
		if element_names^.data <> '' then begin
			item_list := item^.elements_name_list;
			while item_list <> NIL do begin
				if item_list^.data <> '' then begin
					if (CompareText(item_list^.data,element_names^.data) = 0) then
						MakeError(MSG_418, item_list^.data);
				end;
				item_list := item_list^.next;
			end;
		end;
		element_names := element_names^.next;
  end;
end;

procedure type_add_record(item: types; element_names: string_list; element_type: types);
begin
	type_record_check_duplicate_names(item, element_names);
	while element_names <> NIL do begin
		if element_names^.data <> '' then begin
			string_list_append(item^.elements_name_list, element_names^.data);
			type_list_append(item^.elements_type_list, element_type);
		end;
		element_names := element_names^.next;
  end;
end;

procedure add_type(item: block; name: string; type_type: types);
var
	descriptor: identifier;
begin
	descriptor := identifier_create();
	descriptor^.identifier_class := type_name;
	descriptor^.defined_type := type_type;
	name_table_insert(item^.names, name, descriptor);
end;

procedure add_variables(item: block; identifier_list: string_list; var_type: types);
var
	it: string_list;
	descriptor: identifier;
	block_identifier: identifier;
begin
	it := identifier_list;
	if it^.data = '' then Exit;
	if item^.parent_block <> root_block then begin
		block_identifier := name_table_find(item^.parent_block^.names, item^.block_name);
		if block_identifier^.variables = NIL then
			block_identifier^.variables := type_list_create();
	end	else
		block_identifier := NIL;

	while (it <> NIL) do begin

		descriptor := identifier_create();
		descriptor^.identifier_class := variable_name;
		descriptor^.variable_type := type_duplicate(var_type);
		descriptor^.variable_index := item^.next_variable_index;
		descriptor^.belongs_to_program_block := ord(item^.parent_block = root_block);
		name_table_insert(item^.names, it^.data, descriptor);

		if compiling_unit = 0 then
			initialize_variable(item, descriptor, it^.data, 1, 'M')
		else
			initialize_variable(item, descriptor, it^.data, 1, str_program_name);

		if block_identifier <> NIL then
			type_list_append(block_identifier^.variables, descriptor^.variable_type);

		Inc(item^.next_variable_index);

		it := it^.next;
	end;

end;

function type_list_length(item: type_list): integer;
var
  counter: integer;
begin
	counter := 0;
	if item^.data = NIL then begin Result:=0; Exit; end;
	repeat
		item := item^.next;
		Inc(counter);
	until item = NIL;
	Result:= counter;
end;

procedure get_field_descriptor(field_type: types; descriptor: Pchar);
var
  element_type: array [0..128] of char;
  i, len      : integer;

  procedure sprintf(var D: pchar; S: string);
  begin
    Move(S,D[0],Length(S));
  end;

begin

	case (field_type^.type_class) of

    void_type:          sprintf(descriptor,   'V');

    integer_type,
    char_type,
    boolean_type:       sprintf(descriptor,   'I');

    real_type:
                  if (mathType = 1) then
                        sprintf(descriptor,   'I')
                  else
                        sprintf(descriptor,   'LReal;');

    string_type:        sprintf(descriptor,   'Ljava/lang/String;');
    command_type:       sprintf(descriptor,   'Ljavax/microedition/lcdui/Command;');
    stream_type:	      sprintf(descriptor,   'Ljava/io/InputStream;');
    record_store_type: 	sprintf(descriptor,   'Ljavax/microedition/rms/RecordStore;');
    http_type: 		      sprintf(descriptor,   'LH;');
    alert_type:		      sprintf(descriptor,   'Ljavax/microedition/lcdui/AlertType;');
    image_type:		      sprintf(descriptor,   'Ljavax/microedition/lcdui/Image;');

    array_type:
                  begin
                    get_field_descriptor(field_type^.element_type, element_type);
                    len := type_list_length(field_type^.dimensions_list);
                    for i:=1 to len do descriptor[i] := '[';
                    descriptor[len] := #0;
                    Move(element_type[0],descriptor[len],128);
                  end;


    record_type:        sprintf(descriptor,    'LR_'+IntToStr(field_type^.unique_record_ID));
    error_type: ;

	else
		MakeError(MSG_013,'');
  end;

end;

procedure add_parameters(item: block; parameters_list: string_list;
					parameters_type: types; is_parameter_variable: integer);
var
	it: string_list;
	descriptor: identifier;
begin
	it := parameters_list;

	if it^.data = '' then Exit;

	while (it <> NIL) do begin
		descriptor := identifier_create();
		descriptor^.identifier_class := parameter_name;
		descriptor^.parameter_index := item^.next_parameter_index;
		Inc(item^.next_parameter_index);
		descriptor^.parameter_type := type_duplicate(parameters_type);
		descriptor^.is_parameter_variable := is_parameter_variable;
		name_table_insert(item^.names, it^.data, descriptor);
		it := it^.next;
  end;

	item^.next_variable_index := item^.next_parameter_index;

end;

function type_equal(type1,type2: types): integer;
begin

	if (type1 = type2) then begin result:= 1; exit; end;
	if ((type1 = NIL) or (type2 = NIL)) then begin result:= 0; exit; end;
	if (type1^.type_class <> type2^.type_class) then begin result:= 0; exit; end;

	case type1^.type_class of
	array_type:
		begin
			if (type_equal(type1^.element_type, type2^.element_type)=0) then begin result:= 0; exit; end;
			if (type_list_different_parameter(type1^.dimensions_list,type2^.dimensions_list) <> 0) then begin result:= 0; exit; end;
    end;

	record_type: result:=ord(type1^.unique_record_ID = type2^.unique_record_ID);

	interval_type:
		begin
			if (type1^.interval_base_type <> type2^.interval_base_type) then begin result:= 0; exit; end;
			if (type1^.first_element <> type2^.first_element) then begin result:= 0; exit; end;
			if (type1^.last_element <> type2^.last_element) then begin result:= 0; exit; end;
		end;
	end;

	Result:= 1;
end;

function type_list_different_parameter(list1,list2: type_list): integer;
var
	counter: integer;
begin

  counter := 1;
  Result:= 0;

	while list1 <> NIL do begin
		if ((list2 = NIL)
    or (list1^.data = NIL) or (list2^.data = NIL)
    or (type_equal(list1^.data, list2^.data)=0)) then begin
			if ((list2 <> NIL)
      and (list1^.data = NIL)
      and (list2^.data = NIL)) then Exit;
			Result:= counter;
      Exit;
		end;
		Inc(counter);
		list1 := list1^.next;
		list2 := list2^.next;
	end;
	if (list2 <> NIL) then Result:= -1;

end;

procedure check_forward_declaration(declaration: identifier;
							   parameters: type_list;
							   forward_declaration: integer;
							   return_type: types);
var
	different_parameter: integer;
begin

	if ((declaration^.identifier_class <> procedure_name)
		and (declaration^.identifier_class <> function_name)) then exit;

	if forward_declaration<>0 then begin
		MakeError(MSG_401, '');
		Exit;
  end;

	if ((return_type = NIL)
  and (declaration^.identifier_class = function_name)) then
		MakeError(MSG_402, 'function');

	if ((return_type <> NIL)
  and (declaration^.identifier_class = procedure_name)) then
		MakeError(MSG_402, 'procedure');

	different_parameter := type_list_different_parameter(parameters, declaration^.parameters);
	if (different_parameter = -1) then MakeError(MSG_403, '');
	if (different_parameter > 0) then MakeError(MSG_404, '');

	if (return_type <> NIL) then
		if ((declaration^.identifier_class <> function_name) or (type_equal(return_type, declaration^.return_type)=0)) then
			MakeError(MSG_405, '');

end;

procedure add_procedure(item: block; name: string; parameters: type_list;
				   forward_declaration,linenum: integer);
var
	descriptor: identifier;
begin
	descriptor := name_table_find(item^.names, name);

	if descriptor <> NIL then begin
		if ((forward_declaration <> -1)
    and (descriptor^.forward_declaration <> -1)) then
			check_forward_declaration(descriptor, parameters,forward_declaration, NIL);

		if ((descriptor^.forward_declaration = -1)
    or (forward_declaration <> -1)) then
			descriptor^.forward_declaration := forward_declaration;
	end else begin
		descriptor := identifier_create();
		descriptor^.identifier_class := procedure_name;
		descriptor^.parameters := parameters;
		descriptor^.forward_declaration := forward_declaration;
		descriptor^.subprogram_linenum := linenum;
		name_table_insert(item^.names, name, descriptor);
  end;
end;

procedure add_function(item: block; name: string; parameters: type_list;
				              return_type: types; forward_declaration,linenum: integer);
var
	descriptor: identifier;
begin
	descriptor := name_table_find(item^.names, name);

	if descriptor <> NIL then begin
		if ((forward_declaration <> -1)
    and (descriptor^.forward_declaration <> -1)) then
			check_forward_declaration(descriptor, parameters,forward_declaration, return_type);

		if ((descriptor^.forward_declaration = -1)
    or (forward_declaration <> -1)) then
			descriptor^.forward_declaration := forward_declaration;
	end else begin
		descriptor := identifier_create();
		descriptor^.identifier_class := function_name;
		descriptor^.parameters := parameters;
		descriptor^.forward_declaration := forward_declaration;
		descriptor^.subprogram_linenum := linenum;
		name_table_insert(item^.names, name, descriptor);
  end;
end;

function type_equal_cast(type1,type2: types): integer;
begin
  Result:= 0;
	if type_equal(type1, type2)<>0 then Result:= 1;
	if (type1^.type_class = real_type)
		and (type2^.type_class = integer_type) then Result:= 2;
	if ((type1^.type_class = integer_type)
		and (type2^.type_class = real_type)) then Result:= 3;
	if ((type1^.type_class = string_type)
		and (type2^.type_class = char_type)) then Result:= 2;
	if ((type1^.type_class = char_type)
		and (type2^.type_class = string_type)) then Result:= 3;
end;

function type_find_record_element(records: types; cstr: string): types;
var
	names: string_list;
	types: type_list;
begin
	if (records^.type_class <> record_type) then begin Result:= NIL; exit; end;

	names := records^.elements_name_list;
	types := records^.elements_type_list;

	while (names <> NIL) do begin
		if (names^.data <> '') then
			if CompareText(cstr, names^.data)=0 then begin
				Result:= types^.data;
        Exit;
      end;
	
		names := names^.next;
		types := types^.next;
  end;

	Result:= NIL;
end;

procedure replace_aaload_instruction(code: bytecode; element_type: types);
begin
	Dec(code^.bytecode_pos);

	case element_type^.type_class of

    integer_type,
    boolean_type,
    char_type: bytecode_append(code, iaload_);

    real_type:
      if (mathType = 1) then
        bytecode_append(code, iaload_)
      else
        bytecode_append(code, aaload_);

    record_type,
    string_type,
    image_type,
    command_type,
    stream_type,
    record_store_type,
    http_type,
    alert_type: bytecode_append(code, aaload_);

	else
		MakeError(MSG_024,'');
  end;
end;

function type_list_different_parameter_array(real_types: type_list; interval_types: type_list): integer;
var
	counter: integer;
begin
  counter :=1;
  result  :=0;
	while (real_types <> NIL) do begin
		if ((interval_types = NIL)
			or (real_types^.data = NIL)
			or (interval_types^.data = NIL)
			or (real_types^.data^.type_class <> interval_types^.data^.interval_base_type)) then begin
			Result:= counter;
      Exit;
    end;
		Inc(counter);
		real_types := real_types^.next;
		interval_types := interval_types^.next;
  end;
	
	if interval_types <> NIL then result:=-1;
end;

function type_list_different_parameter_cast(list1,list2: type_list): integer;
var
	counter: integer;
begin

  counter := 1;
  Result:= 0;
	while (list1 <> NIL) do begin
		if ((list2 = NIL)
			or (list1^.data = NIL)
			or (list2^.data = NIL)
			or (type_equal_cast(list1^.data, list2^.data) = 0)
			or (type_equal_cast(list1^.data, list2^.data) = 2)) then begin
			if ((list2 <> NIL)
				and (list1^.data = NIL)
				and (list2^.data = NIL)) then begin
				Result:= 0;
        Exit;
      end;
			Result:= counter;
      Exit;
		end;

		Inc(counter);
		list1 := list1^.next;
		list2 := list2^.next;
  end;
	
	if (list2 <> NIL) then Result:= -1;
end;

procedure transform_break_stmts(code: bytecode; start_offset,end_offset,jump_pos: integer);
var
	jump_offset,pos: integer;
begin
  pos := start_offset;
	while (pos < end_offset) do begin
		if ((integer(code^.bytecode[pos]) = ord(break_stmt_))
			and ((pos + 2) < end_offset)) then begin
			if ((code^.bytecode[pos+1] = #0)
				and (code^.bytecode[pos+2] = #0)) then begin
				jump_offset := jump_pos - pos;
				code^.bytecode[pos] := char(goto_);
				code^.bytecode[pos+1] := char(jump_offset shr 8);
				code^.bytecode[pos+2] := char(jump_offset);
			end;
		end;
		Inc(pos);
	end;
end;

function get_variable_type(current_block: block; name: string): types;
var
	found_type: types;
	declaration: identifier;
begin
	found_type := type_create();
	found_type^.type_class := error_type;

	while (current_block <> NIL) do begin

		declaration := name_table_find(current_block^.names, name);

		if (declaration <> NIL) then begin
			if (declaration^.identifier_class = variable_name) then
				found_type := type_duplicate(declaration^.variable_type);
			break;
    end;
	
		current_block := current_block^.parent_block;
	end;
	Result:= found_type;
end;

function type_get_name(basic_type: types): string;
begin

  Result:= '';

	case basic_type^.type_class of
    integer_type      : Result:= 'integer';
    real_type         : Result:= 'real';
    boolean_type      : Result:= 'boolean';
    char_type         : Result:= 'char';
    image_type        : Result:= 'image';
    command_type      : Result:= 'command';
    stream_type       : Result:= 'stream';
    string_type       : Result:= 'string';
    array_type        : Result:= 'array';
    record_type       : Result:= 'record';
    record_store_type : Result:= 'recordStore';
    http_type         : Result:= 'http';
    alert_type        : Result:= 'alert type';
    interval_type     : Result:= 'interval';
	else
		Result:= 'unknown type';
  end;

end;

//######################################################

function initialize_root_block(): block;
var
  root_block: block;
begin
	root_block := block_create(NIL,'');
  Init(root_block);
	result:= root_block;
end;

procedure Init(item: block);
begin

	add_std_type(item, integer_type, 'integer');
	add_std_type(item, real_type, 'real');
	add_std_type(item, boolean_type, 'boolean');
	add_std_type(item, char_type, 'char');
	add_std_type(item, string_type, 'string');
	add_std_type(item, image_type, 'image');
	add_std_type(item, command_type, 'command');
	add_std_type(item, stream_type, 'resource');
	add_std_type(item, record_store_type, 'recordstore');
	add_std_type(item, http_type, 'http');

add_real_constant(item, 3.1415926535897932384626433832795, 'pi');

	add_integer_constant(item, 0, 'KE_NONE');
	add_integer_constant(item, 0, 'GA_NONE');

	add_integer_constant(item, 1, 'GA_UP');
	add_integer_constant(item, 6, 'GA_DOWN');
	add_integer_constant(item, 2, 'GA_LEFT');
	add_integer_constant(item, 5, 'GA_RIGHT');
	add_integer_constant(item, 8, 'GA_FIRE');
	add_integer_constant(item, 9,  'GA_GAMEA');
	add_integer_constant(item, 10, 'GA_GAMEB');
	add_integer_constant(item, 11, 'GA_GAMEC');
	add_integer_constant(item, 12, 'GA_GAMED');

	add_integer_constant(item, 48, 'KE_KEY0');
	add_integer_constant(item, 49, 'KE_KEY1');
	add_integer_constant(item, 50, 'KE_KEY2');
	add_integer_constant(item, 51, 'KE_KEY3');
	add_integer_constant(item, 52, 'KE_KEY4');
	add_integer_constant(item, 53, 'KE_KEY5');
	add_integer_constant(item, 54, 'KE_KEY6');
	add_integer_constant(item, 55, 'KE_KEY7');
	add_integer_constant(item, 56, 'KE_KEY8');
	add_integer_constant(item, 57, 'KE_KEY9');
	add_integer_constant(item, 42, 'KE_STAR');
	add_integer_constant(item, 35, 'KE_POUND');

	add_integer_constant(item, 0, 'FONT_FACE_SYSTEM');
	add_integer_constant(item, 32, 'FONT_FACE_MONOSPACE');
	add_integer_constant(item, 64, 'FONT_FACE_PROPORTIONAL');

	add_integer_constant(item, 8, 'FONT_SIZE_SMALL');
	add_integer_constant(item, 0, 'FONT_SIZE_MEDIUM');
	add_integer_constant(item, 16, 'FONT_SIZE_LARGE');

	add_integer_constant(item, 0, 'FONT_STYLE_PLAIN');
	add_integer_constant(item, 1, 'FONT_STYLE_BOLD');
	add_integer_constant(item, 2, 'FONT_STYLE_ITALIC');
	add_integer_constant(item, 4, 'FONT_STYLE_UNDERLINED');

	add_integer_constant(item, 1, 'CM_SCREEN');
	add_integer_constant(item, 2, 'CM_BACK');
	add_integer_constant(item, 3, 'CM_CANCEL');
	add_integer_constant(item, 4, 'CM_OK');
	add_integer_constant(item, 5, 'CM_HELP');
	add_integer_constant(item, 6, 'CM_STOP');
	add_integer_constant(item, 7, 'CM_EXIT');
	add_integer_constant(item, 8, 'CM_ITEM');

	add_integer_constant(item, 0, 'TF_ANY');
	add_integer_constant(item, 1, 'TF_EMAIL');
	add_integer_constant(item, 2, 'TF_NUMERIC');
	add_integer_constant(item, 3, 'TF_PHONENUMBER');
	add_integer_constant(item, 4, 'TF_URL');
	add_integer_constant(item, $10000, 'TF_PASSWORD');

	add_integer_constant(item, 1, 'CH_EXCLUSIVE');
	add_integer_constant(item, 2, 'CH_MULTIPLE');
	add_integer_constant(item, 3, 'CH_IMPLICIT');

	add_string_constant(item, 'GET', 'GET');
	add_string_constant(item, 'HEAD', 'HEAD');
	add_string_constant(item, 'POST', 'POST');

	add_integer_constant(item, 1, 'DF_DATE');
	add_integer_constant(item, 2, 'DF_TIME');
	add_integer_constant(item, 3, 'DF_DATE_TIME');

	add_integer_constant(item, 1000, 'EOF');

  //------------------------------------------------------------------------

  add_special_function(item, [real_type], real_type, 'sqrt');
	add_special_function(item, [real_type], integer_type, 'trunc');
	add_special_function(item, [real_type], real_type, 'sin');
	add_special_function(item, [real_type], real_type, 'cos');
	add_special_function(item, [real_type], real_type, 'atan');
	add_special_function(item, [real_type], real_type, 'log');
	add_special_function(item, [real_type], real_type, 'exp');
	add_special_function(item, [real_type], real_type, 'asin');
	add_special_function(item, [real_type], real_type, 'acos');
	add_special_function(item, [real_type], real_type, 'tan');
	add_special_function(item, [real_type], real_type, 'toDegrees');
	add_special_function(item, [real_type], real_type, 'toRadians');
	add_special_function(item, [real_type], real_type, 'frac');
  add_special_function(item, [real_type], real_type, 'log10');
  add_special_function(item, [real_type], real_type, 'rabs');

	add_special_function(item, [integer_type], integer_type, 'sqr');
	add_special_function(item, [integer_type], integer_type, 'abs');

	add_special_function(item, [integer_type], boolean_type, 'odd');
	add_special_function(item, [integer_type], char_type, 'chr');
	add_special_function(item, [char_type], integer_type, 'ord');

	add_special_function(item, [void_type], void_type, 'halt');
	add_special_function(item, [void_type], boolean_type, 'isMidletPaused');

	add_special_function(item, [void_type], void_type, 'randomize');
	add_special_function(item, [integer_type], integer_type, 'random');

	add_special_function(item, [string_type], string_type, 'upcase');
	add_special_function(item, [string_type], string_type, 'locase');
	add_special_function(item, [string_type], integer_type, 'length');

	add_special_function(item, [string_type], integer_type, 'stringToInteger');
	add_special_function(item, [integer_type], string_type, 'integerToString');

  add_special_function(item, [real_type, real_type], real_type, 'atan2');
	add_special_function(item, [real_type, real_type], real_type, 'pow');
	add_special_function(item, [string_type, integer_type], real_type, 'stringToReal');
	add_special_function(item, [string_type, integer_type, integer_type], string_type, 'copy');
	add_special_function(item, [string_type, string_type], integer_type, 'pos');
	add_special_function(item, [string_type, integer_type], char_type, 'getChar');
	add_special_function(item, [string_type, char_type, integer_type], string_type, 'setChar');

	add_special_function(item, [string_type], string_type, 'getproperty');

	add_special_function(item, [integer_type], void_type, 'delay');
	add_special_function(item, [void_type], integer_type, 'getRelativeTimeMs');
	add_special_function(item, [void_type], integer_type, 'getCurrentTime');
	add_special_function(item, [integer_type], integer_type, 'getDay');
	add_special_function(item, [integer_type], integer_type, 'getMonth');
	add_special_function(item, [integer_type], integer_type, 'getYear');
	add_special_function(item, [integer_type], integer_type, 'getSecond');
	add_special_function(item, [integer_type], integer_type, 'getMinute');
	add_special_function(item, [integer_type], integer_type, 'getHour');
	add_special_function(item, [integer_type], integer_type, 'getWeekDay');
	add_special_function(item, [integer_type], integer_type, 'getYearDay');

	add_special_function(item, [integer_type, integer_type, integer_type], void_type, 'playTone');

	add_special_function(item, [void_type], integer_type, 'getKeyPressed');
	add_special_function(item, [void_type], integer_type, 'getKeyClicked');
	add_special_function(item, [integer_type], integer_type, 'keyToAction');

	add_special_function(item, [string_type], image_type, 'loadImage');
	add_special_function(item, [image_type], integer_type, 'getImageWidth');
	add_special_function(item, [image_type], integer_type, 'getImageHeight');
	add_special_function(item, [image_type, integer_type, integer_type, integer_type, integer_type], image_type, 'imageFromImage');
	add_special_function(item, [integer_type, integer_type, integer_type, integer_type], image_type, 'imageFromCanvas');
	add_special_function(item, [string_type, integer_type], image_type, 'imageFromBuffer');

	add_special_function(item, [void_type], void_type, 'repaint');

	add_special_function(item, [void_type], integer_type, 'getWidth');
	add_special_function(item, [void_type], integer_type, 'getHeight');
	add_special_function(item, [void_type], boolean_type, 'isColorDisplay');
	add_special_function(item, [void_type], integer_type, 'getColorsNum');


	add_special_function(item, [string_type, integer_type, integer_type], void_type, 'drawText');
	add_special_function(item, [integer_type, integer_type, integer_type, integer_type], void_type, 'setClip');
	add_special_function(item, [integer_type, integer_type, integer_type, integer_type], void_type, 'drawLine');
	add_special_function(item, [integer_type, integer_type, integer_type], void_type, 'setColor');
	add_special_function(item, [integer_type, integer_type, integer_type, integer_type], void_type, 'drawEllipse');
	add_special_function(item, [integer_type, integer_type, integer_type, integer_type], void_type, 'fillEllipse');
	add_special_function(item, [integer_type, integer_type, integer_type, integer_type], void_type, 'drawRect');
	add_special_function(item, [integer_type, integer_type, integer_type, integer_type], void_type, 'fillRect');
	add_special_function(item, [integer_type, integer_type, integer_type, integer_type, integer_type, integer_type], void_type, 'drawArc');
	add_special_function(item, [integer_type, integer_type], void_type, 'plot');
	add_special_function(item, [void_type], integer_type, 'getColorRed');
	add_special_function(item, [void_type], integer_type, 'getColorGreen');
	add_special_function(item, [void_type], integer_type, 'getColorBlue');
	add_special_function(item, [integer_type, integer_type, integer_type, integer_type, integer_type, integer_type], void_type, 'drawRoundRect');
	add_special_function(item, [integer_type, integer_type, integer_type, integer_type, integer_type, integer_type], void_type, 'fillRoundRect');

	add_special_function(item, [integer_type, integer_type, integer_type], void_type, 'setFont');
	add_special_function(item, [void_type], void_type, 'setDefaultFont');

	add_special_function(item, [image_type, integer_type, integer_type], void_type, 'drawImage');

	add_special_function(item, [string_type], integer_type, 'getStringWidth');
	add_special_function(item, [string_type], integer_type, 'getStringHeight');

	add_special_function(item, [string_type], void_type, 'debug');
	add_special_function(item, [boolean_type], void_type, 'assert');

	add_special_function(item, [string_type, integer_type, integer_type], command_type, 'createCommand');
	add_special_function(item, [void_type], command_type, 'getClickedCommand');
	add_special_function(item, [command_type], void_type, 'addCommand');
	add_special_function(item, [command_type], void_type, 'removeCommand');
	add_special_function(item, [void_type], command_type, 'emptyCommand');

	add_special_function(item, [void_type], void_type, 'showForm');
	add_special_function(item, [void_type], void_type, 'showCanvas');

	add_special_function(item, [string_type], void_type, 'setTicker');

	add_special_function(item, [void_type], void_type, 'clearForm');
	add_special_function(item, [integer_type], void_type, 'formRemove');
	add_special_function(item, [string_type], integer_type, 'formAddString');
	add_special_function(item, [void_type], integer_type, 'formAddSpace');
	add_special_function(item, [image_type], integer_type, 'formAddImage');
	add_special_function(item, [string_type, string_type, integer_type, integer_type], integer_type, 'formAddTextField');
	add_special_function(item, [string_type, boolean_type, integer_type, integer_type], integer_type, 'formAddGauge');
	add_special_function(item, [string_type, integer_type], integer_type, 'formAddDateField');
	add_special_function(item, [integer_type, integer_type], void_type, 'formSetDate');
	add_special_function(item, [integer_type], integer_type, 'formGetDate');

	add_special_function(item, [string_type, integer_type], integer_type, 'formAddChoice');

	add_special_function(item, [integer_type], integer_type, 'formGetValue');
	add_special_function(item, [integer_type], string_type, 'formGetText');
	add_special_function(item, [integer_type, integer_type], void_type, 'formSetValue');
	add_special_function(item, [integer_type, string_type], void_type, 'formSetText');

	add_special_function(item, [integer_type, string_type], integer_type, 'choiceAppendString');
	add_special_function(item, [integer_type, string_type, image_type], integer_type, 'choiceAppendStringImage');
	add_special_function(item, [integer_type, integer_type], boolean_type, 'choiceIsSelected');
	add_special_function(item, [integer_type], integer_type, 'choiceGetSelectedIndex');

	add_special_function(item, [string_type], void_type, 'setFormTitle');
	add_special_function(item, [void_type], void_type, 'removeFormTitle');
	add_special_function(item, [void_type], string_type, 'getFormTitle');

	add_special_function(item, [string_type, string_type, integer_type, integer_type], void_type, 'showTextBox');
	add_special_function(item, [void_type], string_type, 'getTextBoxString');

	add_special_function(item, [string_type, string_type, image_type, alert_type], void_type, 'showAlert');
	add_special_function(item, [void_type], void_type, 'playAlertSound');
	add_special_function(item, [void_type], alert_type, 'ALERT_INFO');
	add_special_function(item, [void_type], alert_type, 'ALERT_WARNING');
	add_special_function(item, [void_type], alert_type, 'ALERT_ERROR');
	add_special_function(item, [void_type], alert_type, 'ALERT_ALARM');
	add_special_function(item, [void_type], alert_type, 'ALERT_CONFIRMATION');

	add_special_function(item, [string_type, integer_type], void_type, 'showMenu');
	add_special_function(item, [string_type], integer_type, 'menuAppendString');
	add_special_function(item, [string_type, image_type], integer_type, 'menuAppendStringImage');
	add_special_function(item, [integer_type], boolean_type, 'menuIsSelected');
	add_special_function(item, [void_type], integer_type, 'menuGetSelectedIndex');

	add_special_function(item, [string_type], record_store_type, 'openRecordStore');
	add_special_function(item, [record_store_type], void_type, 'closeRecordStore');
	add_special_function(item, [string_type], void_type, 'deleteRecordStore');
	add_special_function(item, [record_store_type, integer_type], void_type, 'deleteRecordStoreEntry');
	add_special_function(item, [record_store_type, string_type], integer_type, 'addRecordStoreEntry');
	add_special_function(item, [record_store_type, integer_type], string_type, 'readRecordStoreEntry');
	add_special_function(item, [record_store_type], integer_type, 'getRecordStoreSize');
	add_special_function(item, [record_store_type, string_type, integer_type], void_type, 'modifyRecordStoreEntry');
	add_special_function(item, [record_store_type], integer_type, 'getRecordStoreNextId');

	add_special_function(item, [http_type, string_type], boolean_type, 'openHttp');
	add_special_function(item, [http_type], boolean_type, 'isHttpOpen');
	add_special_function(item, [http_type], void_type, 'closeHttp');
	add_special_function(item, [http_type, string_type, string_type], void_type, 'addHttpHeader');
	add_special_function(item, [http_type, string_type], void_type, 'setHttpMethod');
	add_special_function(item, [http_type], integer_type, 'sendHttpMessage');
	add_special_function(item, [http_type, string_type], string_type, 'getHttpHeader');
	add_special_function(item, [http_type, string_type], void_type, 'addHttpBody');
	add_special_function(item, [http_type], string_type, 'getHttpResponse');

	add_special_function(item, [string_type], stream_type, 'openResource');
	add_special_function(item, [stream_type], void_type, 'closeResource');
	add_special_function(item, [stream_type], integer_type, 'readByte');
	add_special_function(item, [stream_type], string_type, 'readLine');
	add_special_function(item, [stream_type], boolean_type, 'resourceAvailable');

	add_special_function(item, [string_type, string_type], boolean_type, 'openPlayer');
	add_special_function(item, [void_type], boolean_type, 'startPlayer');
	add_special_function(item, [void_type], void_type, 'stopPlayer');
	add_special_function(item, [integer_type], boolean_type, 'setPlayerCount');
	add_special_function(item, [void_type], integer_type, 'getPlayerDuration');

	add_special_function(item, [string_type, string_type], boolean_type, 'smsStartSend');
	add_special_function(item, [void_type], boolean_type, 'smsIsSending');
	add_special_function(item, [void_type], boolean_type, 'smsWasSuccessfull');
end;

end.
